/* ipv6_tools.c -  misc. tools for handling IPv6 addresses */

/* Written 1998-1999 by Mohsen.Souissi@inria.fr, INRIA */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>

#include <net/if.h>
#ifndef if_type
#include <net/if_var.h> 
#endif
#include <netinet/in.h>
#include <netinet/in6_var.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>
#include <netinet/if_atm.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#define COMPONENT "MARS_TOOLS"
#define IN6_ADDR_SIZE 16

int get_my_ipv6(struct in6_addr *my_ia6, int itf)
{ 
    struct in6_ifreq ifr;
    struct sockaddr_in6 *s_in = NULL;
    int s, error;
    
    bzero(&ifr, sizeof(ifr));
    (void) snprintf(ifr.ifr_name, IFNAMSIZ, ENNAME, itf);
    if ((s = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
	perror("socket(inet)");
	return(-1);
    }
    error = ioctl(s, SIOCGIFADDR6, (char *)&ifr);
    (void) close(s);
    if (error < 0) {
	perror("SIOCGIFADDR6");
	return(error);
    }
    s_in = (struct sockaddr_in6 *)&ifr.ifr_Addr;
    bcopy(&s_in->sin6_addr, my_ia6, sizeof(struct in6_addr));
    return(0);
}


int set_my_cmi(u_int16_t cmi, int itf)
{
    struct ifreq ifr;
    int s, error;

/*    (void) setuid(0); */

    bzero(&ifr, sizeof(ifr));
    (void) snprintf(ifr.ifr_name, IFNAMSIZ, ENNAME, itf);
    if ((s = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
	perror("socket(inet)");
	return(-1);
    }

    ifr.ifr_scid6[0] = MADDR6_SCP_CMI;
    ifr.ifr_scid6[1] = cmi;

    error = ioctl(s, SIOCSIFSCID6, (char *)&ifr);
    (void) close(s);
    if (error < 0) {
	perror("SIOCSIFSCID6");
	return(error);
    }
    return(0);
}

int in6_addr_cmp(struct in6_addr *a, struct in6_addr *b)
{
    int i;
    int  diff;

    for(i = 0; i < IN6_ADDR_SIZE; i++)
	if ((diff = (a->s6_addr[i] - b->s6_addr[i])) != 0)
	    return(diff);
    return(0);
}

/* return true if a is just before b (ie b = 'a+1') */
int is_just_before(struct in6_addr *a, struct in6_addr *b)
{
    int i, diff;
    
    for (i = 0; i < IN6_ADDR_SIZE - 1; i++)
	if (b->s6_addr[i] != a->s6_addr[i])
	    return(0);
    diff = b->s6_addr[i] - a->s6_addr[i];
    return (diff == 1);
}

int is_multicast_pair(struct in6_addr *min, struct in6_addr *max)
{
    int res;
    if ((min == NULL) || (max == NULL))
	return(0);
    res = ((IN6_IS_ADDR_MULTICAST(min)) 
	   && (IN6_IS_ADDR_MULTICAST(max))
	   && (in6_addr_cmp(min, max) <= 0));
    return(res);
}

/* store 'a + 1' in b, and check that b remains in the same scope as a */
int increment_addr(struct in6_addr *a, struct in6_addr *b)
{
    int i;
    u_int v = 1;
    char str_addr[64];

    for (i = IN6_ADDR_SIZE - 1; i >= 2; i--) {
	v += a->s6_addr[i];
	b->s6_addr[i] = v & 0xff;
	v >>= 8;
    }
    if (v != 0) {
	printf("increment_addr: cannot increment address %s\n",
	       inet_ntop(AF_INET6, a, str_addr, sizeof(str_addr)));
	return(-1);
    }
    b->s6_addr[1] = a->s6_addr[1];
    b->s6_addr[0] = a->s6_addr[0];
    return(0);
}


int decrement_addr(struct in6_addr *a, struct in6_addr *b)
{
    int i, j;
    char str_addr[64];

    for (i = IN6_ADDR_SIZE - 1; i >= 2; i--) {
	if (a->s6_addr[i] > 0x00) {
	    b->s6_addr[i] = a->s6_addr[i] - 1;
	    break;
	}
	b->s6_addr[i] = 0xff;
    }
    if (i == 2) {
	printf("decrement_addr: cannot decrement address %s\n",
	       inet_ntop(AF_INET6, a, str_addr, 64));
	return(-1);
    }
    for (j = i - 1; j >= 0; j--)
	b->s6_addr[i] = a->s6_addr[i];
    return(0);
}
