/* frommclient.c  - Processing of MARS messages from mars_clients (CLMB/MCS) */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>

#include <net/if.h>
#include <net/if_atm.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmsap.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_def.h"
#include "multi_data.h"
#include "mars_msg.h"
#include "tables.h"

#include "marsd_tab.h"
#include "to_mclient.h"
#include "marsd_io.h"

#define COMPONENT "FROM_MCL"


/* 
 * this routine is called 2 times for each new mars client:
 * first to allocate new structures and to initiate adding new 
 * corrsponding leaf (mars_buf != NULL),
 * second, upon confirming the "ADD_PARTY" operation by signaling, confirm the 
 * registration and send the confirmation back to MARS 
 * client (mars_buf == NULL)
 */
int register_mars_client(enum mars_opcode opcode,
			 enum mcl_type mtype,
			 ATMCALL *mcl_p2p,
			 struct svc_addr *addr, 
			 struct in6_addr *ia6,
			 u_int8_t *mars_buf,
			 int mars_len)
{
    struct mcl_entry *mcl_e, *next_mcl_e;
    struct mcl_head *mcl_h;
    struct mars_memb *memb_msg;
    u_int8_t *to_send;
    ATMCALL *Cvc;
    PARTY *new_p;
    int i, error = 0, send_len;
    u_int32_t CurrSN;
    u_int16_t flags, mcl_id = 0;

    /* get mars client entry if available */
    mcl_e = mcl_p2p->mcl_e;

    if (mcl_e != NULL) {
	
	/* 
	 * check if we are receiving a new registartion msg while we are 
	 * still in `registering' state. If that's the case ignore this msg 
	 */
	if ((mars_buf != NULL) && (mcl_e->mcl_state == mcl_registering)) {
	    diag(COMPONENT, DIAG_WARN,
		 "register_mars_client: mclient has already "
		 "started registering");
	    return(0);
	}
	
	/* 
	 * check if we are still in 'registering' state. 
	 * This is typically the case if MARSD is waiting
	 * for ADD_PARTY to be confirmed, and here's the second call for
	 * registration procedure (from signaling client code)
	 */
	if ((mars_buf == NULL) && (mcl_e->mcl_state == mcl_registering)) {
	    if ((to_send = mcl_e->to_send) == NULL) { 
		/* this shouldn't happen */
		diag(COMPONENT, DIAG_ERROR, 
		     "register_mars_client: no message to send back "
		     "to mclient");
		return(-1);
	    }
	 
	    /* copy Current Sequence Number (CurrCSN/CurrSSN) to msn field */
	    memb_msg = (struct mars_memb *) (to_send + ATMLLC_LEN);

	    CurrSN = (mtype == clmb_type) ? CurrCSN : CurrSSN;
	    memb_msg->msn = htonl(CurrSN);
	    /* send the message privately */
	    diag(COMPONENT, DIAG_DEBUG, "TO %s (MCLId = %d): %s", 
		 mcl_name[mtype], mcl_e->mcl_id, mop_name[opcode]);
	    send_len = mcl_e->send_len;
	    error = send_msg_to_mclient(mcl_p2p->sd, 
					to_send, 
					send_len);
	    free(to_send);
	    mcl_e->to_send = NULL;
	    mcl_e->send_len = 0;
	    if (error !=0)
	    {
		diag(COMPONENT, DIAG_ERROR, 
		     "register_mars_client: failed to send message %s "
		     "to %s (MCLId = %d)", 
		     mop_name[opcode], mcl_name[mtype], mcl_e->mcl_id);
		return(error);
	    }

	    /* change state */
	    mcl_e->mcl_state = mcl_registered;
	    
	    OutputPkts[opcode]++;
	    diag(COMPONENT, DIAG_DEBUG, 
		 "register_mars_client: new %s registered with [%s = %d]", 
		 mcl_name[mtype], 
		 (mtype == clmb_type) ? "CMI" : "MCSId",
		 mcl_e->mcl_id);
	    return(0);
	}
	/* continue */
    }
    
    /* build the whole message to send back to mclient */
    send_len = ATMLLC_LEN + mars_len;
    to_send = (u_int8_t *) malloc(send_len);
    bzero(to_send, send_len);
    bcopy(&llc_mars_ctl, to_send, ATMLLC_LEN);
    bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
    
    memb_msg = (struct mars_memb *) (to_send + ATMLLC_LEN);
    flags = ntohs(memb_msg->flags) | FL_COPY;
    memb_msg->flags = htons(flags);
    
    /* check if already registered mclient */
    if ((mcl_e != NULL) 
	&& (mars_buf != NULL) 
	&& (mcl_e->mcl_state == mcl_registered)) {
	
	/* copy CMI field to original JOIN message */
	if (mtype == clmb_type) {
	    memb_msg->cmi = htons(mcl_e->mcl_id);
	    CurrSN = CurrCSN;
	} else
	    CurrSN = CurrSSN;
	/* copy CSN to msn field */
	memb_msg->msn = htonl(CurrSN);
	/* send the message privately to mclient */
	diag(COMPONENT, DIAG_DEBUG, "TO %s (MCLId = %d): %s", 
	     mcl_name[mtype], mcl_e->mcl_id, mop_name[opcode]);
        error = send_msg_to_mclient(mcl_p2p->sd, to_send, send_len);
	free(to_send);
        mcl_e->to_send = NULL;
	mcl_e->send_len = 0;
	if (error !=0)
            diag(COMPONENT, DIAG_ERROR, 
                 "register_mars_client: failed to send message %s "
		 "to %s (MCLId = %d)", 
		 mop_name[opcode], mcl_name[mtype], mcl_e->mcl_id);
	else
	    OutputPkts[opcode]++;
        return(error);
    }
    
    /* 
     * it's a new mars client (CLMB/MCS)
     */
    
    /* allocate a new mcl entry and initialize it */
    mcl_e = new_mcl_entry(mtype, addr, ia6, mcl_p2p);

    /* allocate a new MARS Client ID */
    next_mcl_e = alloc_new_mcl_id(mtype);	/* modifies CurrCMI/CurrMCSId */
    
    if (mtype == clmb_type) {
	mcl_id = (u_int16_t) CurrCMI;
	Cvc = Ccvc;
    } else {
	mcl_id = (u_int16_t) CurrMCSId;
	Cvc = Scvc;
    }
    mcl_h = &MCL_Table[mtype - 1];
    mcl_e->mcl_id = mcl_id;
    /* 
     * insert the new mcl_entry at its place, list is sorted by MCLId
     * in ascending order
     */
    if (next_mcl_e == NULL)
	TAILQ_INSERT_TAIL(mcl_h, mcl_e, links);	/* list end */ 
    else
	TAILQ_INSERT_BEFORE(next_mcl_e, mcl_e, links);
	
    /*
     * initiate adding the new mclient as a leaf to 
     * the ClusterControlVC/ServerControlVC 
     */
    
    /* if it is the first party, we shall call `atmcall_setup' */
    if (Cvc == NULL) {
	error = atmcall_setup(addr, NULL, NULL, NULL, 0, NULL, mcl_e);
	if (error != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "register_mars_client: failed to initiate opening %s",
		 (mtype == clmb_type) ? "CCVC" : "SCVC");
	    return(-1);
	}
	/* assign last created atm call to CCVC/SCVC */
	if (mtype == clmb_type)
	    Cvc = Ccvc = AtmCalls;
	else 
	    Cvc = Scvc = AtmCalls;
	
	new_p = Ccvc->parties;
    } else {
	new_p = active_add_party(Cvc, addr, NULL, NULL, mcl_e);
	if (new_p == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "register_mars_client: failed to initiate adding leaf: ");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", addr->prv[i]);
	    printf("\n");
	    return(-1);
	}
    }
    
    /* 
     * store the msg to send to mclient, return and come back 
     * when SETUP/ADD_PARTY is confirmed by signaling
     */


    if (mtype == clmb_type) 
	memb_msg->cmi = htons(mcl_id);    /* cmi = 0 for MCS */
    mcl_e->to_send = to_send;
    mcl_e->send_len = send_len;
    mcl_e->mcl_state = mcl_registering;
    diag(COMPONENT, DIAG_DEBUG, 
	 "register_mars_client: registering a new %s with [%s = %d]", 
	 mcl_name[mtype], 
	 (mtype == clmb_type) ? "CMI" : "MCSId",
	 mcl_e->mcl_id);
    return(error);
}

/* 
 * this routine is called 2 times for each de-registering mars client:
 * first to initiate dropping corrsponding leaf,
 * second, upon confirming the "DROP_PARTY" operation by signaling, confirm the 
 * de-registration and send the confirmation back to cluster member
 */
int deregister_mars_client(enum mars_opcode opcode,
			   u_int8_t *mars_buf,
			   int mars_len,
			   struct mcl_entry *mcl_e)
{
    struct mars_memb *memb_msg;
    u_int8_t *to_send;
    int i, error = 0, send_len;
    u_int16_t flags;
    enum mcl_type mtype = mcl_e->mcl_type;
    
    /* 
     * check if we are receiving another de-registeration msg while 
     * in `deregistering' state. If that's the case ignore this msg 
     */
    
    if ((mars_buf != NULL) && (mcl_e->mcl_state == mcl_deregistering)) {
        diag(COMPONENT, DIAG_WARN,
             "deregister_mars_client: %s is already in 'deregistering' state", 
	     mcl_name[mtype]);
        return(0);
    }
    /* 
     * if it is the first time we call this function, we shall 
     * update mars tables : remove mars client's atm address wherever
     *  it appears and drop mars client from CCVC/SCVC
     */
    if ((to_send = mcl_e->to_send) == NULL) {
	/* change mars client state */
        mcl_e->mcl_state = mcl_deregistering;
	if (opcode == mars_leave) {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "deregister_mars_client: de-registering CLMB " 
		 "with [CMI = %d]", mcl_e->mcl_id);
	    remove_from_all_joined(mcl_e);
	} else {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "deregister_mars_client: de-registering MCS");
	    remove_from_server_map(mcl_e);
	}
	
	/* initiate dropping CLMB/MCS from CCVC/SCVC */
	error = active_drop_party(0, 0, mcl_e->party);
	if (error != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "deregister_mars_client: failed to drop leaf: ");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", mcl_e->svc_addr.prv[i]);
	    printf("\n");
	    return(error);
	}

	/* build the whole message to send back to mclient */
	send_len = ATMLLC_LEN + mars_len;
	to_send = (u_int8_t *) malloc(send_len);
	bzero(to_send, send_len);
	bcopy(&llc_mars_ctl, to_send, ATMLLC_LEN);
	bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
	memb_msg = (struct mars_memb *) (to_send + ATMLLC_LEN);
	flags = ntohs(memb_msg->flags) | FL_COPY;
	memb_msg->flags = htons(flags);

	/* 
	 * store the msg to send to mclient, return and come back when 
	 * DROP_PARTY is confirmed by signaling
	 */
	if (mtype == clmb_type) 
	    memb_msg->cmi = htons(mcl_e->mcl_id);   /* cmi = 0 for MCS */
	mcl_e->to_send = to_send;
	mcl_e->send_len = send_len;
	
	return(0);
    }

    /* 
     * it's the second time this routine is called. Normally,
     * at this stage, DROP PARTY has been confirmed by signaling
     * cut the cross references and free the whole mars client entry
     */
    
    if (mcl_e->p2p != NULL) {
	memb_msg = (struct mars_memb *) (to_send + ATMLLC_LEN);
	/* copy Sequence Number (CSN/SSN) to msn field */
	memb_msg->msn = (mtype == clmb_type) 
	    ? htonl(CurrCSN) : htonl(CurrSSN);
	diag(COMPONENT, DIAG_DEBUG, "TO %s (MCLId = %d): %s", 
	     mcl_name[mtype], mcl_e->mcl_id, mop_name[opcode]);
	send_len = mcl_e->send_len;
	error = send_msg_to_mclient(mcl_e->p2p->sd, 
				    to_send, 
				    send_len);
    }

    free(to_send);
    mcl_e->to_send = NULL;
    mcl_e->send_len = 0;

    if (error !=0)
    {
	diag(COMPONENT, DIAG_ERROR, 
	     "deregister_mars_client: failed to send message %s"
	     "to %s (MCLId = %d)",
	     mop_name[opcode], 
	     mcl_name[mtype],
	     mcl_e->mcl_id);
	return(error);
    }

    OutputPkts[opcode]++;
    return(0);
}

static int process_membership(ATMCALL *mcl_p2p, 
			      u_int8_t *mars_buf,
			      int mars_len,
			      enum mars_opcode opcode)
{
    struct mars_memb *mmsg;
    struct fixed_hdr *fxh;
    u_int16_t shtl, sstl, spln, tpln, flags, pnum;
    u_int16_t copy, fl_layer3grp;
    u_int8_t *start, *sha, *spa, *to_send, *to_send_cp;
    struct svc_addr source_atm;
    struct in6_addr source_ia6;
    struct mcl_entry *mcl_e;
    enum mcl_type mtype;
    u_int32_t CurrSN;
    struct minmax_pair *minmax;
    struct map_entry *host_map_e, *server_map_e, *new_map_e;
    char v[64], v1[64];
    int error = 0, i;
    int send_len, theo_var_len, actual_var_len;
    int is_single_grp;
    int found_s = 0, found_h = 0, removed = 0;

    /* check if MARS Client exists */
    mcl_e = mcl_p2p->mcl_e;    
    
    /* if mclient existe, check mclient type and MARS opcode consistency */
    if (mcl_e != NULL) {
	mtype = mcl_e->mcl_type;

	if ( ((mtype == clmb_type) 
	      && ((opcode == mars_mserv) || (opcode == mars_unserv))) 
	    || ((mtype == mcs_type) && 
		((opcode == mars_join) || (opcode == mars_leave)))) {
	
	    diag(COMPONENT, DIAG_ERROR, 
		 "register_mars_client: unexpected MARS opcode %s for a %s",
		 mop_name[opcode], mcl_name[mtype]);
	    return(-1);
	}
    } else {
	if ((opcode == mars_join) || (opcode == mars_leave))
	    mtype = clmb_type;
	else 
	    mtype = mcs_type;
    }

    mmsg = (struct mars_memb *) mars_buf;
    fxh = &mmsg->fxh;

    /*
     * Get parameters to compute the theoritical variable length
     */
    shtl = fxh->shtl & TL_LEN; 
    sstl = fxh->sstl & TL_LEN;
    spln = mmsg->spln & TL_LEN;  
    tpln = mmsg->tpln & TL_LEN;
    pnum = ntohs(mmsg->pnum);

    /* @@@ only IPv6 is accepted as a layer-3 protocol */
    if (spln != IN6_ADDR_LEN) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_membership: layer-3 protocol not supported yet");
	return(-1);
    }
    
    theo_var_len = shtl + sstl + spln + pnum*2*tpln;
    actual_var_len = mars_len - sizeof(struct  mars_memb);
    
    /* check it's the actual variable length we've read */
    if (actual_var_len != theo_var_len) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_membership: variable length part inconsistency");
	return(-1);
    }
    
    /* 
     * Process flags
     */
    flags = ntohs(mmsg->flags);
    
    /* @@@ only IPv6 is accepted as a layer-3 protocol */
    if (!(flags & FL_REGISTER) && (tpln != IN6_ADDR_LEN)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_membership: layer-3 protocol not supported yet");
	return(-1);
    }
    
    /* make sure 'copy' flag is reset (cf 6.1.2) */
    copy = flags & FL_COPY;
    if (copy != 0) {
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_membership: received membership with a non "
	     "reset copy flag");
	return(0);	/* drop the message silently */
    }
    
    if (pnum > 1) {
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_membership: received membership with more than "
	     "one min_max pair");
	return(0);		/* drop the message silently */
    }
    
    /* get ATM source address */
    start = (u_int8_t *) (mars_buf + sizeof(struct mars_memb));
    sha = get_info(&start, shtl); 
    spa = get_info(&start, spln); 
    bzero(&source_atm, sizeof(source_atm));
    bcopy(sha, source_atm.prv, shtl);
    bzero(&source_ia6, sizeof(source_ia6));
    bcopy(spa, &source_ia6, spln);
    
    /* check if it is a registration/deregistration message */
    if ((flags & FL_REGISTER) && 
	((opcode == mars_join) || (opcode == mars_mserv))) {
	error = register_mars_client(opcode,
				     mtype,
				     mcl_p2p, 
				     &source_atm, 
				     &source_ia6,
				     mars_buf, 
				     mars_len);
	return(error);
    }
    
    if (mcl_e == NULL) {
	/* drop the message silently, cf 6.1.2 */
	diag(COMPONENT, DIAG_DEBUG, "No such a MARS Client: ");
	for (i = 0; i < ATM_ESA_LEN; i++)
	    printf("%02X", source_atm.prv[i]);
	printf("\n");
	return(0);
    }
    
    if (mcl_e->mcl_state != mcl_registered) {
	/* drop the message silently, cf 6.1.2 */
	diag(COMPONENT, DIAG_DEBUG, 
	     "%s (MCLId = %d) is not in `registered' state, message ignored",
	     mcl_name[mtype], mcl_e->mcl_id);
	for (i = 0; i < ATM_ESA_LEN; i++)
	    printf("%02X", source_atm.prv[i]);
	printf("\n");
	return(0);
    }
    
    /* is it a de-registration message ? */
    if ((flags & FL_REGISTER) && 
	((opcode == mars_leave) || (opcode == mars_unserv))) {
	error = deregister_mars_client(opcode, 
				       mars_buf,
				       mars_len,
				       mcl_e);
	return(error);
    }
    
    /* Now, it's a real {JOIN, LEAVE, MSERV, UNSERV} message */
    if (pnum == 0) {
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_membership: no min_max pair has been specified");
	return(0);		/* drop the message silently */
    }
    
    /* get the min_max pair (only a single pair may be accepted */
    minmax = get_info(&start, sizeof(struct minmax_pair));
    if (!is_multicast_pair(&minmax->min, &minmax->max)) {
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_membership: not a multicast minmax pair :  [%s, %s]",
	     inet_ntop(AF_INET6, &minmax->min, v, sizeof(v)),
	     inet_ntop(AF_INET6, &minmax->max, v1, sizeof(v1)));
	return(0);		/* drop the message silently */
    }
    
    /* 
     * build the message to send back to mclient(s). 
     * We don't know yet if it is to be sent over p2p or CCVC or SCVC
     */
    send_len = ATMLLC_LEN + mars_len;
    to_send = (u_int8_t *) malloc(send_len);
    bzero(to_send, send_len);
    bcopy(&llc_mars_ctl, to_send, ATMLLC_LEN);
    flags = ntohs(mmsg->flags) | FL_COPY;
    CurrSN = (mtype == clmb_type) ? CurrCSN : CurrSSN; 

    /* check if duplicate membership request */
    switch (opcode) {
    case mars_join:
    case mars_mserv:
	/* 
	 * check if the minmax pair is encompassed by the pairs already 
	 * joined (supported) by CLMB (MCS) (e.g. duplicate JOIN/MSERV).
	 */
	if (is_pair_encompassed(minmax, &mcl_e->minmax_head) != 0) {
	    /* send the message back to mclient privately */
	    diag(COMPONENT, DIAG_DEBUG, 
		 "TO %s (MCLId = %d): %s. Minmax pair [%s, %s] is "
		 "already %s by %s", 
		 mcl_name[mtype],
		 mcl_e->mcl_id,
		 mop_name[opcode],
		 inet_ntop(AF_INET6, &minmax->min, v, sizeof(v)),
		 inet_ntop(AF_INET6, &minmax->max, v1, sizeof(v1)),
		 (mtype == clmb_type) ? "joined" : "supported",
		 mcl_name[mtype]);
	    flags &= ~FL_PUNCHED;	/* make sure it's reset */
	    mmsg->flags = htons(flags);
	    mmsg->msn = htonl(CurrSN);
	    bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
	    error = send_msg_to_mclient(mcl_p2p->sd, 
					to_send, 
					send_len);
	    free(to_send);
	    if (error !=0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_membership: failed to send %s messsage "
		     "to %s (MCLId = %d)",
		     mop_name[opcode], 
		     mcl_name[mtype],
		     mcl_e->mcl_id);
		return(-1);
	    }
	    OutputPkts[opcode]++;
	    return(0);
	}
	break;
    case mars_leave:
    case mars_unserv:
	/* 
	 * check if the minmax pair no longer exists in CLMB's (MCS's) joined
	 * (supported) list (e.g. duplicate LEAVE/UNSERV).
	 */
	if ((TAILQ_FIRST(&mcl_e->minmax_head) == NULL) ||
	    (is_pair_excluded(minmax, &mcl_e->minmax_head) != 0)) {
	    /* send the message back to mclient privately */
	    diag(COMPONENT, DIAG_DEBUG,
		 "TO %s (MCLId = %d): %s. Minmax pair [%s, %s] is not "
		 "%s by %s", 
		 mcl_name[mtype],
		 mcl_e->mcl_id,
		 mop_name[opcode],
		 inet_ntop(AF_INET6, &minmax->min, v, sizeof(v)),
		 inet_ntop(AF_INET6, &minmax->max, v1, sizeof(v1)),
		 (mtype == clmb_type) ? "joined" : "supported",
		 mcl_name[mtype]);
	    flags &= ~FL_PUNCHED;	/* make sure it's reset */
	    mmsg->flags = htons(flags);
	    mmsg->msn = htonl(CurrSN);
	    bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
	    error = send_msg_to_mclient(mcl_p2p->sd, 
					to_send, 
					send_len);

	    bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
	    error = send_msg_to_mclient(mcl_p2p->sd, 
					to_send, 
					send_len);
	    free(to_send);
	    if (error !=0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_membership: failed to send message %s "
		     "to %s (MCLId = %d)", 
		     mop_name[opcode], mcl_name[mtype], mcl_e->mcl_id);
		return(-1);
	    }
	    OutputPkts[opcode]++;
	    return(0);
	}
	break;
    default:
	 diag(COMPONENT, DIAG_ERROR, 
	      "process_membership: unexpected MARS operation code: %s", 
	      mop_name[opcode]);
	return(-1);
    }

	
    /* Now, we know it's a new {JOIN, LEAVE, MSERV, UNSERV} message  */

    /* check if a single group is being joined / left / mserved / unserved */
    is_single_grp = (in6_addr_cmp(&minmax->min, &minmax->max) == 0);
    fl_layer3grp = flags & FL_LAYER3GRP;
    
    if (is_single_grp) {
	server_map_e = lookup_map_entry(&Map_Table[mcs_type - 1], 
					&minmax->min, 
					NULL, 
					&found_s);
	host_map_e = lookup_map_entry(&Map_Table[clmb_type - 1], 
				      &minmax->min, 
				      NULL, 
				      &found_h);

	switch (opcode) {

	case mars_join:
	    /* update CLMB joined minmax pairs list */
	    insert_new_mgroup(&mcl_e->minmax_head, &minmax->min, fl_layer3grp);
	    
	    /* is there a matching host_map_entry for that group ? */
	    if (found_h) {
		update_host_map_entry(&source_atm, 
				      host_map_e, 
				      mars_join, 
				      &removed);
		
	    } else {
		/* create a new map entry */
		new_map_e = new_map_entry(&minmax->min, &source_atm);
		if (host_map_e != NULL) {
		    /* 
		     * insert the new map entry just after the 
		     * entry returned by lookup function
		     */
		    TAILQ_INSERT_AFTER(&Map_Table[clmb_type - 1], 
				       host_map_e, 
				       new_map_e, 
				       links);
		} else {
		    /* 
		     * either Host Map_Table is empty or the new entry 
		     * is to be added at the end
		     */
		    TAILQ_INSERT_HEAD(&Map_Table[clmb_type - 1], 
				      new_map_e, 
				      links);
		}
	    }

	    /* is there a server_map_entry for that group ? */
	    if (found_s) {
		/*
		 * modified response to MARS_JOIN (cf 6.2.4)
		 * transform the JOIN message into a SJOIN message and 
		 * send it over SCVC
		 */
		fxh->op = htons((u_int16_t) mars_sjoin);
		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = htonl(CurrSSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		diag(COMPONENT, DIAG_DEBUG, 
		     "To All MCSs: %s", mop_name[mars_sjoin]);
		error = send_msg_to_all(scvc_sd, to_send, send_len);
		if (error !=0)
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s on SCVC",
			 mop_name[mars_sjoin]);
		else {
		    CurrSSN++;
		    OutputPkts[mars_sjoin]++;
		}
		
		/* send the original message back to CLMB privately */
		fxh->op = htons((u_int16_t) mars_join);
		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = htonl(CurrCSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		diag(COMPONENT, DIAG_DEBUG, 
		     "To CLMB (CMI = %d): %s", mcl_e->mcl_id, mop_name[opcode]);
		error = send_msg_to_mclient(mcl_p2p->sd, to_send, send_len);
		free(to_send);
		if (error !=0)
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s "
			 "to CLMB (CMI = %d)",
			 mop_name[opcode], mcl_e->mcl_id);
		else 
		    OutputPkts[opcode]++;
		
	    } else { /* no server_map_entry for the group */
		/* send the original JOIN message over CCVC */
		diag(COMPONENT, DIAG_DEBUG, 
		     "To all CLMBs: %s", mop_name[opcode]);

		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = htonl(CurrCSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		error = send_msg_to_all(ccvc_sd, to_send, send_len);
		free(to_send);
		if (error !=0)
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s on CCVC",
			 mop_name[opcode]);
		else {
		    CurrCSN++;
		    OutputPkts[opcode]++;
		}
	    }
	    
	    return(error);
	    
	case mars_leave:

	    /* update CLMB joined minmax pairs list */
	    error = remove_mgroup(&mcl_e->minmax_head, &minmax->min);
	    if (error != 0) {
		diag(COMPONENT, DIAG_WARN,
		     "process_membership: WARN: got trouble when updating "
		     "CLMB's min_max pairs");
		return(0);
	    }

	    if (!found_h) {
		diag(COMPONENT, DIAG_WARN,
		     "process_membership: no matching host map entry");
		return(0);
	    }
	    
	    update_host_map_entry(&source_atm, 
				  host_map_e, 
				  mars_leave, 
				  &removed);
	    /* is there a server_map_entry for that group ? */
	    if (found_s) {
		/*
		 * modified response to MARS_LEAVE (cf 6.2.4)
		 * transform the LEAVE message into a SLEAVE message and 
		 * send it over SCVC
		 */
		fxh->op = htons((u_int16_t) mars_sleave);
		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = htonl(CurrSSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		diag(COMPONENT, DIAG_DEBUG, 
		     "To all MCSs: %s", mop_name[mars_sleave]);
		error = send_msg_to_all(scvc_sd, to_send, send_len);
		if (error !=0)
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s on SCVC",
			 mop_name[mars_sleave]);
		else {
		    CurrSSN++;
		    OutputPkts[mars_sleave]++;
		}

		/* send the original message back to CLMB privately */
		fxh->op = htons((u_int16_t) mars_leave);
		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = htonl(CurrCSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		diag(COMPONENT, DIAG_DEBUG, 
		     "To CLMB (CMI = %d): %s", mcl_e->mcl_id, mop_name[opcode]);
		error = send_msg_to_mclient(mcl_p2p->sd, to_send, send_len);
		free(to_send);
		if (error !=0)
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s "
			 "to CLMB (CMI = %d)", 
			 mop_name[opcode], mcl_e->mcl_id);
		else
		    OutputPkts[opcode]++;
		
	    } else {
		/* send the original LEAVE message over CCVC */
		diag(COMPONENT, DIAG_DEBUG, 
		     "To all CLMBs: %s", mop_name[opcode]);
		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = htonl(CurrCSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		error = send_msg_to_all(ccvc_sd, to_send, send_len);
		free(to_send);
		if (error !=0)
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s on CCVC",
			 mop_name[opcode]);
		else {
		    CurrCSN++;
		    OutputPkts[opcode]++;
		}
	    }
	    return(error);

	case mars_mserv:

	    /* is there a server_map_entry for that group ? */
	    if (found_s) {

		/* 
		 * RFC 2022 specifies that this MCS address will be added
		 * to the server map list. But that assumes some 
		 * synchronization between different MCSs. 
		 */

		/* for the moment only one MCS can support a given mgroup */
		diag(COMPONENT, DIAG_DEBUG,
		     "process_membership: multicast group [%s] is already "
		     "supported by another MCS",
		     inet_ntop(AF_INET6, &minmax->min, v, sizeof(v)));

		/* 
		 * send the MSERV message to MCS privately.
		 * Actually, it is supposed to be transmitted on SCVC.
		 * This isn't specified by RFC, but the underlying idea is
		 * to inform MCS that "mserv request" hasn't been taken into
		 * account.
		 */
		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = htonl(CurrSSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		diag(COMPONENT, DIAG_DEBUG, 
		     "To MCS (MCSId = %d): %s", 
		     mcl_e->mcl_id, mop_name[opcode]);
		error = send_msg_to_mclient(mcl_p2p->sd, to_send, send_len);
		free(to_send);
		if (error !=0)
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s "
			 "to MCS (MCSId = %d)",
			 mop_name[opcode], 
			 mcl_e->mcl_id);
		else
		    OutputPkts[opcode]++;

		return(error);
	    }
	    
	    /* update MCS mserved minmax pairs list */
	    insert_new_mgroup(&mcl_e->minmax_head, &minmax->min, fl_layer3grp);

	    /* create a new server map entry */
	    new_map_e = new_map_entry(&minmax->min, &source_atm);
	    if (server_map_e != NULL) {
		/* 
		 * insert the new map entry just after the 
		 * entry returned by lookup function
		 */
		TAILQ_INSERT_AFTER(&Map_Table[mcs_type - 1], 
				   server_map_e, 
				   new_map_e, 
				   links);
	    } else {
		/* 
		 * either Server Map_Table is empty or the new entry 
		 * is to be added at the end
		 */
		TAILQ_INSERT_HEAD(&Map_Table[mcs_type - 1], new_map_e, links);
	    }

	    /* 
	     * if a non null host map entry exists for the group, generate 
	     * a MARS_MIGRATE message and send it on CCVC (cf 6.2.2) 
	     */
	    if (found_h) {
		error = send_migrate(new_map_e);
		if (error != 0) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: send_migrate failed");
		    free(to_send);
		    return(error);
		}
		OutputPkts[mars_migrate]++;
	    }

	    /* send MARS_MSERV over SCVC (cf 6.2.2) */
	    flags &= ~FL_PUNCHED;	/* make sure it's reset */
	    mmsg->flags = htons(flags);
	    mmsg->msn = htonl(CurrSSN);
	    bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
	    diag(COMPONENT, DIAG_DEBUG, 
		 "To all MCSs: %s", mop_name[opcode]);
	    error = send_msg_to_all(scvc_sd, to_send, send_len);
	    free(to_send);
	    if (error !=0)
		diag(COMPONENT, DIAG_ERROR, 
		     "process_membership: failed to send msg %s to all MCSs",
		     mop_name[opcode]);
	    else {
		CurrSSN++;
		OutputPkts[opcode]++;
	    }
	    
	    return(error);
	    
	case mars_unserv:

	    /* update MCS mserved minmax pairs list */
	    error = remove_mgroup(&mcl_e->minmax_head, &minmax->min);
	    if (error != 0) {
		diag(COMPONENT, DIAG_WARN,
		     "process_membership: WARN: got trouble when updating MCS "
		     "min_max pairs");
		return(0);
	    }

	    /* is there a server_map_entry for that group ? */
	    if (!found_s) {
		diag(COMPONENT, DIAG_WARN,
		     "process_membership: no matching server map entry");
		return(0);
	    }	

	    /* is there a host_map_entry for that group ? */

	    if (found_h) {
		/*
		 * transform the UNSERV message into a LEAVE message and 
		 * send it over CCVC (cf 6.2.2)
		 */
		fxh->op = htons((u_int16_t) mars_leave);
		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = htonl(CurrCSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		diag(COMPONENT, DIAG_DEBUG, 
		     "To all CLMBs: %s", mop_name[mars_leave]);
		error = send_msg_to_all(ccvc_sd, to_send, send_len);
		if (error !=0) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s on CCVC",
			 mop_name[mars_leave]);
		    free(to_send);
		    return(error);
		} else {
		    CurrCSN++;
		    OutputPkts[mars_leave]++;
		}
	    }
	    
	    /* send original MARS_UNSERV back over SCVC (cf 6.2.2) */
	    fxh->op = htons((u_int16_t) mars_unserv);
	    flags &= ~FL_PUNCHED;	/* make sure it's reset */
	    mmsg->flags = htons(flags);
	    mmsg->msn = htonl(CurrSSN);
	    bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
	    diag(COMPONENT, DIAG_DEBUG, 
		 "To all MCSs: %s", mop_name[opcode]);
	    error = send_msg_to_all(scvc_sd, to_send, send_len);
	    free(to_send);
	    if (error !=0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_membership: failed to send msg %s to all MCSs",
		     mop_name[opcode]);
		return(error);
	    }
	    
	    CurrSSN++;
	    OutputPkts[opcode]++;
	    
	    /* 
	     * remove serve map entry from table, since only one MCS 
	     * may support a given multicast group, for the moment 
	     */
	    TAILQ_REMOVE(&Map_Table[mcs_type - 1], server_map_e, links);
	    
	    return(error);

	default:
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_membership: unexpected MARS operation code: %s", 
		 mop_name[opcode]);
	    return(-1);
	    
	}
	
    } else {
	/* 
	 * a block of multiple addresses is being 
	 * joined / left / mserved / unserved
	 */
	
	struct in6_addr min_multi, max_multi;
	int ret1, ret2, cmp1, cmp2, send_len_cp;
	struct minmax_entry *walk, *next;
	
	switch (opcode) {
	case mars_join:
	case mars_mserv:
	    
	    diag(COMPONENT, DIAG_WARN,
		 "process_membership: multiple group %s not supported yet",
		 mop_name[opcode]);
	    break;
	    
	case mars_leave:
	case mars_unserv:

	    ret1 = inet_pton(AF_INET6, MIN_MULTI_ADDR, &min_multi);
	    ret2 = inet_pton(AF_INET6, MAX_MULTI_ADDR, &max_multi);
	    if ((ret1 < 0) || (ret2 < 0)) {
		diag(COMPONENT, DIAG_ERROR,
		     "process_membership: inet_pton: %s\n", strerror(errno));
		return(-1);
	    }
	    if ((ret1 == 0) || (ret2 == 0)) {
		/* paranoid, unlikely to happen */
		diag(COMPONENT, DIAG_ERROR, 
		     "process_membership: unparsable multicast min_max pair");
		return(-1);
	    }
	    cmp1 = in6_addr_cmp(&minmax->min, &min_multi);
	    cmp2 = in6_addr_cmp(&minmax->max, &max_multi);
	    if ((cmp1 ==0) && (cmp2 == 0)) {
		/* 
		 * mars client is leaving all that it has joined/mserved
		 */
		
		/* 
		 * if it is a MCS, remove all corresponding map entries 
		 * from Server Map Table
		 */
		if (mtype == mcs_type)
		    remove_from_server_map(mcl_e);

		/* 
		 * build a 'hole punched' MARS_LEAVE message and 
		 * send it on CCVC
		 */

		/* how many minmax pairs are there */
		pnum = 0;
		for (walk = TAILQ_FIRST(&mcl_e->minmax_head); 
		     walk; 
		     walk = TAILQ_NEXT(walk, links))
		    pnum++;
		
		/* 
		 * allocate a MARS_LEAVE message big enough to hold all 
		 * minmax pairs
		 */
		send_len_cp = send_len + 2*(pnum - 1)*tpln;
		to_send_cp = (u_int8_t *) malloc(send_len_cp);
		bzero(to_send_cp, send_len_cp);
		bcopy(&llc_mars_ctl, to_send_cp, ATMLLC_LEN);
		bcopy(mars_buf, to_send_cp + ATMLLC_LEN, mars_len);
		mmsg = (struct mars_memb *) (to_send_cp + ATMLLC_LEN);
		fxh = &mmsg->fxh;
		/* force opcode to MARS_LEAVE */
		fxh->op = htons((u_int16_t) mars_leave);
		flags |= FL_PUNCHED;
		mmsg->flags = htons(flags);
		mmsg->pnum = htons(pnum);
		mmsg->msn = htonl(CurrCSN);
		start = to_send_cp + ATMLLC_LEN + sizeof(struct mars_memb) 
		    + shtl + sstl + spln;

		/* put min_max pair in message and free min_max entry */
		for (walk = TAILQ_FIRST(&mcl_e->minmax_head); 
		     walk; 
		     walk = next) {
		    
		    next = TAILQ_NEXT(walk, links);
		    set_info(&start, &walk->pair, sizeof(struct minmax_pair));
		    TAILQ_REMOVE(&mcl_e->minmax_head, walk, links);
		    free(walk);
		}

		/* send the message on CCVC */
		diag(COMPONENT, DIAG_DEBUG, "To all CLMBs: MARS_LEAVE");
		error = send_msg_to_all(ccvc_sd, to_send_cp, send_len_cp);
		free(to_send_cp);
		if (error !=0) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send message "
			 "MARS_LEAVE on CCVC");
		    return(error);
		} else {
		    CurrCSN++;
		    OutputPkts[mars_leave]++;
		}
		

		/* 
		 * now send original mars_leave/mars_unserv back to 
		 * mars client privately
		 */
		mmsg = (struct mars_memb *) mars_buf;
		flags &= ~FL_PUNCHED;	/* make sure it's reset */
		mmsg->flags = htons(flags);
		mmsg->msn = (opcode == mars_leave) 
		    ? htonl(CurrCSN) : htonl(CurrSSN);
		bcopy(mars_buf, to_send + ATMLLC_LEN, mars_len);
		diag(COMPONENT, DIAG_DEBUG, 
		     "To %s (MCLId = %d): %s", 
		     mcl_name[mtype], mcl_e->mcl_id, mop_name[opcode]);
		error = send_msg_to_mclient(mcl_p2p->sd, to_send, send_len);
		free(to_send);
		if (error !=0)
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_membership: failed to send msg %s "
			 "to %s (MCLId = %d)",
			 mop_name[opcode], mcl_name[mtype], mcl_e->mcl_id);
		else
		    OutputPkts[opcode]++;

		return(error);
	    }
	default:
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_membership: unexpected MARS operation code: %s", 
		 mop_name[opcode]);
	    return(-1);
	}
    }
    
    return(error);
}



/* 
 * MARS_REQUEST may be sent either by a CLMB or by a MCS.
 * Depending on sender, marsd returns one or multiple mars_multi messages 
 * filled with either host map entry content or server map entry 
 * content (if any). (cf 6.2.1) 
 */
static int process_request(ATMCALL *mcl_p2p, 
			   u_int8_t *buffer, 
			   int buf_len)
{
    u_int8_t *mars_buf;
    struct mars_req *msg;
    struct fixed_hdr *fxh;
    int mars_len;
    u_int16_t shtl, sstl, thtl, tstl, spln, tpln;
    u_int8_t *start, *sha, *ssa, *spa, *tpa = NULL;
    int i, error = 0;
    struct svc_addr source_atm;
    struct in6_addr ip6m;
    struct mcl_entry *mcl_e;
    struct map_entry *host_map_e, *server_map_e;
    enum mcl_type mtype;
    struct svc_addr_entry *addr_e;
    char v[64];
    int theo_var_len, actual_var_len;
    int found_h = 0, found_s;


    mars_buf = (u_int8_t *) (buffer + ATMLLC_LEN);
    mars_len = buf_len - ATMLLC_LEN;
    msg = (struct mars_req *) mars_buf;
    fxh = &msg->fxh;

    /*
     * Get parameters to compute the message variable length
     */
    shtl = fxh->shtl & TL_LEN; 
    sstl = fxh->sstl & TL_LEN;
    thtl = msg->thtl & TL_LEN; 
    tstl = msg->tstl & TL_LEN; 
    spln = msg->spln & TL_LEN;  
    tpln = msg->tpln & TL_LEN;

    /* @@@ disable sub-address field temporarily */
    if (thtl || tstl) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_request: target hardware address must be null");
	return(-1);
    }

    /* @@@ only IPv6 is accepted as a layer-3 protocol */
    if ((spln != IN6_ADDR_LEN) || (tpln != IN6_ADDR_LEN)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_request: layer-3 protocol not supported yet");
	return(-1);
    }

    theo_var_len =  shtl + sstl + spln + tpln;
    actual_var_len = mars_len - sizeof(struct  mars_req);

    /* check it's the actual variable length we've read */
    if (actual_var_len != theo_var_len) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_request: variable length part inconsistency");
	return(-1);
    }

    start = (u_int8_t *) (mars_buf + sizeof(struct mars_req));
    sha = get_info(&start, shtl); 
    ssa = get_info(&start, sstl);
    spa = get_info(&start, spln); 
    tpa = get_info(&start, tpln);

    bzero(&source_atm, sizeof(struct svc_addr));
    bcopy(sha, source_atm.prv, ATM_ESA_LEN);
    bcopy(tpa, &ip6m, sizeof(struct in6_addr));

    /* check the message is coming from a registered mclient (CLMB/MCS) */
    mcl_e = mcl_p2p->mcl_e;
    if (mcl_e == NULL) {
	/* drop and ignore message, cf 6.1.1 */
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_request: No such a cluster member: ");
	for (i = 0; i < ATM_ESA_LEN; i++)
	    printf("%02X", source_atm.prv[i]);
	printf("\n");
	return(0);
    }

    mtype = mcl_e->mcl_type;

    /* lookup the matching server map entry (if any) */
    server_map_e = lookup_map_entry(&Map_Table[mcs_type - 1], 
				    &ip6m,
				    NULL, 
				    &found_s);

    /* if the requester is a MCS */
    if (mtype == mcs_type) {
	if (!found_s) {
	
	    /* drop the message silently and ignore it */
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_request: no matching server map_entry for %s ", 
		 inet_ntop(AF_INET6, &ip6m, v, sizeof(v)));
	    return(0);
	}
	/* 
	 * if MCS does not belong to the 
	 * corresponding server map entry then drop the message and 
	 * ignore it (cf 6.2.1) 
	 */

	/* 
	 * for the moment, server map entry is supposed to contain only 
	 * one server ATM address. Let's compare it to the requester's addr
	 */

	addr_e = LIST_FIRST(&server_map_e->addr_head);
	if (!atm_equal(&source_atm, &addr_e->addr, 0, 0)) {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_request: MCS is not a member of the server "
		 "map for the  group [%s]: ", 
		 inet_ntop(AF_INET6, &ip6m, v, sizeof(v)));
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", source_atm.prv[i]);
	    printf("\n");
	    return(0);
	}
    }
    
    /* lookup the matching host map entry (if any) */
    host_map_e = lookup_map_entry(&Map_Table[clmb_type - 1], 
				  &ip6m, 
				  NULL, 
				  &found_h);

    /* if no host map entry was found */ 
    if (!found_h) {
	
	/* 
	 * transform the request message into a `mars_nak' and send 
	 * it to mclient privately
	 */
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_request: no matching host map_entry for %s ", 
	     inet_ntop(AF_INET6, &ip6m, v, sizeof(v)));
	fxh->op = htons((u_int16_t) mars_nak);
	diag(COMPONENT, DIAG_DEBUG, "TO %s (MCLId = %d): %s", 
	     mcl_name[mtype], mcl_e->mcl_id, mop_name[mars_nak]);
	error = send_msg_to_mclient(mcl_p2p->sd, buffer, buf_len);
	if (error  != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_request: failed to send message %s ", 
		 "to %s (MCLId = %d)", 
		 mop_name[mars_nak], mcl_name[mtype], mcl_e->mcl_id);
	    return(-1);
	}

	OutputPkts[mars_nak]++;
	return(0);
    }

    if (mtype == clmb_type) {
	if (!found_s) {
	    /* 
	     * create and send a mars_multi containing the host map
	     * privately back to CLMB
	     */
	    error = send_map_list(mcl_p2p, msg, sha, ssa, spa, host_map_e);
	    if (error  != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "process_request: failed to send host map list "
		     "to CLMB (CMI = %d)", mcl_e->mcl_id);
		return(-1);
	    }

	    OutputPkts[mars_multi]++;
	    return(0);
	}
	/* 
	 * create and send a mars_multi containing the server map
	 * privately back to CLMB
	 */
	error = send_map_list(mcl_p2p, msg, sha, ssa, spa, server_map_e);
	if (error  != 0) {
	    diag(COMPONENT, DIAG_ERROR,
		 "process_request: failed to send server map list "
		 "to CLMB (CMI = %d)", mcl_e->mcl_id);
	    return(-1);
	}
	
	OutputPkts[mars_multi]++;
	return(0);
    }

    /* 
     * now mclient is actually a MCS and there are both server and host map 
     * entries for the group in question. 
     */
    
    /* 
     * create and send a mars_multi containing the host map privately to MCS 
     */
    
    error = send_map_list(mcl_p2p, msg, sha, ssa, spa, host_map_e);
    if (error  != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "process_request: failed to send host map list "
	     "to MCS (MCSId = %d)", mcl_e->mcl_id);
	return(-1);
    }

    OutputPkts[mars_multi]++;
    return(0);
}

static int is_legal_destination(struct in6_addr *dst)
{
    return(1);
}

static ATMCALL **get_p2p_active(struct map_entry *map_e)
{
    struct svc_addr_entry *addr_e;
    struct mcl_entry *mcl_e;
    int i, j, size_tab = 4;
    ATMCALL **p2p_tab, **new_tab;

    p2p_tab = (ATMCALL **) malloc(size_tab*(sizeof(ATMCALL **)));
    bzero(p2p_tab, size_tab*(sizeof(ATMCALL **)));

    for (addr_e = LIST_FIRST(&map_e->addr_head), i = 0; 
	 addr_e; 
	 addr_e = LIST_NEXT(addr_e, links), i++) {
	
	mcl_e = lookup_mcl_entry(&addr_e->addr, &MCL_Table[clmb_type - 1]);
	if (mcl_e == NULL) {
	    diag(COMPONENT, DIAG_ERROR,
		 "get_p2p_active: no matching CLMB for address:");
	    for (j = 0; i < ATM_ESA_LEN; j++)
		printf("%02x", addr_e->addr.prv[j]);
	    printf("\n");
	    free(p2p_tab);
	    return(NULL);
	}
	
	if ((mcl_e->p2p == NULL) || 
	    (mcl_e->p2p->sd < 0) || 
	    (mcl_e->p2p->state != connected)) {

	    free(p2p_tab);
	    return(NULL);
	}

	if (i >= size_tab) {
	    new_tab = (ATMCALL **) malloc(2*size_tab*(sizeof(ATMCALL **)));
	    bzero(new_tab, 2*size_tab*(sizeof(ATMCALL **)));
	    bcopy(p2p_tab, new_tab, size_tab*(sizeof(ATMCALL **)));
	    free(p2p_tab);
	    size_tab =  2*size_tab;
	    p2p_tab = new_tab;
	}
	
	p2p_tab[i] = mcl_e->p2p;
    }

    return(p2p_tab);
}

static int process_multi_ipv6(u_int8_t *mbuffer, int size)
{
    int min_len;
    struct ip6_hdr *ip6h;
    struct in6_addr *dst;
    struct map_entry *map_e;
    ATMCALL *p2p_call, **p2p_tab;
    int found = 0, wrote, i;
    char v[64];

    min_len =  ATMMLLC_LEN + sizeof(struct ip6_hdr);
    if (size < min_len) {
	diag(COMPONENT, DIAG_ERROR,
	     "process_multi_ipv6: packet too short (got %d < %d)", 
	     size, min_len);
	NbInputErrors++;
	return(-1);
    }
    
    ip6h = (struct ip6_hdr *) (mbuffer + ATMMLLC_LEN);
    dst = &ip6h->ip6_dst;

    if (!is_legal_destination(dst)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_multi_ipv6: illegal destination %s",
	     inet_ntop(AF_INET6, dst, v, sizeof(v)));
	NbInputErrors++;
	return(-1);
    }

    InputPkts[mData]++;
    /* lookup matching host map entry (if any) */
    map_e = lookup_map_entry(&Map_Table[clmb_type - 1], dst, NULL, &found);
    if (!found)
	goto broadcast;

    
    /* check we have an active p2p VC to each member of mgroup in question */
    p2p_tab = get_p2p_active(map_e);
    if (p2p_tab == NULL)
	goto broadcast;

    /* send a copy of the present message to each CLMB privately */
    for (i = 0, p2p_call = p2p_tab[0]; 
	 p2p_call != NULL; 
	 p2p_call = p2p_tab[++i]) {

	diag(COMPONENT, DIAG_DEBUG, 
	     "sending Multicast Data to CLMB (CMI = %d):", 
	     p2p_call->mcl_e ? p2p_call->mcl_e->mcl_id : -1);
	if ((wrote = write(p2p_call->sd, mbuffer, size)) == size) {
	    OutputPkts[mData]++;
	    NbOutput++;
	    continue;
	}
	if (wrote < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_multi_ipv6: to CLMB (CMI = %d) write: %s", 
		 p2p_call->mcl_e ? p2p_call->mcl_e->mcl_id : -1,
		 strerror(errno));
	    NbOutputErrors++;
	    return(wrote);
	}

	diag(COMPONENT, DIAG_ERROR, 
	     "process_multi_ipv6: bad write: wanted %d, wrote %d\n", 
	     size, wrote);
	NbOutputErrors++;
	NbOutput++;
	return(wrote);
    }


 broadcast:
    /* forward the present message on CCVC */    
    if ((wrote = write(Ccvc->sd, mbuffer, size)) == size) {
	OutputPkts[AllNodes]++;
	NbOutput++;
	return(0);
    }
    if (wrote < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_multi_ipv6: write: %s", strerror(errno));
	NbOutputErrors++;
	return(wrote);
    }
    
    diag(COMPONENT, DIAG_ERROR, 
	 "process_multi_ipv6: bad write: wanted %d, wrote %d\n", size, wrote);
    NbOutputErrors++;
    NbOutput++;
    return(wrote);
}

int dispatch_mars_msg(ATMCALL *mcl_p2p, u_int8_t *buffer, int buf_len)
{
    struct atmllc *alc;
    struct atmmllc *amlc;
    u_int8_t *mars_buf;
    struct fixed_hdr *fxh;
    enum mars_opcode opcode;
    int error, mars_len;
    struct mcl_entry *mcl_e;

    /* 
     * Check if the remote end-point is a known mars client.
     * If it's the case, and if the mars client entry isn't linked
     * to any p2p VC, then link it to the present one (mcl_p2p)
     */

    /* try first with MCS Table */
    mcl_e = lookup_mcl_entry(&mcl_p2p->rsvcaddr, &MCL_Table[mcs_type -1]);

    if (mcl_e != NULL) {
	if (mcl_e->p2p == NULL) {
	    mcl_e->p2p = mcl_p2p;
	    mcl_p2p->mcl_e = mcl_e;
	}
    } else {
	mcl_e = lookup_mcl_entry(&mcl_p2p->rsvcaddr, &MCL_Table[clmb_type -1]);
	if (mcl_e != NULL) {
	    if (mcl_e->p2p == NULL) {
		mcl_e->p2p = mcl_p2p;
		mcl_p2p->mcl_e = mcl_e;
	    }
	}
    }

    /* strip off llc/snap encapsulation and check it */
    /* is it a MARS control message ? */
    alc = (struct atmllc *) buffer;
    if (bcmp(alc, &llc_mars_ctl, ATMLLC_LEN) != 0) {
	/* is it multicast data packet (e.g. NDP) ? */ 
	amlc = (struct atmmllc *) buffer;
	if (check_llc_multi_ipv6(amlc) != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "dispatch_mars_msg: bad multicast encapsulation");
	    /* discard the message */
	    InputPkts[0]++;
	    return(-1);
	} else {
	    error = process_multi_ipv6(buffer, buf_len);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "dispatch_mars_msg: failed to process multi IPv6 packet");
		/* discard the message */
		InputPkts[0]++;
		return(-1);
	    }
	    return(0);
	}
	diag(COMPONENT, DIAG_ERROR, 
	     "dispatch_mars_msg: bad mars encapsulation");
	/* discard the message */
	InputPkts[0]++;
	return(-1);
    }
    
    /* Now it is a mars control message */

    mars_buf = (u_int8_t *) (buffer + ATMLLC_LEN);
    mars_len = buf_len - ATMLLC_LEN;
    
    /* check fixed header is valid */

    if ((error = check_fxh(mars_buf, mars_len)) != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "dispatch_mars_msg: invalid mars message fixed header");
	InputPkts[0]++;
	return(-1);
    }


    fxh = (struct fixed_hdr *) mars_buf;
    opcode = (enum mars_opcode) ntohs(fxh->op);

    switch (opcode) {
    case mars_request :
	/* pass the whole buffer (perhaps mars_nak will be sent) */
	diag(COMPONENT, DIAG_DEBUG, 
	     "dispatch_mars_msg: incoming mars_request message");
	InputPkts[opcode]++;
	error = process_request(mcl_p2p, buffer, buf_len);
	break;
    case mars_join:
    case mars_leave:
    case mars_mserv:
    case mars_unserv:
	InputPkts[opcode]++;
	diag(COMPONENT, DIAG_DEBUG, 
	     "dispatch_mars_msg: incoming %s message", mop_name[opcode]);
	error = process_membership(mcl_p2p, mars_buf, mars_len, opcode);
	break;
    default:
	InputPkts[0]++;
	diag(COMPONENT, DIAG_ERROR, 
	     "ERROR: dispatch_mars_msg: invlalid MARS operation, %s",
	     mop_name[opcode]);
	error = -1;
    }
    return(error);
}
