/* marsd.c - MARS daemon */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <sys/wait.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "atm.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_def.h"
#include "mars_msg.h"
#include "tables.h"

#include "marsd_tab.h"
#include "marsd_io.h"

#define COMPONENT	"MAIN"

int debug = 0;
int myitf = 0;

const char *sig_qos = NULL;
const char *dump_dir = NULL;

/* Cluster/Server Control VC */
ATMCALL *Ccvc, *Scvc;
int ccvc_sd, scvc_sd;

/* stats for mars messages */
u_int8_t mData = 14;
u_int8_t AllNodes = 15;
u_int32_t InputPkts[16], OutputPkts[16];
u_int32_t NbInput, NbOutput, NbInputErrors, NbOutputErrors;

/* multicast group Host/Server map entries */
struct map_head Map_Table[2];		/* Map_Table[0|1] for Host|Server */  
/* Current Cluster Members (CLMB) and MCSs */
struct mcl_head MCL_Table[2];		/* MCL_Table[0|1] for CLMB|MCS */  

int CurrCMI, CheckedCMI, CurrMCSId, CheckedMCSId;
u_int32_t CurrCSN, CurrSSN;

void init_marsd(void)
{

    /* initialize tables */
    TAILQ_INIT(&MCL_Table[0]);
    TAILQ_INIT(&MCL_Table[1]);
    TAILQ_INIT(&Map_Table[0]);
    TAILQ_INIT(&Map_Table[1]);

    /* initialize global variables */
    bzero(InputPkts, 14*sizeof(u_int32_t));
    bzero(OutputPkts, 14*sizeof(u_int32_t));
    NbOutputErrors = NbInputErrors = NbOutput = NbInput = 0;

    CurrMCSId = CurrCMI = 0;
    CheckedMCSId = CheckedCMI = CMI_MAX;
    CurrCSN = 1;	/* free choice of initial value (cf 5.1.4.2) */
    CurrSSN = 1;
    
    /* initialize CCVC/SCVC parameters */
    Ccvc = NULL;
    Scvc = NULL;
    ccvc_sd = -1;
    scvc_sd = -1;
}


static int main_loop(void)
{
    int error = 0, fin = 0, fds;
    fd_set set, perm;
    ATMCALL *call;
    pid_t pid;
    int sig_down = 0;
    enum atmsvc_msg_type type;
    
    if (signal(SIGPIPE, sig_pipe) == SIG_ERR)
	diag(COMPONENT, DIAG_ERROR, "sig_pipe: error");
  
    FD_ZERO(&perm);
    FD_SET(sig_sd, &perm);
    fds = FD_SETSIZE;
    (void) gettimeofday(&now, NULL);

    if ((error = socketpair(AF_UNIX, SOCK_STREAM, 0, psd)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "main_loop: failed to create a socketpair");
	return(-1);
    }
    if ((pid = fork()) < 0) {
	diag(COMPONENT, DIAG_ERROR, "fork error: %s", strerror(errno));
	exit(1);
    }
    if (pid == 0) {		/* child */
	(void) close(sig_sd);	/* useless for child */
	if ((fin = user_input()) == 1) {
	    diag(COMPONENT, DIAG_DEBUG, "Child exited normally");
	    exit(0);
	}
	exit(1);
    }
    if (pid > 0) {
	while (fin <= 0) {
	    set = perm;
	    FD_SET(psd[0], &set);
	
	    for (call = AtmCalls; call; call =  call->next)
		if (call->sd > 0)
		    FD_SET(call->sd, &set);

	    error = select(fds, &set, NULL, NULL, next_timer());
	    if (error < 0) {
		if (errno != EINTR) {
		    perror("select");
		    /* abort(); */
		    
		    if (kill(pid, SIGTERM) < 0) /* kill the child  */
			perror("kill");
		    abort();
		}
		continue;
	    }
	    
	    gettimeofday(&now, NULL);

	    if (FD_ISSET(psd[0], &set))
		fin = process_cmd();

	    if (FD_ISSET(sig_sd, &set)) {
		if ((error = recv_atmsigd(&type)) < 0) {
		    if (error == -2) { 
			/* sigd has closed its tcp socket with client */
			sig_down = 1;
			diag(COMPONENT, DIAG_ERROR, 
			     "\n\nmain_loop: !!! CONNECTION WITH ATMSIGD "
			     "IS DOWN !!!");
			diag(COMPONENT, DIAG_ERROR, 
			     "TEARING DOWN CURRENT ATM CALLS");
			(void) close_all(sig_down);
			(void) close(sig_sd);
			if (kill(pid, SIGTERM) < 0) /* kill the child  */
			    perror("kill");
		    }
		    return(-2);
		}
	    }
	    for (call = AtmCalls; call; call =  call->next) {
		if ((call->sd > 0) && (FD_ISSET(call->sd, &set))) {
		    if ((error = recv_mclient(call)) != 0) {
			diag(COMPONENT, DIAG_ERROR, 
			     "main_loop: failed to read message from mclient");
		    }
		}
	    }

	    /* expire timers after handling messges to make sure we don't
               time out unnecessarily because of scheduling delays */
            expire_timers();
	}
    }
    return(0);
}


/* ------------------------------- main ...  ------------------------------- */

static void usage(const char *name)
{
    fprintf(stderr,
	    "usage: %s [ -s selector ] \n", name);
    exit(1);
}


int main(int argc, char **argv)
{
    int c, error;
    char *end;
    long long_sel;
    u_int8_t selector;

    set_application("MARSD");
    dump_dir = NULL;
    set_verbosity(NULL, DIAG_DEBUG);
    
/* initialize default selector value */
    selector = 0x00;
    
    if (argc > 1) {
	while ((c = getopt(argc, argv, "s:")) != EOF)
	    switch (c) {
	    case 's':		/* choose a selector for local ATM address */
		long_sel = strtol(optarg, &end, 0);
		if (*end)
		    usage(argv[0]);
		if (long_sel > 0xff) {
		    diag(COMPONENT, DIAG_ERROR,
			 "Bad selector value!  0x00 <= selector <= 0xff");
		    exit(-1);
		}
		selector = (u_int8_t) long_sel;
		break;
	    default:
		usage(argv[0]);
	    }
    }

    diag(COMPONENT, DIAG_INFO, "\nMARSD "
#ifdef UNI30
	 "UNI 3.0"
#endif
#ifdef UNI31
	 "UNI 3.1"
#ifdef ALLOW_UNI30
	 "+3.0compat"
#endif
#endif
#ifdef UNI40
	 "UNI 4.0"
#ifdef Q2963_1
	 "+Q.2931.1"
#endif
#endif
	 "/AAL5, version " VERSION "\n");
  
    init_current_time();
    init_addr(myitf, selector);
    CurrVcc = 0;
    
    if ((error = register_with_atmsigd()) != 0) {
	diag(COMPONENT, DIAG_ERROR, "main: failed to register with atmsigd");
	exit(error);
    }
    diag(COMPONENT, DIAG_DEBUG, "\tWaiting for incoming messages.....");
    /* initialize marsd internal tables and global variables */
    init_marsd();
    error = main_loop();
    diag(COMPONENT, DIAG_DEBUG, "deregistering with atmsigd");
    deregister_with_atmsigd();
    (void) close(sig_sd);
    diag(COMPONENT, DIAG_DEBUG, "End of session");
    return (error);
}
