/* marsd_tab.c - MARS Daemon's table management */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "../linux-atm.h"

#include "atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"

#include "mars_msg.h"
#include "mars_def.h"
#include "tables.h"

#include "marsd_io.h"
#include "marsd_tab.h"

#define COMPONENT "TAB"

/* 
 * This routine allocates a free Mars CLient Identifier (CurrMCLId) 
 * and returns the mars client entry which the new one is to be 
 * inserted before.
 */ 
struct mcl_entry *alloc_new_mcl_id(enum mcl_type mtype)
{
    int CurrMCLId, CheckedMCLId; 
    struct mcl_entry *mcl_e;
    struct mcl_head *mcl_h;
    
    /* 
     * Inspired by new Process ID allocation (cf /usr/src/sys/kern/kern_fork.c).
     * Find an unused MCLId.  We remember a range of unused MCLIds
     * ready to be used (from CurrMCLId+1 through CheckedMCLId-1).
     */
    
    /* 
     * First get a copy of global variables (Current and Checked IDs) 
     * The updated values will be wrtitten back to global variables
     */

    if (mtype == clmb_type) {
	CurrMCLId = CurrCMI;
	CheckedMCLId = CheckedCMI;

    } else {
	CurrMCLId = CurrMCSId;
	CheckedMCLId = CheckedMCSId;
    }

    mcl_h = &MCL_Table[mtype - 1];
    CurrMCLId++;
 retry:
        /*
         * If the CurrMCLId has wrapped around,
         * restart from 1
         */
    if (CurrMCLId > CMI_MAX) {
	CurrMCLId = 1;
	CheckedMCLId = 1;
    }
    if (CurrMCLId >= CheckedMCLId) {
	CheckedMCLId = CMI_MAX;
	/*
	 * Scan MARS Client entries to check whether this MCLId
	 * is in use.  Remember the lowest MCLId that's greater
	 * than CurrMCLId, so we can avoid checking for a while.
	 */

	for (mcl_e = TAILQ_FIRST(mcl_h); 
	     mcl_e != NULL; 
	     mcl_e = TAILQ_NEXT(mcl_e, links)) {
	    while (mcl_e->mcl_id == CurrMCLId) {
		CurrMCLId++;
		if (CurrMCLId >= CheckedMCLId)
		    goto retry;
	    }
	    if (mcl_e->mcl_id > CurrMCLId && CheckedMCLId > mcl_e->mcl_id)
		CheckedMCLId = mcl_e->mcl_id;
	}
    } else {
	/* find the mcl entry such that (mcl_id == CheckedMCLId) */
	for (mcl_e = TAILQ_FIRST(mcl_h); 
	     mcl_e != NULL; 
	     mcl_e = TAILQ_NEXT(mcl_e, links))
	    if (mcl_e->mcl_id == CheckedMCLId) 
		break;
    }

    /* write back (Current and Checked IDs) to corresponding global variables */
    if (mtype == clmb_type) {
	CurrCMI = CurrMCLId;
	CheckedCMI = CheckedMCLId;
    } else {
	CurrMCSId = CurrMCLId;
	CheckedMCSId = CheckedMCLId;
    }

    return(mcl_e);
}

/* find a mars client (CLMB/MCS) entry given its ATM address */
struct mcl_entry *lookup_mcl_entry(struct svc_addr *addr, 
				   struct mcl_head *mcl_h)
{
    struct mcl_entry *walk;

    for (walk = TAILQ_FIRST(mcl_h); walk; walk = TAILQ_NEXT(walk, links))
	if (atm_equal(&walk->svc_addr, addr, 0, 0) != 0)
	    break;
    return (walk);
}    

/* find a mars client (CLMB/MCS) entry given its MARS Client Id */
struct mcl_entry *lookup_mcl_by_id(u_int16_t mcl_id, 
				   struct mcl_head *mcl_h)
{
    struct mcl_entry *walk;

    for (walk = TAILQ_FIRST(mcl_h); walk; walk = TAILQ_NEXT(walk, links))
	if (walk->mcl_id == mcl_id)
	    break;
    return (walk);
} 

/* 
 * find a map entry matching a given multicast group, starting from 
 * a specified map_entry (start_e). 
 * Map entries are stored by group address in ascending order 
 */
struct map_entry *lookup_map_entry(struct map_head *mh,
				   struct in6_addr *ip6m,
				   struct map_entry *start_e,
				   int *found)
{
    struct map_entry *walk, *last = NULL;
    int cmp;

    *found = 0;
    if ((walk = start_e) == NULL)
	walk = TAILQ_FIRST(mh);
    while (walk != NULL) {
	cmp = in6_addr_cmp(&walk->ip6m, ip6m);
	if (cmp == 0) {
	    *found = 1;
	    return(walk);
	}
	if (cmp > 0)
	    return(last);	/* the map entry just before */

	last = walk;
	walk = TAILQ_NEXT(walk, links);
    }
    return (last);
}

struct mcl_entry *new_mcl_entry(enum mcl_type mtype,
				struct svc_addr *addr, 
				struct in6_addr *ia6,
				ATMCALL *mcl_p2p)
{
    struct mcl_entry *new;

    new = alloc_t(struct mcl_entry);
    bzero(new, sizeof(struct mcl_entry));
    new->mcl_type = mtype;
    TAILQ_INIT(&new->minmax_head);
    new->mcl_state = mcl_registering;
    bcopy(addr, &new->svc_addr, sizeof(struct svc_addr));
    bcopy(ia6, &new->ia6, sizeof(struct in6_addr));
    
    new->p2p = mcl_p2p;
    mcl_p2p->mcl_e = new;

    return(new);
}

struct map_entry *new_map_entry(struct in6_addr *ip6m, struct svc_addr *addr)

{
    struct map_entry *new;
    struct svc_addr_entry *addr_e;

    new = alloc_t(struct map_entry);
    bzero(new, sizeof(struct map_entry));
    LIST_INIT(&new->addr_head);
    bcopy(ip6m, &new->ip6m, sizeof(struct in6_addr));
    addr_e = alloc_t(struct svc_addr_entry);
    bzero(addr_e, sizeof(struct svc_addr_entry));
    bcopy(addr, &addr_e->addr, sizeof(struct svc_addr));
    LIST_INSERT_HEAD(&new->addr_head, addr_e, links);
    return(new);
}

/* 
 * check if the minmax pair is encompassed by one pair from the 
 * given minmax list.
 */

int is_pair_encompassed(struct minmax_pair *pair, struct minmax_head *mmh)
{
    struct minmax_entry *walk;

    for (walk = TAILQ_FIRST(mmh);
	 walk;
	 walk = TAILQ_NEXT(walk, links)) {
	if (in6_addr_cmp(&pair->min, &walk->pair.min) < 0)
	    return(0);
	if (in6_addr_cmp(&pair->min, &walk->pair.max) <= 0) {
	    if (in6_addr_cmp(&pair->max, &walk->pair.max) <= 0)
		return(1);
	    return(0);
	}
    }
    return(0);
}

/* 
 * check if the minmax pair has no intersection with the 
 * given minmax list.
 */

int is_pair_excluded(struct minmax_pair *pair, struct minmax_head *mmh)
{
    struct minmax_entry *walk;

    for (walk = TAILQ_FIRST(mmh);
	 walk;
	 walk = TAILQ_NEXT(walk, links)) {
	if (in6_addr_cmp(&pair->min, &walk->pair.min) < 0) {
	    if (in6_addr_cmp(&pair->max, &walk->pair.min) < 0)
		return(1);
	    else
		return(0);
	}
	if (in6_addr_cmp(&pair->min, &walk->pair.max) <= 0)
	    return(0);
    }
    return(1);
}

void update_host_map_entry(struct svc_addr *addr,
			   struct map_entry *map_e,
			   enum mars_opcode op,
			   int *removed)
{
    struct svc_addr_entry *addr_e;

    /* find the corresponding addr_entry if any */
    for (addr_e = LIST_FIRST(&map_e->addr_head);
	 addr_e;
	 addr_e = LIST_NEXT(addr_e, links)) {

	if (atm_equal(addr, &addr_e->addr, 0, 0) != 0)
	    break;
    }
    
    switch(op) {
    case mars_join:
	if (addr_e != NULL) {	/* this shouldn't happen */
	    diag(COMPONENT, DIAG_WARN, 
                 "update_map_entry: ATM addr is already member");
	    return;
	}  
	/* add the new addr as a member of the indicated group */
	addr_e = alloc_t(struct svc_addr_entry);
	bzero(addr_e, sizeof(struct svc_addr_entry));
	bcopy(addr, &addr_e->addr, sizeof(struct svc_addr));
	LIST_INSERT_HEAD(&map_e->addr_head, addr_e, links);
	return;
    case mars_leave:
	if (addr_e == NULL) {	/* this shouldn't happen */
	    diag(COMPONENT, DIAG_WARN, 
                 "update_map_entry: ATM address is no longer a memeber");
	    return;
	}
	/* remove the addr entry from the CLMB list of the indicated group */
	LIST_REMOVE(addr_e, links);
	free(addr_e);

	/* if the addr list in the map entry has become empty, release it */
	if (LIST_FIRST(&map_e->addr_head) == NULL) {
	    TAILQ_REMOVE(&Map_Table[clmb_type - 1], map_e, links);
	    free(map_e);
	    *removed = 1;
	}
	return;
    default:
	diag(COMPONENT, DIAG_WARN, 
                 "update_map_entry: unknown MARS operation code: %d", op);
    }
    return;
}

void remove_from_all_joined(struct mcl_entry *clmb_e)
{
    struct minmax_entry *minmax_e;
    struct minmax_head *mmh;
    struct map_entry *map_e, *start_e;
    struct in6_addr ip6m1, ip6m2;
    int cmp = -1, found = 0, removed, error, equal;
    char v[64];
    
    mmh = &clmb_e->minmax_head;
    start_e = NULL;
    for (minmax_e = TAILQ_FIRST(mmh); 
	 minmax_e; 
	 minmax_e = TAILQ_NEXT(minmax_e, links)) {
	bcopy(&minmax_e->pair.min, &ip6m1, sizeof(ip6m1));
	equal = (in6_addr_cmp(&minmax_e->pair.min, &minmax_e->pair.max) == 0);
	while (cmp <= 0) {
	    map_e = lookup_map_entry(&Map_Table[clmb_type - 1], 
				     &ip6m1, 
				     start_e, 
				     &found);
	    if (found) {
		removed = 0;
		start_e = TAILQ_NEXT(map_e, links);
		update_host_map_entry(&clmb_e->svc_addr, 
				      map_e, 
				      mars_leave, 
				      &removed);
		if (!removed)
		    start_e = map_e;
	    } else {
		diag(COMPONENT, DIAG_WARN, 
		     "remove_from_all_joined: WARN: haven't found host map "
		     "entry for address [%s]", 
		     inet_ntop(AF_INET6, &ip6m1, v, sizeof(v)));
		start_e = NULL;
	    }
	    if (equal)
		break;
	    else {
		error = increment_addr(&ip6m1, &ip6m2);
		if (error != 0) {
		    diag(COMPONENT, DIAG_WARN,
			 "remove_from_all_joined: increment failed");
		    return;
		}
		bcopy(&ip6m2, &ip6m1, sizeof(ip6m1));
		cmp = in6_addr_cmp(&ip6m1, &minmax_e->pair.max);
	    }
	}
    }
}

void remove_from_server_map(struct mcl_entry *mcs_e)
{
    struct minmax_entry *minmax_e;
    struct minmax_head *mmh;
    struct map_entry *map_e, *start_e;
    struct in6_addr ip6m1, ip6m2;
    int cmp = -1, found = 0, error, equal;
    char v[64];
    
    mmh = &mcs_e->minmax_head;
    start_e = NULL;
    for (minmax_e = TAILQ_FIRST(mmh); 
	 minmax_e; 
	 minmax_e = TAILQ_NEXT(minmax_e, links)) {
	bcopy(&minmax_e->pair.min, &ip6m1, sizeof(ip6m1));
	equal = (in6_addr_cmp(&minmax_e->pair.min, &minmax_e->pair.max) == 0);
	while (cmp <= 0) {
	    map_e = lookup_map_entry(&Map_Table[mcs_type - 1], 
				     &ip6m1, 
				     start_e, 
				     &found);
	    if (found) {
		start_e = TAILQ_NEXT(map_e, links);
		/* for the moment, only one MCS may support a multicast group */
		TAILQ_REMOVE(&Map_Table[mcs_type - 1], map_e, links);
		free(map_e);
	    } else {
		diag(COMPONENT, DIAG_WARN, 
		     "remove_from_server_map: WARN: haven't found server map "
		     "entry for address [%s]", 
		     inet_ntop(AF_INET6, &ip6m1, v, sizeof(v)));
		start_e = NULL;
	    }
	    if (equal)
		break;
	    else {
		error = increment_addr(&ip6m1, &ip6m2);
		if (error != 0) {
		    diag(COMPONENT, DIAG_WARN,
			 "remove_from_all_joined: increment failed");
		    return;
		}
		bcopy(&ip6m2, &ip6m1, sizeof(ip6m1));
		cmp = in6_addr_cmp(&ip6m1, &minmax_e->pair.max);
	    }
	}
    }
}


/* minmax pair entries are sorted in ascending order of [min, Max] */	
void update_minmax_list(struct minmax_head *mmh,
			 struct minmax_pair *pair,
			 struct minmax_entry *marked_e)
{
    struct minmax_entry *walk, *next, *new_minmax_e;
    struct in6_addr *min, *max, *new_min, *new_max;

    if (marked_e == NULL) {
	/* allocate a new minmax entry and insert at the tail */
	new_minmax_e = alloc_t(struct minmax_entry);
	bcopy(pair, &new_minmax_e->pair, sizeof(struct minmax_pair));
	TAILQ_INSERT_TAIL(mmh, new_minmax_e, links);
	return;
    }
    
    /* 
     * determine position of the new entry in relation to the 
     * given marked entry  
     */
    
    min = &marked_e->pair.min;
    new_min = &pair->min;
    max = &marked_e->pair.max;
    new_max = &pair->max;

    if (in6_addr_cmp(new_min, min) < 0) {
	if (in6_addr_cmp(new_max, min) < 0) {
	    /* allocate a new minmax entry and insert at the tail */
	    new_minmax_e = alloc_t(struct minmax_entry);
	    bcopy(pair, &new_minmax_e->pair, sizeof(struct minmax_pair));
	    TAILQ_INSERT_BEFORE(marked_e, new_minmax_e, links);
	    return;
	}
	
	/* 
	 * it's actually interleaving with 'marked_e'. 
	 * extend the 'min' boundary of 'marked_e'
	 */
	bcopy(new_min, min, sizeof(struct in6_addr));
    }
    
    /* 
     * here, its is also interleaving
     * now, determine how far the new 'max' goes
     */
    if (in6_addr_cmp(new_max, max) <= 0)	/* we're done */
	return;
    
    for (walk = TAILQ_NEXT(marked_e, links);  walk; walk = next) {
	next = TAILQ_NEXT(walk, links);
	if (in6_addr_cmp(new_max, &walk->pair.min) >= 0) {
	    bcopy(&walk->pair.max, max, sizeof(struct in6_addr));
	    TAILQ_REMOVE(mmh, walk, links);
	    free(walk);
	    continue;
	} else {
	    break;
	}
    }
    if (in6_addr_cmp(new_max, max) >= 0)
	bcopy(new_max, max, sizeof(struct in6_addr));
    return;
}

void free_addr_entries(struct svc_addr_head *svc_head)
{
    struct svc_addr_entry *walk, *next;
    
    for (walk = LIST_FIRST(svc_head); walk; walk = next) {
	next = LIST_NEXT(walk, links);
	LIST_REMOVE(walk, links);
	free(walk);
    }
}
		
void free_map_entry(struct map_entry *map_e)
{
    if (map_e != NULL) {
	free_addr_entries(&map_e->addr_head);
	free(map_e);
    }
}
	
void free_mcl_entry(struct mcl_entry *mcl_e)
{
    enum mcl_type mtype = mcl_e->mcl_type;

    diag(COMPONENT, DIAG_DEBUG, "Releasing %s with [%s = %d]",
	 mcl_name[mtype], 
	 (mtype == clmb_type) ? "CMI" : "MCSId",
	 mcl_e->mcl_id);
 
    if (mcl_e->party != NULL)
	diag(COMPONENT, DIAG_WARN,
	     "corresponding party still exists in %s for %s",
	     (mtype == clmb_type) ? "CCVC" : "SCVC", mcl_name[mtype]);
    if (mcl_e->to_send != NULL) {
	free(mcl_e->to_send);
	mcl_e->to_send = NULL;
    }
    free_minmax_entries(&mcl_e->minmax_head);
    free(mcl_e);
}

int dispose_of_mclient(enum mcl_type mtype, u_int16_t mcl_id)
{
    struct mcl_entry *mcl_e;
    struct mcl_head *mcl_h;
    ATMCALL *p2p_call;
    PARTY *party;
    int error;

    mcl_h = &MCL_Table[mtype - 1];
    if ((mcl_e = lookup_mcl_by_id(mcl_id, mcl_h)) == NULL) {
	diag(COMPONENT, DIAG_ERROR, 
	     "dispose_of_mclient: No matching %s for [MCLId = %d]",
	     mcl_name[mtype], mcl_id);
	return(-1);
    }

    diag(COMPONENT, DIAG_DEBUG, "Disposing of %s with [MCLId = %d]",
	 mcl_name[mtype], mcl_id);

    /* 
     * force closing private p2p call with mclient if connected and 
     * cut all cross references
     */
    if ((p2p_call = mcl_e->p2p) != NULL) {
	if (p2p_call->state == connected) {
	    error =  active_close_atmcall(0, p2p_call->vcc, p2p_call);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "dispose_of_mclient: failed to attempt closing private "
		     "p2p call with %s (MCLId = %d)", mcl_name[mtype], mcl_id);
		return(-1);
	    }
	}
	/* 
	 * anyway, cut cross references, signaling will carry on 
	 * the cleaning job
	 */
	p2p_call->mcl_e = NULL;
	mcl_e->p2p = NULL;
    }

    /* 
     * force dropping corresponding party from CCVC/SCVC if active and 
     * cut all cross references
     */
    if ((party = mcl_e->party) != NULL) {
	if (party->pstate == ps_active) {
	    error = active_drop_party(0, 0, party);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "dispose_of_mclient: failed to drop %s from %s: ",
		     mcl_name[mtype], (mtype == clmb_type) ? "CCVC" : "SCVC");
		return(error);
	    }
	}
	/* 
	 * anyway, cut cross references, signaling will carry on 
	 * the cleaning job
	 */
	party->mcl_e = NULL;
	mcl_e->party = NULL;
    }

    /* Now we can release mcl entry! */
    
    if (mcl_e->mcl_type == clmb_type)
	remove_from_all_joined(mcl_e);
    else
	remove_from_server_map(mcl_e);

    TAILQ_REMOVE(&MCL_Table[mtype - 1], mcl_e, links);
    free_mcl_entry(mcl_e);
    
    return(0);
}

/* force releasing mars client table (CLMB/MCS) */
static void release_mcl_table(enum mcl_type mtype)
{
    struct mcl_head *mclh;
    struct mcl_entry *walk, *next;
    
    mclh = &MCL_Table[mtype - 1];

    diag(COMPONENT, DIAG_DEBUG, "releasing %s Table", mcl_name[mtype]);
    for (walk = TAILQ_FIRST(mclh); walk; walk = next) {
	next = TAILQ_NEXT(walk, links);
	TAILQ_REMOVE(mclh, walk, links);
	free_mcl_entry(walk);
    }
    return;
}

/* force releasing Map Table (for CLMB/MCS) */
static void release_map_table(enum mcl_type mtype)
{
    struct map_head *map_h;
    struct map_entry *walk, *next;
    
    diag(COMPONENT, DIAG_DEBUG, "releasing %s Map Table", 
	 (mtype == clmb_type) ? "Host" : "Server");
    map_h = &Map_Table[mtype - 1];
    for (walk = TAILQ_FIRST(map_h); walk; walk = next) {
	next = TAILQ_NEXT(walk, links);
	TAILQ_REMOVE(map_h, walk, links);
	free_svc_addr_entries(&walk->addr_head);
	free(walk);
    }
    return;
}

/* 
 * this routine may do nothing, since, in principle everything 
 * has already been cleaned before calling this routine
 */
void release_all_tables(void)
{

    /* force releasing CLMB and MCS Tables */
    release_mcl_table(clmb_type);
    release_mcl_table(mcs_type);
    
    /* force releasing Map Tables (for CLMB and MCS) */
    release_map_table(clmb_type);
    release_map_table(mcs_type);
    
    /* force resetting CCVC and SCVC */
    Ccvc = NULL;
    Scvc = NULL;

    return;
}

/* ****************** Display Internal Structures ********************* */

void show_all_mclients(enum mcl_type mtype)
{
    struct mcl_entry *mcl_e;
    struct mcl_head *mh;
    struct minmax_entry *mme;
    int fl3, i;
    char v[64], v1[64];
    ATMCALL *private, *p2mp;
    u_int32_t msn;

    if (mtype == clmb_type) {
	p2mp = Ccvc;
	msn = CurrCSN;
    } else {
	p2mp = Scvc;
	msn = CurrSSN;
    }
    mh = &MCL_Table[mtype - 1];

    printf("\t\t************  %s Table  **************\n\n", mcl_name[mtype]);
    if (p2mp != NULL) {
	printf("%s Control VC:\t<vcc, vpi, vci> = <%ld, %d, %d>\t<%s = %d>\n\n",
	       (mtype == clmb_type) ? "Cluster" : "Server",
	       p2mp->vcc, p2mp->vpi, p2mp->vci,
	       (mtype == clmb_type) ? "CSN" : "SSN", msn);
	
	for (mcl_e = TAILQ_FIRST(mh); mcl_e; mcl_e = TAILQ_NEXT(mcl_e, links)) {
	    printf("\t-----------------------------------------------------\n");
	    printf("\tIPv6-ATM address:\t%s\n",
		   inet_ntop(AF_INET6, &mcl_e->ia6, v, sizeof(v)));
	    printf("\tATM address:\t\t");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02x", mcl_e->svc_addr.prv[i]);
	    printf("\n\n");
	    
	    printf("\t<%s state = %s>\t<%s = %d>\n\n", 
		   mcl_name[mtype], 
		   mcl_state_name[mcl_e->mcl_state], 
		   (mtype == clmb_type) ? "CMI" : "MCSId",
		   mcl_e->mcl_id);
	    
	    if ((private = mcl_e->p2p) != NULL)
		printf("\tP2P VC:\t\t\t<vcc, vpi, vci> = <%ld, %d, %d>\n", 
		       private->vcc, private->vpi, private->vci);
	    if (mcl_e->party != NULL)
		printf("\tParty:\t\t\t[ep_ref = %d]\n\n", mcl_e->party->ep_ref);
	    
	    printf("\n\t*********     %s Multicast Groups    ***********\n",
		   (mtype == clmb_type) ? "Joined" : "Supported");
	    printf("\tMin_Max pair\t\t\t\tLayer_3_Group flag\n\n");
	    
	    for (mme = TAILQ_FIRST(&mcl_e->minmax_head); 
		 mme; 
		 mme = TAILQ_NEXT(mme, links)) {
		fl3 = ((mme->fl_layer3grp &  FL_LAYER3GRP) != 0);
		printf("\t[%s, %s]\t\t%d\n", 
		       inet_ntop(AF_INET6, &mme->pair.min, v, sizeof(v)),
		       inet_ntop(AF_INET6, &mme->pair.max, v1, sizeof(v1)),
		       fl3);
	    }
	    printf("\n\t---------------------------------------------------\n");
	}
    }
    return;
}

void show_map_table(enum mcl_type mtype)
{
    struct map_entry *map_e;
    struct svc_addr_entry *addr_e;
    char v[64];
    int i;

    printf("\t\t*********   %s Map Table    ***********\n\n",
	   (mtype == clmb_type) ? "Host" : "Server");
    for (map_e = TAILQ_FIRST(&Map_Table[mtype - 1]);
	 map_e; 
	 map_e = TAILQ_NEXT(map_e, links)) {
	
	printf("%s\n",
	       inet_ntop(AF_INET6, &map_e->ip6m, v, sizeof(v)));
	for (addr_e = LIST_FIRST(&map_e->addr_head);
	     addr_e;
	     addr_e = LIST_NEXT(addr_e, links)) {
	    printf("\t");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02x", addr_e->addr.prv[i]);
	    printf("\n");
	}
	printf("\n");
    }
    
    return;
}

void show_stats(void)
{

    int i;
    
    printf("\t\t*********     MARS messages Stats     **************\n\n");
    printf("\t\tInput MARS messages\t\tOutput MARS messages\n\n");

    printf("--------------------------------------------------------------"
	   "-----\n");

    for (i = 1; i < 10; i++) {
	printf("%s\t\t%d\t\t\t\t%d\n", 
	       mop_name[i],  InputPkts[i], OutputPkts[i]);
	
	printf("--------------------------------------------------------------"
	       "-----\n");
    }
    printf("%s\t\t%d\t\t\t\t%d\n", 
	   mop_name[13],  InputPkts[13], OutputPkts[13]);
    printf("--------------------------------------------------------------"
	       "-----\n");
    printf("mData\t\t\t%d\t\t\t\t%d\n", InputPkts[14], OutputPkts[14]);
    printf("--------------------------------------------------------------"
	       "-----\n");
    printf("AllNodes\t\t%d\t\t\t\t%d\n", InputPkts[15], OutputPkts[15]);
    printf("--------------------------------------------------------------"
	       "-----\n");

    printf("%s\t\t%d\t\t\t\t%d\n", 
	   mop_name[0],  InputPkts[0], OutputPkts[0]);
	

    printf("\n================================================================"
	   "========\n");
    printf("Total number of input MARS messages:\t%d\n", NbInput);
    printf("Total number of output MARS messages:\t%d\n", NbOutput);
    printf("Total number of input errors:\t\t%d\n", NbInputErrors);
    printf("Total number of output errors:\t\t%d\n", NbOutputErrors);

    printf("\n---------------------------------------------------------------"
	   "-----\n");

    printf("MARS Sequence Numbers:\t<CSN = %d>\t<SSN = %d>\n", 
	   CurrCSN, CurrSSN);
    printf("Last assigned CMI: %d\n", CurrCMI);
    printf("Last assigned MCSId: %d\n\n", CurrMCSId);

    return;
}
