/* frommars.c - Processing of incoming MARS messages */

/* Written 1998-1999 by Mohsn.Souissi@inria.fr, INRIA */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h> 
#include <sys/time.h>
#include <sys/types.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>


#include <net/if.h>
#include <net/if_atm.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmsap.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_def.h"
#include "mars_msg.h"
#include "tables.h"

#include "mcl_tab.h"
#include "mcl_timeout.h"
#include "to_mars.h"
#include "mcl_io.h"


#define COMPONENT "FROMMARS"

static int process_multi(int sd, u_int8_t *mars_buf, int mars_len)
{
    struct mars_multi *msg;
    struct fixed_hdr *fxh;
    u_int8_t shtl, sstl, thtl, tstl, spln, tpln;
    u_int16_t tnum;
    int i, error = 0;
    u_int8_t *start, *sha, *ssa, *spa, *tpa = NULL, *tha, *tsa;
    struct in6_addr ip6m;
    u_int32_t msn; 
    int msn_diff;
    u_int16_t seq_diff, curr_seq;
    struct resolv_entry *res_e;
    struct ept_entry *new_ept;
    struct svc_addr eptaddr;
    int theo_var_len, actual_var_len;;

    msg = (struct mars_multi *) mars_buf;
    fxh = &msg->fxh;

    /*
     * Get parameters to compute the message variable length
     */
    shtl = fxh->shtl & TL_LEN; 
    sstl = fxh->sstl & TL_LEN;
    thtl = msg->thtl & TL_LEN; 
    tstl = msg->tstl & TL_LEN; 
    spln = msg->spln & TL_LEN;  
    tpln = msg->tpln & TL_LEN;
    tnum = ntohs(msg->tnum);

    /* @@@ disable sub-address field temporarily */
    if (tstl) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_multi: can't handle sub-address field yet");
	return(-1);
    }

    /* @@@ only IPv6 is accepted as a layer-3 protocol */
    if ((spln != IN6_ADDR_LEN) || (tpln != IN6_ADDR_LEN)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_multi: layer-3 protocol not supported yet");
	return(-1);
    }

    theo_var_len = shtl + sstl + spln + tpln + tnum*(thtl + tstl);    
    actual_var_len = mars_len - sizeof(struct  mars_multi);

    /* check it's the actual variable length we've read */
    if (actual_var_len != theo_var_len) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_multi: variable length part inconsistency");
	return(-1);
    }

    start = (u_int8_t *) (mars_buf + sizeof(struct mars_multi));
    sha = get_info(&start, shtl); 
    ssa = get_info(&start, sstl);
    spa = get_info(&start, spln); 
    tpa = get_info(&start, tpln);
    bcopy(tpa, &ip6m, tpln);
    if ((res_e = lookup_ip6m(&ip6m)) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "process_multi: no matching resolv_entry");
	return(-1);
    }

    /* "rs_resolv" and "rs_revalid" are the only expectable states */

    if ((res_e->rstate != rs_resolv) && (res_e->rstate != rs_revalid)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_multi: unexpected MARS MULTI message");
	return(-1);
    }
    
    /* get msn and sequence number */
    msn = ntohl(msg->msn);
    curr_seq = ntohs(msg->seqxy) & SEQ_Y;
    
    /* 
     * initialize gathering multi sequences and set the 
     * timer on the first received sequence (res_e->seq not modified yet)
     */
    /* save msn value for the first multi seq */
    if (res_e->seq == 0) {
	res_e->msn = msn;
	START_TIMER(T_LAST_SEQ, on_T_LAST_SEQ, res_e);
    }
    seq_diff = curr_seq - res_e->seq;
    res_e->seq = curr_seq;
    /* if sequence jump detected (ie lost sequence(s)), 
     * set discard flag, wait for the last multi sequence 
     * and discard all sequences
     */
    if (seq_diff > 1)
	res_e->discard = 1;  
    
/* 
 * if MSN changes during reception of multi sequences, revalidate all open
 * p2mp VCs, wait for the last multi sequence and discard all sequences
 */
    msn_diff = msn - res_e->msn;
    if ((msn_diff != 0) && (msn_diff != 1)) {
	reval_all = 1;
	res_e->discard = 1;
    }

    /* ignore duplicate sequences */
    if (seq_diff == 0) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_multi: received a duplicate multi msg, seq = %d", 
	     curr_seq);
	return(0);		/* discard message silently */
    }

    /* check if it is the last multi msg seq */
    if ((ntohs(msg->seqxy) & LAST_SEQ)) { 
	/* first, stop the timer */
	STOP_TIMER(res_e);
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_multi: receiving last sequence of MULTI MESSAGE");
	if (res_e->discard) {
	    discard_ept_list(res_e);
	    if (reval_all) {
		revalidate_all();
		reval_all = 0;
	    } else {
		error = retransmit_mars_req(res_e);
		if (error != 0) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_multi: failed to retransmit request");
		    return(-1);
		}
	    }
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_multi: mars multi sequences have been discarded");
	    return(-1);
	}
    }
    diag(COMPONENT, DIAG_DEBUG, 
	 "process_multi: receiving MULTI MESSAGE sequence number %d", 
	 curr_seq);
    
    for (i=1; i <= tnum; i++) {
	tha = get_info(&start, thtl);
	tsa = get_info(&start, tstl);
	bzero(&eptaddr, sizeof(eptaddr));
	bcopy(tha, eptaddr.prv, ATM_ESA_LEN);
	/* create and insert the new ATM address */
	new_ept = new_ept_entry(&eptaddr, res_e);
	if (res_e->rstate == rs_resolv)
	    LIST_INSERT_HEAD(&res_e->ept_head, new_ept, links);
	else 
	    LIST_INSERT_HEAD(&res_e->new_head, new_ept, links);	/* rs_revalid */
    }
/*
 * Update resolve entry upon receiving the last sequence
 */
    if ((ntohs(msg->seqxy) & LAST_SEQ)) {
	mcl.mcl_sn = msn;
	if (res_e->rstate == rs_resolv) {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_multi:  ATM resolve list  built successfully!");
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_multi:  building corresponding p2mp VC...");
	    error = build_mvc(res_e);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_multi: failed to build Multicast VC");
		LIST_REMOVE(res_e, links);
		free_resolv_entry(res_e);
	    }
	} else {
	    /* rs_revalid case */
	    error = update_revalid(res_e);
	    if (error == 0) {
		diag(COMPONENT, DIAG_DEBUG, 
		     "process_multi: revalidation update succeeded!");
		res_e->rstate = rs_valid;
		res_e->mvc->revalidate = 0;
	    } else 
		diag(COMPONENT, DIAG_ERROR, 
		     "process_multi: revalidation update failed, "
		     "nothing has changed!");
	}
    }

    return(error);
}

/* 
 * process mars membership messages for mars client (CLMB/MCS): 
 * {JOIN, LEAVE, MSERV, UNSERV, SJOIN, SLEAVE}
 */
static int process_membership(int sd, u_int8_t *mars_buf, int mars_len)
{
    struct mars_memb *msg;
    struct fixed_hdr *fxh;
    u_int8_t shtl, sstl, spln, tpln;
    u_int16_t flags, pnum, cmi = 0, copy, regist, punched, fl_layer3grp;
    u_int8_t seq;
    int i, error = 0;
    u_int8_t *start, *sha, *spa;
    u_int32_t msn;
    int msn_diff = 0;
    struct unacked_req *unacked = NULL;
    struct svc_addr source_atm;
    struct in6_addr *max;
    struct minmax_pair *first_minmax = NULL, *minmax;
    struct minmax_entry  *new_entry;
    struct minmax_head mpairs;
    int theo_var_len, actual_var_len;
    enum mars_opcode opcode;
    int originator = 0, remove_unacked = 0;
    char v[64], v1[64];
    
    msg = (struct mars_memb *) mars_buf;
    fxh = &msg->fxh;
    opcode = (enum mars_opcode) ntohs(fxh->op);

    /*
     * Get parameters to compute the message variable length
     */

    shtl = fxh->shtl & TL_LEN; 
    sstl = fxh->sstl & TL_LEN;
    spln = msg->spln & TL_LEN;  
    tpln = msg->tpln & TL_LEN;
    pnum = ntohs(msg->pnum);
    
    /* @@@ only IPv6 is accepted as a layer-3 protocol */
    if (spln != IN6_ADDR_LEN) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_memebrship: layer-3 protocol not supported yet");
	return(-1);
    }

    theo_var_len = shtl + sstl + spln + pnum*2*tpln;
    actual_var_len = mars_len - sizeof(struct  mars_memb);

    /* check it's the actual variable length we've read */
    if (actual_var_len != theo_var_len) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_membership: variable length part inconsistency");
	return(-1);
    }

    /* 
     * extract info. from mars_memb message 
     */
    start = (u_int8_t *) (mars_buf + sizeof(struct mars_memb));

    /* extract ATM source ATM address */
    sha = get_info(&start, shtl); 
    spa = get_info(&start, spln); 
    bzero(&source_atm, sizeof(source_atm));
    bcopy(sha, source_atm.prv, shtl);

    /* get first min_max pair (if any) */
    if (pnum != 0) {
	/* pointer to the first minmax_pair */
	first_minmax = (struct minmax_pair *) start;
    }
    
    /* 
     * Process flags
     */
    flags = ntohs(msg->flags);

    /* @@@ only IPv6 is accepted as a layer-3 protocol */
    if (!(flags & FL_REGISTER) && (tpln != IN6_ADDR_LEN)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_membership: layer-3 protocol not supported yet");
	return(-1);
    }

    /* make sure it's a valid 'copy' (cf 5.2.2) */
    copy = flags & FL_COPY;
    if (copy == 0) {
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_membership: client received %s with copy flag reset",
	     mop_name[opcode]);
	return(0);	/* discard message silently */
    }
    
    /*
     * Track MARS Sequence Number
     */
    msn = ntohl(msg->msn);
    if (mcl.mcl_sn != 0)
	msn_diff = msn - mcl.mcl_sn;
    mcl.mcl_sn = msn;
    
    /* if MSN jumps, revalidate all p2mp VCs before returning */
    if ((msn_diff != 0) && (msn_diff != 1))
	reval_all = 1;
    
    /*
     * Check if originator of the {JOIN, LEAVE, MSERV, UNSERV}  request
     */
    if (atm_equal(&mcl.svc_addr, &source_atm, 0, 0) != 0) {	
	/* we are the originator */
	originator = 1;
	/* first check for 'punched' flag */
	punched = flags & FL_PUNCHED;
	if (punched != 0) {		
	    /* just ignore it (cf 6.2.4), we'll receive a private copy  */
	    goto out;
	}
	
	/* 
	 * make sure this msg matches a previously sent 
	 * {JOIN, LEAVE, MSERV, UNSERV} msg 
	 */
	regist = flags & FL_REGISTER;
	seq = (u_int8_t) (flags & FL_SEQUENCE);
	if ((unacked = lookup_unacked((u_int16_t) opcode, regist, seq, pnum, 
				      &source_atm, first_minmax)) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "No matching previously sent %s message", 
		 mop_name[opcode]);
	    error = -1;
	    goto out;
	}
	/* 
	 * As soon as we make sure it's the unacked entry in question,
	 * we stop its timer 
	 */
	 
	STOP_TIMER(unacked);
	
    }

    /* 
     * Check if is a registration message 
     */
    if ((regist = flags & FL_REGISTER) != 0) {
	/* make sure we are originator of this message */
	if (!originator) {
	    diag(COMPONENT, DIAG_WARN, 
		 "process_membership: not originator of registration msg %s", 
		 mop_name[opcode]);
	    goto out;
	}

	/* make sure it's received on p2p2mars */
	if (sd != p2p_sd) {
	    /* Just ignore the message and drop it silently */
	    diag(COMPONENT, DIAG_WARN, 
		 "process_membership: registration msg %s not "
		 "received on p2p2mars", mop_name[opcode]);
	    
	    /* restart corresponding timer */
	    START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
	    
	    goto out;
	}

	if (pnum != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "No multicast group addresses may be included in a "
		 "registration message");
	
	switch(opcode) {
	case mars_leave:
	case mars_unserv:
	    if ((error = active_close_atmcall(0, p2p_sd, p2p2mars)) != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "process_membership: failed to initiate "
		     "closing p2p2mars VC");
		
		/* restart corresponding timer */
		START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
		goto out;
	    } else {
		diag(COMPONENT, DIAG_DEBUG, 
		     "process_membership: %s de-registered", 
		     mcl_name[mcl.mcl_type]);
		mcl.mcl_state = mcl_none;
	    }
	    break;

	case mars_join:		/* it must be a CLMB */
	    
	    if (mcl.mcl_type != clmb_type) {
		diag(COMPONENT, DIAG_ERROR, 
		     "unexpected mars operation code %s", mop_name[opcode]);
		/* restart corresponding timer */
		START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
		error = -1;
		goto out;
	    }
	    
	    cmi = ntohs(msg->cmi);
	    if (cmi == 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "(CMI) Must be non-zero");
		/* restart corresponding timer */
		START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
		error = -1;
		goto out;
	    }
	    
	    /* update mcl structure */
	    mcl.cmi = cmi;
	    mcl.mcl_state = mcl_registered;
	    
	    /* let interface learn this cmi value */
	    error = set_my_cmi(cmi, myitf);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "process_membership: failed to set interface CMI");
		error = -1;
	    }

	    diag(COMPONENT, DIAG_DEBUG, 
		 "CLMB registered with MARS with CMI = %d", cmi);
	    /* join the link-local `all nodes' group */
	    error = join_former_mgroups(myitf);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "process_membership: failed to initiate joining" 
		     "formerly joined multicast groups");
		error = -1;
	    }
	    break;
	    
	case mars_mserv:		/* it must be a MCS */
	    
	    if (mcl.mcl_type != mcs_type) {
		diag(COMPONENT, DIAG_ERROR, 
		     "unexpected mars operation code %s", mop_name[opcode]);
		/* restart corresponding timer */
		START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
		error = -1;
		goto out;
	    }
	    
	    /* update MCS state */
	    mcl.mcl_state = mcl_registered;
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_membership: MCS registered with MARS successfully");
	    
	    break;
	    
	default:	/* unlikely to happen since opcode already filtered */
	    diag(COMPONENT, DIAG_ERROR, 
		 "unexpected mars operation code %s", mop_name[opcode]);
	    /* restart corresponding timer */
	    START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
	    error = -1;
	    goto out;
	}
	    
	remove_unacked = 1;
	goto out;
    } else {	/* not a registration message */
	/* make sure there is one min_max pair at least */
	if (first_minmax == NULL) {
	    diag(COMPONENT, DIAG_WARN, 
		 "%s message contains no min_max pair", mop_name[opcode]);
	    goto out;
	}
    }

    /* 
     * Updating internal tables and and involved Multicast VCs 
     */
    max = NULL;
    fl_layer3grp = (mcl.mcl_type == clmb_type) ? flags & FL_LAYER3GRP : 0;

    /* get the list of minmax pairs */
    TAILQ_INIT(&mpairs);
    for (i = 0; i < pnum; i++) {
	minmax = get_info(&start, sizeof(struct minmax_pair)); 
	if (is_multicast_pair(&minmax->min, &minmax->max) == 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_membership: Not a multicast minmax pair");
	    free_minmax_entries(&mpairs);
	    if ((originator == 1) && (unacked != NULL))
		/* restart corresponding timer */
		START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
	    error = -1;
	    goto out;;
	}
	if ((max != NULL)
	    && (in6_addr_cmp(&minmax->min, max) < 0)) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_membership: multicast minmax pairs are not sorted!");
	    free_minmax_entries(&mpairs);
	    if ((originator == 1) && (unacked != NULL))
		/* restart corresponding timer */
		START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
	    error = -1;
	    goto out;
	}
	new_entry = alloc_t(struct minmax_entry);
	bcopy(minmax, &new_entry->pair, 
	      sizeof(struct minmax_pair));
	new_entry->fl_layer3grp = fl_layer3grp;
	TAILQ_INSERT_TAIL(&mpairs, new_entry, links);
	max = &minmax->max;
    }

    if (mcl.mcl_type == clmb_type) {
	
	switch (opcode) {    

	case mars_join:
	    /* 
	     * Now, MARS client can update its membership table if 
	     * originator of the mars_join message
	     */
	    if ((originator == 1) && (unacked != NULL))
		update_membership(unacked, fl_layer3grp);
	    
	    if (sd == cvc_sd) {
		/* 
		 * add this cluster member given by its ATM address as a 
		 * leaf for the multicast group(s) which the member has 
		 * joined and which we have an open p2m VC to
		 */
		add_new_leaves(&source_atm, &mpairs);
		break;
	    } else {	/* sd == p2p_sd */
		/* 
		 * this is a confirmation of a previous JOIN request for 
		 * a group which is currently supported by a MCS
		 */
		break;
	    }

	case mars_leave:
	    if ((sd == cvc_sd) || (mcl.mcl_state == mcl_deregistering)) {
		/* 
		 * drop this cluster member as a leaf from the multicast 
		 * group(s) which the member has left and which we have an 
		 * open p2m VC to
		 */
		drop_leaves(&source_atm, &mpairs);
	    } 

	    /* else (sd == p2p_sd) 
	     * this is a confirmation of a previous LEAVE request for 
	     * a group which is currently supported by a MCS
	     */

	    /* 
	     * Now, CLMB can update its Minmax Table if originator of 
	     * the mars_leave message
	     */
	    if ((originator == 1) && (unacked != NULL)) {
		update_membership(unacked, fl_layer3grp);

		/* 
		 * if we've left all join_groups and cluster member is 
		 * in `deregistering' state,  send de-register request to MARS
		 */
		if ((TAILQ_FIRST(&Minmax_Table) == NULL) 
		    && (mcl.mcl_state == mcl_deregistering)) {
		    error = deregister_mars_client(mars_leave);
		    if (error != 0)
			diag(COMPONENT, DIAG_ERROR,
			     "process_membership: failed to send "
			     "de-registration request");
		}
	    }
	    break;

	default:	/* unlikely to happen since opcode already filtered */
	    diag(COMPONENT, DIAG_ERROR, 
		 "unexpected mars operation code %s", mop_name[opcode]);
	    /* restart corresponding timer */
	    START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
	    error = -1;
	}
	goto out;
    }

    /* now it's a MCS  */
    switch (opcode) {
    case mars_mserv:
	/* make sure it's received on SCVC */
	if (sd != cvc_sd) {
	    /* 
	     * just ignore the message, drop it silently and 
	     * release corresponding unacked entry 
	     */
	    diag(COMPONENT, DIAG_WARN, 
		 "process_membership: membership msg %s not "
		 "received on SCVC. Ignored.", mop_name[opcode]);
	    remove_unacked = 1;
	    goto out;
	}
	
	if ((originator == 1) && (unacked != NULL)) {
	    /* 
	     * Now, MCS can update its Minmax Table and initiate resolving 
	     * the multicast group(s) for which it has just received "MSERV" 
	     * confirmation
	     */
	    
	    update_membership(unacked, 0);

	    diag(COMPONENT, DIAG_DEBUG,
		 "process_membership: resolving mgroups bloc [%s, %s]",
		 inet_ntop(AF_INET6, &first_minmax->min, v, sizeof(v)),
		 inet_ntop(AF_INET6, &first_minmax->max, v1, sizeof(v1)));
	    error = resolve_minmax_mgroups(first_minmax);
	    if (error != 0)
		diag(COMPONENT, DIAG_ERROR, 
		     "process_membership: failed to initiate resolving mgroups "
		     "bloc [%s, %s]", 
		     inet_ntop(AF_INET6, &first_minmax->min, v, sizeof(v)),
		     inet_ntop(AF_INET6, &first_minmax->max, v1, sizeof(v1)));
	}
	break;

    case mars_sjoin:
	/* 
	 * add this cluster member given by its ATM address as a leaf for the 
	 * multicast group(s) which the member has joined and which we 
	 * have an open p2m VC to
	 */

	add_new_leaves(&source_atm, &mpairs);
	break;
	
    case mars_sleave:
	/* 
	 * drop this cluster member as a leaf from the multicast group(s) 
	 * which the member has left and which we have an open 
	 * p2m VC to
	 */
	drop_leaves(&source_atm, &mpairs);
	break;
	
    case mars_unserv:
	/* make sure it's received on SCVC if not originator */
	if ((!originator) && (sd != cvc_sd)) {
	    /* 
	     * just ignore the message, drop it silently and 
	     * release corresponding unacked entry 
	     */
	    diag(COMPONENT, DIAG_WARN, 
		 "process_membership: membership msg %s not "
		 "received on SCVC. Ignored.", mop_name[opcode]);
	    remove_unacked = 1;
	    goto out;
	}
	
	/* 
	 * Now, MARS client can update its membership table and remove 
	 * resolve entry(ies) if originator of the mars_unserv message
	 */
	if ((originator == 1) && (unacked != NULL)) {
	    update_membership(unacked, 0);
	    diag(COMPONENT, DIAG_DEBUG,
		 "process_membership: closing MVCs bloc [%s, %s]",
		 inet_ntop(AF_INET6, &first_minmax->min, v, sizeof(v)),
		 inet_ntop(AF_INET6, &first_minmax->max, v1, sizeof(v1)));
	    error = close_minmax_mvc(first_minmax);
	    if (error != 0)
		diag(COMPONENT, DIAG_ERROR,
		     "process_membership: failed to close MVCs bloc [%s, %s]",
		     inet_ntop(AF_INET6, &first_minmax->min, v, sizeof(v)),
		     inet_ntop(AF_INET6, &first_minmax->max, v1, sizeof(v1)));
	    
	}
	
	/* 
	 * if we've `unserved' all supported mgroups and MCS is 
	 * in `deregistering' state,  send de-register request to MARS
	 */
	if ((TAILQ_FIRST(&Minmax_Table) == NULL) 
	    && (mcl.mcl_state == mcl_deregistering)) {
	    error = deregister_mars_client(mars_unserv);
	    if (error != 0)
		diag(COMPONENT, DIAG_ERROR,
		     "process_membership: failed to send de-registration "
		     "request");
	}
	break;
	
    default:	/* unlikely to happen since opcode already filtered */
	diag(COMPONENT, DIAG_ERROR, 
	     "process_membership: unexpected mars operation code %s", 
	     mop_name[opcode]);
	/* restart corresponding timer */
	START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
	error = -1;
	goto out;
    }
    
    
 out:
    /* if reval_all flag is set, revalidate all open p2m VCs */
    if (reval_all) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_membership:  MSN jump. Revalidating all open p2m VCs");
	revalidate_all();
	reval_all = 0;
    }
    
    if ((unacked != NULL) && (remove_unacked)) {
	LIST_REMOVE(unacked, links);
	if (unacked->msg != NULL)
	    free(unacked->msg);
	free(unacked);
    }
    return(error);
}

/* 
 * Upon processing this message which means that the group 
 * is currently empty, delay sending out a new MARS_REQUEST
 * for that group 	(cf 5.1.1)
 */
static int process_nak(int sd, u_int8_t *mars_buf, int mars_len)
{
    struct mars_req *msg;
    struct fixed_hdr *fxh;
    u_int8_t shtl, sstl, thtl, tstl, spln, tpln;
    u_int8_t *start, *sha, *ssa, *spa, *tpa = NULL;
    struct in6_addr ip6m;
    struct resolv_entry *entry;
    char v[64];
    int theo_var_len, actual_var_len;
    int error;

    msg = (struct mars_req *) mars_buf;
    fxh = &msg->fxh;

    /*
     * Get parameters to compute the message variable length
     */
    shtl = fxh->shtl & TL_LEN; 
    sstl = fxh->sstl & TL_LEN;
    thtl = msg->thtl & TL_LEN; 
    tstl = msg->tstl & TL_LEN; 
    spln = msg->spln & TL_LEN;  
    tpln = msg->tpln & TL_LEN;

    /* @@@ disable sub-address field temporarily */
    if (tstl) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_nak: can't handle sub-address field yet");
	return(-1);
    }

    /* @@@ only IPv6 is accepted as a layer-3 protocol */
    if ((spln != IN6_ADDR_LEN) || (tpln != IN6_ADDR_LEN)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_nak: layer-3 protocol not supported yet");
	return(-1);
    }

    theo_var_len =  shtl + sstl + spln + tpln;
    actual_var_len = mars_len - sizeof(struct  mars_req);

    /* check it's the actual variable length we've read */
    if (actual_var_len != theo_var_len) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_nak: variable length part inconsistency");
	return(-1);
    }

    /* 
     * extract info. from mars_nak message 
     */
    start = (u_int8_t *) (mars_buf + sizeof(struct mars_req));
    sha = get_info(&start, shtl); 
    ssa = get_info(&start, sstl);
    spa = get_info(&start, spln); 
    tpa = get_info(&start, tpln);
    bcopy(tpa, &ip6m, sizeof(struct in6_addr));
    if ((entry = lookup_ip6m(&ip6m)) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "process_nak: no matching resolv_entry");
	return(-1);
    }
    diag(COMPONENT, DIAG_DEBUG, "Multicast group [%s] is empty",
	 inet_ntop(AF_INET6, &entry->ia6, v, sizeof(v)));
    
    /* delete route to that destination */
    (void) setuid(0);
    error = sendrtmsg(RTM_DELETE, 1, &entry->ia6, 0, 0, 0);
    if (error != 0) 
	diag(COMPONENT, DIAG_ERROR, 
	     "delete_ipatmroute: failed to delete route to "
	     "destination %s: %s", 
	     inet_ntop(AF_INET6, &entry->ia6, v, sizeof(v)), strerror(error));
    else 
	diag(COMPONENT, DIAG_DEBUG, 
	     "Deleted route to destination %s", 
	     inet_ntop(AF_INET6, &entry->ia6, v, sizeof(v)));

    /* mark this resolve entry in "holddown" state */
    entry->rstate = rs_holddown;
   /* delay sending out a new MARS_REQUEST for that group : (cf 5.1.1) */
    START_TIMER(T_HOLD_REQ, on_T_HOLD_REQ, entry);
    
    return(0);
}

/*
 * REDIRECT_MAP messages MAY be generated by the MARS if synchronization
 * with backup MARS's is enabled. As this is not the case, disable this 
 * functionality for the moment 
 */
static int process_redirect_map(int sd, u_int8_t *mars_buf, int mars_len)
{
    diag(COMPONENT, DIAG_ERROR, 
	     "process_redirect_map: Operation not supported yet");
    return(0);
}

/* this routine is specific to CLMB */
static int process_migr(int sd, u_int8_t *mars_buf, int mars_len)
{
    struct mars_migr *msg;
    struct fixed_hdr *fxh;
    u_int8_t shtl, sstl, thtl, tstl, spln, tpln;
    u_int16_t tnum;
    int i, error = 0;
    u_int8_t *start, *sha, *ssa, *spa, *tpa = NULL, *tha, *tsa;
    struct in6_addr ip6m;
    u_int32_t msn;
    int msn_diff;
    struct resolv_entry *res_e;
    struct ept_entry *new_ept;
    ATMCALL *mvc;
    struct svc_addr eptaddr;
    int theo_var_len, actual_var_len;
    char v[64];

    msg = (struct mars_migr *) mars_buf;
    fxh = &msg->fxh;

    /*
     * Get parameters to compute the message variable length
     */
    shtl = fxh->shtl & TL_LEN; 
    sstl = fxh->sstl & TL_LEN;
    thtl = msg->thtl & TL_LEN; 
    tstl = msg->tstl & TL_LEN; 
    spln = msg->spln & TL_LEN;  
    tpln = msg->tpln & TL_LEN;
    tnum = ntohs(msg->tnum);

    /* @@@ disable sub-address field temporarily */
    if (tstl) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_migr: can't handle sub-address field yet");
	return(-1);
    }

    /* @@@ only IPv6 is accepted as a layer-3 protocol */
    if ((spln != IN6_ADDR_LEN) || (tpln != IN6_ADDR_LEN)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_migr: layer-3 protocol not supported yet");
	return(-1);
    }

    theo_var_len = shtl + sstl + spln + tpln + tnum*(thtl + tstl);    
    actual_var_len = mars_len - sizeof(struct  mars_migr);

    /* check it's the actual variable length we've read */
    if (actual_var_len != theo_var_len) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_migr: variable length part inconsistency");
	return(-1);
    }

    /*
     * Track MARS Sequence Number
     */
    msn = ntohl(msg->msn);
    msn_diff = msn - mcl.mcl_sn;
    mcl.mcl_sn = msn;
    
    /* if MSN jumps, revalidate all open p2mp VCs before returning */
    if ((msn_diff != 0) && (msn_diff != 1))
	reval_all = 1;
    
    /* 
     * extract info. from mars_migr message 
     */
    start = (u_int8_t *) (mars_buf + sizeof(struct mars_migr));
    sha = get_info(&start, shtl); 
    ssa = get_info(&start, sstl);
    spa = get_info(&start, spln); 
    tpa = get_info(&start, tpln);
    bcopy(tpa, &ip6m, tpln);
    
    /* do we have any open p2m VC open to that group */
    if ((res_e = lookup_ip6m(&ip6m)) == NULL) {
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_migr: no open p2mp VC to group [%s] Message ignored.",
	     inet_ntop(AF_INET6, &ip6m, v, sizeof(v)));
	goto out;
    }

    /* "rs_valid" is the only expectable state */

    if (res_e->rstate != rs_valid) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_migr: MARS MIGRATE message received. Resolve entry "
	     "still in %s", resolv_state_name[res_e->rstate]);
	goto out;
    }
    
    diag(COMPONENT, DIAG_DEBUG, "process_migr: receiving MIGRATE MESSAGE");
    
    for (i=1; i <= tnum; i++) {
	tha = get_info(&start, thtl);
	tsa = get_info(&start, tstl);
	bzero(&eptaddr, sizeof(eptaddr));
	bcopy(tha, eptaddr.prv, ATM_ESA_LEN);
	/* create and insert the new ATM address */
	new_ept = new_ept_entry(&eptaddr, res_e);
	/* insert it as if we were revaliditaing the group */
	LIST_INSERT_HEAD(&res_e->new_head, new_ept, links);
    }
    /*
     * Update resolve entry
     */
    mcl.mcl_sn = msn;
    error = migrate_mvc(res_e);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_migr: failed to initiate migrataing "
	     "Multicast VC");
	LIST_REMOVE(res_e, links);
	if ((mvc = res_e->mvc) != NULL) {
	    res_e->mvc->res_e = NULL;
	    res_e->mvc = NULL;
	    if (mvc->state == connected)
		(void) active_close_atmcall(0, mvc->vcc, mvc);
	    else
		free_atmcall(mvc);
	}
	free_resolv_entry(res_e);
	error = -1;
	goto out;
    }

 out:
    /* if reval_all flag is set, revalidate all open p2m VCs */
    if (reval_all) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_migrate:  MSN jump. Revalidating all open p2m VCs");
	revalidate_all();
	reval_all = 0;
    }
    
    return(error);
}

/* 
 * upon receiving a MARS MIGRATE message, shut down the current p2mp VC to that
 * group and buid the new one based on MIGRATE message
 */
/*
 * This routine will be called 2 times : 
 * 1) upon receiving a MARS MIGRATE message, shut down the current p2mp VC 
 * corresponding to the mgroup in question;
 * 2) buid the new one based on MIGRATE message
 */
int migrate_mvc(struct resolv_entry *res_e)
{
    struct ept_entry *walk, *next, *start;
    struct ept_head *head;
    ATMCALL *mvc;
    u_long rnd_delay;
    int error, retry = 1;
    
    mvc = res_e->mvc;

    /* first call to this routine */
    if (mvc != NULL) {
	res_e->rstate = rs_migr;
    	/* shut down the current p2mp VC */
	error = active_close_atmcall(0, mvc->vcc, mvc);
	if (error != 0) {
	    diag(COMPONENT, DIAG_ERROR,
		 "migrate_mvc: failed to close multicast VC [vcc = %d]", 
		 mvc->vcc);
	    return(-1);
	}
	/* return here and come back later when the old MVC id torn down */
	return(0);
    }
 
    /* force freeing current end point entries */
    free_ept_entries(&res_e->ept_head);
    LIST_INIT(&res_e->ept_head);

    /* 
     * copy the new list of end point entries (it is likely that this new
     * list contain only 1 member)
     */
    for (walk = LIST_FIRST(&res_e->new_head); walk; walk = next) {
	next = LIST_NEXT(walk, links);
	LIST_REMOVE(walk, links);
	LIST_INSERT_HEAD(&res_e->ept_head, walk, links);
    }
    
    /* reset new_head */
    LIST_INIT(&res_e->new_head);

    /* build the new one (towards MCS) */
       
    /* 
     * find the first non tagged address entry and attempt to initiate 
     * the p2mp call with it (cf 5.1.3) 
     */
    head = &res_e->ept_head;
    start = LIST_FIRST(head);
    while (retry) {
	for (walk = start; walk; walk = LIST_NEXT(walk, links))
	    if (walk->tag_flag == 0)
		break;
	if (walk != NULL) {
	    error = atmcall_setup(&walk->addr, NULL, &res_e->ia6, 
				  NULL, 0, walk, NULL);
	    if (error != 0) {
		/* next time we'll start retrying from here */
		start = LIST_NEXT(walk, links);
		diag(COMPONENT, DIAG_ERROR, 
		     "migrate_mvc: failed to initiate p2mp VC");
		/* drop this endpoint from the list */
		LIST_REMOVE(walk, links);
		free(walk);
		continue;
	    } else {
		diag(COMPONENT, DIAG_DEBUG, 
		     "migrate_mvc: waiting for response from atmsigd");
		return(0);
	    }
	} else {
	    diag(COMPONENT, DIAG_ERROR, 
		 "migrate_mvc: failed, no more endpoint to be tried: "
		 "all addresses are tagged");
	    /* start random timer on all tagged ept entries to be retried */
	    for (walk = LIST_FIRST(head); walk; walk = LIST_NEXT(walk, links)) {
		walk->new_delay = get_random_delay(5*T_SEC, 10*T_SEC);
		rnd_delay = walk->new_delay;
		start_timer(rnd_delay, on_RANDOM_FIRST_EPT, walk);
	    }
	    return(-1);
	}
    }
    return(0);
}

int dispatch_mars_msg(int sd, u_int8_t *mars_buf, int mars_len)
{
    struct fixed_hdr *fxh;
    enum mars_opcode opcode;
    int error;
    

    /* check fixed header is valid */

    if ((error = check_fxh(mars_buf, mars_len)) != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "dispatch_mars_msg: invalid mars message fixed header");
	return(-1);
    }


    fxh = (struct fixed_hdr *) mars_buf;
    opcode = (enum mars_opcode) ntohs(fxh->op);

    diag(COMPONENT, DIAG_DEBUG, "receiving %s message on %s", 
	 mop_name[opcode], 
	 sd == p2p_sd ? "p2p VC with MARS" : 
	 (mcl.mcl_type == clmb_type) ? "CCVC" : "SCVC");

    switch (opcode) {
    case mars_multi :
	error = process_multi(sd, mars_buf, mars_len);
	break;

    case mars_join:
    case mars_leave:
	if (mcl.mcl_type != clmb_type) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "dispatch_mars_msg: unexpected mars operation code %s", 
		 mop_name[opcode]);
	    return(-1);
	}
	error = process_membership(sd, mars_buf, mars_len);
	break;

    case mars_mserv:
    case mars_unserv:
    case mars_sjoin:
    case mars_sleave:
	if (mcl.mcl_type != mcs_type) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "dispatch_mars_msg: unexpected mars operation code %s", 
		 mop_name[opcode]);
	    return(-1);
	}
	error = process_membership(sd, mars_buf, mars_len);
	break;

    case mars_nak:
	error = process_nak(sd, mars_buf, mars_len);
	break;
    case mars_redirect_map:
	error = process_redirect_map(sd, mars_buf, mars_len);
	break;
    case mars_migrate:
	if (mcl.mcl_type != clmb_type) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "dispatch_mars_msg: unexpected mars operation code %s", 
		 mop_name[opcode]);
	    return(-1);
	}
	error = process_migr(sd, mars_buf, mars_len);
	break;

    default:
	diag(COMPONENT, DIAG_ERROR, 
	     "dispatch_mars_msg: invlalid MARS operation, %s",
	     mop_name[opcode]);
	error = -3;
    }

    return(error);
}
