/* mcl_io.c - common MARS client (CLMB/MCS) I/O operations */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <net/if.h>
#ifndef if_type
#include <net/if_var.h>
#else
#define ifma_msghdr	ifa_msghdr
#define ifmam_msglen	ifam_msglen
#define ifmam_addrs	ifam_addrs
#define ifmam_index	ifam_index
#endif
#include <net/route.h>

#include <net/if_atm.h>
#include <netinet/in.h>
#include <netinet/in6_var.h>
#include <netinet/ip6.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/queue.h>

#include "../linux-atm.h"

#include "atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_def.h"
#include "multi_data.h"
#include "mars_msg.h"
#include "tables.h"

#include "mcl_tab.h"
#include "mcl_io.h"

#define COMPONENT "MCL_IO"

/* VCs for communication with marsd */
ATMCALL *p2p2mars, *Cvc;	/* Cvc = Control VC (CCVC/SCVC) */
int p2p_sd, cvc_sd;		/* associated socket descriptors */

char reval_all;			/* flag, revalidating all open p2mp VCs */

struct mars_client mcl;
/* multicast group resolve entries */
struct resolv_head Resolv_Table;
/* current resolve entry, result of the last lookup in Resolve Table */ 
struct resolv_entry *Curr_Resolved;

/* multicast groups joined/mserved by mars client */
struct minmax_head Minmax_Table;

/* Unacknowledged JOIN/LEAVE/MSERV/UNSERV requests */
struct unacked_head Unacked;



static int get_mars_addresses(void)
{
    FILE *file;
    char mars_ip6_str[INET6_ADDRSTRLEN + 1];
    char mars_name_txt[MAX_ATM_NAME_LEN + 1];
    int len;

    if ((file = fopen("marsd.addr","r")) == NULL) {
	perror("fopen");
	exit(-1);
    }

    bzero(mars_ip6_str, INET6_ADDRSTRLEN);
    if (!fgets(mars_ip6_str, INET6_ADDRSTRLEN, file)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "get_mars_addresse: failed to read MARS IPv6 address");
	fclose(file);
	exit(-1);
    }
    len = strlen(mars_ip6_str);
    mars_ip6_str[len -1] = '\0';

    if (inet_pton(AF_INET6, mars_ip6_str, &mcl.mars_ia6) != 1) {
	diag(COMPONENT, DIAG_ERROR, 
	     "get_mars_addresses: something is wrong with MARS IPv6 address");
	fclose(file);
	exit(-1);
    }
    
    if (!fgets(mars_name_txt, MAX_ATM_NAME_LEN, file)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "get_mars_addresse: failed to read MARS ATM address");
	fclose(file);
	exit(-1);
    }

    len = strlen(mars_name_txt);
    mars_name_txt[len - 1] = '\0';
    if (text2atm_svc(mars_name_txt, 
		     &mcl.mars_svc_addr, 
		     sizeof(struct svc_addr),
		     T2A_SVC | T2A_LOCAL)  < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "get_mars_addresse: failed to convert ATM address");
	fclose(file);
	exit(-1);
    }
    
    fclose(file);
    return(0);
}

/* capture a multicast address and parse it */
int get_mgroup_addr(struct in6_addr *ip6m)
{
    char ip6m_str[INET6_ADDRSTRLEN + 1];
    int ret, len;
    char v[64];


 retry_multi_addr:
    printf("\n\n\n\n");
    printf("\n Multicast Group Address: ");
    if (fgets(ip6m_str, sizeof(ip6m_str), stdin) == NULL)
	goto retry_multi_addr;
    if ((len = strlen(ip6m_str)) < 2)
	goto retry_multi_addr;

    ip6m_str[len - 1] = '\0'; /* replacing '\n' */
    ret = inet_pton(AF_INET6, ip6m_str, ip6m);
    if (ret < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "get_mgroup_addr: inet_pton: %s", strerror(errno));
	return(-1);
    }
    if (ret == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "get_mgroup_addr: unparsable IPv6 address %s", ip6m_str);
	return(-1);
    }
    if (IN6_IS_ADDR_MULTICAST(ip6m) == 0) {
        diag(COMPONENT, DIAG_ERROR,  
         "get_mgroup_addr: not a multicast address [%s]", 
             inet_ntop(AF_INET6, ip6m, v, sizeof(v)));
        return(-1);
    }
    return(0);
}


/* At these statements, mars client is supposed to be already connected 
 * to atmsigd 
 */
void init_mars_client(enum mcl_type mcl_type)
{
    int error;

    /* initialize some global parameters */

    p2p2mars = NULL;
    p2p_sd = -1;
    Cvc = NULL;
    cvc_sd = -1;

    /* initialize mcl (mars client) structure */
    bzero(&mcl, sizeof(mcl));
    mcl.mcl_type = mcl_type;
    mcl.mcl_state = mcl_none;
    bcopy(&my_local_atm_addr, &mcl.svc_addr, sizeof(struct svc_addr));
    bcopy(&my_local_ipv6_addr, &mcl.ia6, sizeof(struct in6_addr));

    if ((error = get_mars_addresses()) != 0) {
	diag(COMPONENT, DIAG_ERROR, "init_mars_client: failed to locate MARS");
	exit;
    }

    /* 
     * initialize mars client's tables 
     */

    LIST_INIT(&Resolv_Table);
    LIST_INIT(&Unacked);
    TAILQ_INIT(&Minmax_Table);

    reval_all = 0;
    return;
}


/* 
 * This routine causes a newly registered CLMB to get all multicast addresses
 * of mgroups the interface has already joined
 */

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))

#define SIN6(x)	((struct sockaddr_in6 *)x)

/*
 * get the list of multicast groups, the interface has alreday joined 
 * before registering with MARS
 */
int get_former_mgroups(int itf, struct in6_addr_head **in6hp) 
{
    int mib[] = { CTL_NET, PF_ROUTE, 0, AF_INET6, NET_RT_IFMADDR, 0 };
    u_int if_index;
    struct ifma_msghdr *ifmam;
    struct sockaddr *sa, *ifa;
    char *buf, *cp, *cp1;
    char str_name[IFNAMSIZ + 1];
    struct in6_addr_head *in6h;
    struct in6_addr_entry *in6e;
    int size, len, i;

    bzero(str_name, IFNAMSIZ + 1);
    (void) snprintf(str_name, IFNAMSIZ, ENNAME, itf);
    if_index = if_nametoindex(str_name);
    mib[5] = if_index;
	
    if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0) {
	perror("sysctl0");
	return(-1);
    }
    
    if ((buf = malloc(size)) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "get_former_mgroups: malloc failed");
	return(-1);
    }

    if (sysctl(mib, 6, buf, &size, NULL, 0) < 0) {
	perror("sysctl");
	return(-1);
    }
    
    in6h = alloc_t(struct in6_addr_head);
    TAILQ_INIT(in6h);
    for (cp = buf; cp < buf + size; cp += ifmam->ifmam_msglen) {
	ifmam = (struct ifma_msghdr *)cp;
	
	/* get the ifa */
	cp1 = (char *)(ifmam + 1);
	len = ifmam->ifmam_msglen;
	ifa = NULL;

	for (i = 0; i < RTAX_MAX; i++)
	    if (ifmam->ifmam_addrs & (1 << i)) {
		sa = (struct sockaddr *)cp1;
		if (i == RTAX_IFA)
		    ifa = sa;
		ADVANCE(cp1, sa);
	    }
	
	if (ifa) {
	    /* create a new minmax entry for the new multicast address */
	    in6e = alloc_t(struct in6_addr_entry);
	    bcopy(&SIN6(ifa)->sin6_addr, &in6e->ia6, sizeof(struct in6_addr));
	    TAILQ_INSERT_TAIL(in6h, in6e, links);
	}
    }
    *in6hp = in6h;
    return(0);
}

int send_msg_to_mars(u_int8_t *msg, u_int msg_len, enum mars_opcode opcode)
{
    int wrote;
   
    if ((p2p2mars == NULL) || (p2p_sd <= 0)) {
	diag(COMPONENT, DIAG_WARN, 
	     "p2p VC with MARS is down, go and re-establish it");
	return(0);
    }
    diag(COMPONENT, DIAG_DEBUG, "TO MARS: %s", mop_name[opcode]);
    if ((wrote = write(p2p_sd, msg, msg_len)) == msg_len)
        return(0);
    if (wrote < 0) {
        diag(COMPONENT, DIAG_ERROR, 
	     "send_msg_to_mars: to_mars write: %s", strerror(errno));
        if (errno == EPIPE)
            diag(COMPONENT, DIAG_ERROR, 
		 "send_msg_to_mars:  MARS is down\n");
        return(wrote);
    }
    diag(COMPONENT, DIAG_ERROR, 
	 "bad MARS write: wanted %d, wrote %d\n", msg_len, wrote);
    return(wrote);

}

/* 
 * MARS client may receive a message either on p2p VC (with MARS)
 * or on CCVC. Distinction is made by using the socket descriptor (sd)
 */
int recv_mars(int sd)
{
    static long abuffer[MAX_BUF_LEN / sizeof(long) + 1];
    u_int8_t *buffer = (u_int8_t *)abuffer;
    u_int8_t *mars_buf;
    int size, error, min_len, mars_len;
    struct atmllc *llc_snap;

    if ((size = read(sd, buffer, MAX_BUF_LEN)) < 0) {
	diag(COMPONENT, DIAG_ERROR, "recv mars: %s", strerror(errno));
	return(-1);       
    }
    if (size == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "recv mars: natm_socket (%d) closed by peer", sd);
	return(-1);
    }

    /* strip off llc/snap encapsulation and check it */
    llc_snap = (struct atmllc *) buffer;
    if (bcmp(llc_snap, &llc_mars_ctl, ATMLLC_LEN) != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "recv_mars: bad mars encapsulation");
	/* discard the message */
	return(-1);
    }

    min_len =  ATMLLC_LEN + sizeof(struct mars_msg);
    if (size < min_len) {
	diag(COMPONENT, DIAG_ERROR,
	     "mars message is too short (got %d < %d)", size, min_len);
	return(-1);
    }
    
    mars_buf = (u_int8_t *) (buffer + ATMLLC_LEN);
    mars_len = size - ATMLLC_LEN;
    
    error = dispatch_mars_msg(sd, mars_buf, mars_len);
    return error < 0 ? -1 : 0;
}

/* specific to MCS */
int forward_multi_data(int sd, u_int8_t *mbuffer, int size)
{
    struct ip6_hdr *ip6h;
    struct in6_addr *dst;
    struct resolv_entry *res_e;
    ATMCALL *mvc;
    int wrote;
    char v[64];

    ip6h = (struct ip6_hdr *) (mbuffer + ATMMLLC_LEN);
    dst = &ip6h->ip6_dst;
    res_e = lookup_ip6m(dst);
    if (res_e == NULL) {
	diag(COMPONENT, DIAG_ERROR,
	     "forward_multi_data: unsupported multicast group %s",
	     inet_ntop(AF_INET6, dst, v, sizeof(v)));
	     return(-1);
    }
    
    mvc = res_e->mvc;
    diag(COMPONENT, DIAG_DEBUG, 
	 "To mgroup %s : (%d bytes)", 
	 inet_ntop(AF_INET6, dst, v, sizeof(v)), size);

    if ((mvc == NULL) || (mvc->sd < 0) || (mvc->state != connected)) {
	diag(COMPONENT, DIAG_ERROR,
	     "forward_multi_data: multicast VC to group %s is down!\n",
	     inet_ntop(AF_INET6, dst, v, sizeof(v)));
	return(-1);
    }
    
    if ((wrote = write(mvc->sd, mbuffer, size)) == size)
        return(0);

    if (wrote < 0) {
        diag(COMPONENT, DIAG_ERROR, 
	     "forward_multi_data: to_mgroup write: %s", strerror(errno));
        return(wrote);
    }

    diag(COMPONENT, DIAG_ERROR, 
	 "forward_multi_data: bad write: wanted %d, wrote %d\n", size, wrote);
    return(wrote);

}
