/* mcl_tab.c - MARS client (CLMB/MCS) table management */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>

#include <net/if.h>
#ifndef if_type
#include <net/if_var.h> 
#endif

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>


#include "../linux-atm.h"

#include "atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_def.h"
#include "mars_msg.h"
#include "tables.h"

#include "mcl_io.h"
#include "mcl_timeout.h"
#include "mcl_tab.h"

#define COMPONENT "MCL_TAB"

const char *resolv_state_name[] = { "NONE", "RESOLV", "VALID",
					 "HOLDDOWN", "REVALID", "MIGR"
				 };

struct resolv_entry *lookup_ip6m(struct in6_addr *ip6m)
{
    struct resolv_entry *walk;
    /* see   IN6_ARE_ADDR_EQUAL(&pa,&pb) */
    /* see /usr/src/local/gated-work/src/bgp4+_sync.c */
    for (walk = LIST_FIRST(&Resolv_Table); walk; walk = LIST_NEXT(walk, links))
	if (SAME_ADDR6(walk->ia6, *ip6m))
	    break;
    return (walk);
}

/*
 * Lookup an unacknowledged outstanding (JOIN/LEAVE) request
 */
struct unacked_req *lookup_unacked(u_int16_t opcode, u_int16_t regist,
				   u_int8_t seq, u_int16_t pnum,
				   struct svc_addr *source_atm, 
				   struct minmax_pair *first_minmax)
{
    struct unacked_req *walk;

    for (walk = LIST_FIRST(&Unacked); walk; walk = LIST_NEXT(walk, links))
	if ((walk->op == opcode) 
	    && (walk->regist == regist) 
	    && (walk->seq == seq) 
	    && (walk->pnum == pnum) 
	    && (atm_equal(&walk->source_atm, source_atm, 0, 0) != 0)) {
	    if (first_minmax != NULL) {
		if (!bcmp(&walk->pair, 
			  first_minmax, sizeof(struct minmax_pair)))
		    break;
	    } else
		break;
	}
    return (walk);
}

int belongs_to_block_list(struct in6_addr *a, struct minmax_head *mpairs)
{
    struct minmax_entry *walk;
    
    /* pairs are supposed to be sorted in ascending order */
    for (walk = TAILQ_FIRST(mpairs); walk; walk = TAILQ_NEXT(walk, links)) {
	if (in6_addr_cmp(a, &walk->pair.min) < 0)
	    return(0);		/* not found */
	if (in6_addr_cmp(a, &walk->pair.max) <= 0) {
	    return(1);
	}
    }	/* out of range */
    return(0);
}
    
struct ept_entry *new_ept_entry(struct svc_addr *rsvcaddr, 
				struct resolv_entry *res_e)

{
    struct ept_entry *new;

    new = alloc_t(struct ept_entry);
    bzero(new, sizeof(struct ept_entry));
    bcopy(rsvcaddr, &new->addr, sizeof(struct svc_addr));
    new->res_e = res_e;
    return(new);
}


/* get the matching ept_entry structure (if any) of the given ATM address */
struct ept_entry *find_ept_entry(struct svc_addr *atm_addr, 
				 struct resolv_entry *res_e)
{
    struct ept_entry *walk  = NULL;
    
    for (walk = LIST_FIRST(&res_e->ept_head); 
	 walk; 
	 walk = LIST_NEXT(walk, links))
	if (atm_equal(atm_addr, &walk->addr, 0, 0) != 0)
	    break;
    return(walk);
}

void free_ept_entries(struct ept_head *head)
{
    struct ept_entry *walk, *next;
    
    for (walk = LIST_FIRST(head); walk; walk = next) {
	next = LIST_NEXT(walk, links);
	STOP_TIMER(walk);
	LIST_REMOVE(walk, links);
	free(walk);
    }
}


/*
 * Discard the whole list of ATM addresses mapping to an IPv6 multicast group
 */
void discard_ept_list(struct resolv_entry *entry)
{
    struct ept_head *head;
    
    if (entry->rstate == rs_resolv)
	head = &entry->ept_head;
    else {
	if (entry->rstate == rs_revalid)
	    head = &entry->new_head;
	else {
	    diag(COMPONENT, DIAG_ERROR, 
		 "discard_ept_list: bad entry multicast state");
	    return;
	}
    }
    free_ept_entries(head);
    entry->discard = 0;
    entry->conn_timer = NULL;
    return;
}


struct resolv_entry *new_resolv_entry(struct in6_addr *ip6m,
				      enum resolv_state rstate,
				      struct atm_qos *qos)

{
    struct resolv_entry *new;

    new = alloc_t(struct resolv_entry);
    bzero(new, sizeof(struct resolv_entry));
    new->rstate = rstate;
    bcopy(ip6m, &new->ia6, IN6_ADDR_LEN);
    new->qos = qos;
    LIST_INIT(&new->ept_head);
    LIST_INIT(&new->new_head);
    return(new);
}

/* duplicate resolv entry for a new p2m call with QoS */
struct resolv_entry *dup_resolv_entry(struct resolv_entry *res_e, 
				      struct atm_qos *qos)
{
    struct resolv_entry *new_res_e;
    struct ept_entry *walk, *new_ept_e;

    new_res_e = new_resolv_entry(&res_e->ia6, rs_valid, qos);
    for (walk = LIST_FIRST(&res_e->ept_head); 
	 walk; 
	 walk = LIST_NEXT(walk, links)) {
	new_ept_e = new_ept_entry(&walk->addr, new_res_e);
	LIST_INSERT_HEAD(&new_res_e->ept_head, new_ept_e, links);
    }

    return(new_res_e);
}
void free_resolv_entry(struct resolv_entry *res_e)
{
    ATMCALL *mvc;

    if (res_e != NULL) {
	if ((mvc = res_e->mvc) != NULL)
	    mvc->res_e = NULL;
	free_ept_entries(&res_e->ept_head);
	/* @@@ paranoid, it should be already done */
	free_ept_entries(&res_e->new_head);
	if (res_e->qos != NULL)
	    free(res_e->qos);
	free(res_e);
    }
}
	
int release_resolved_mgroup(struct in6_addr *ip6m)
{
    struct resolv_entry *res_e;
    ATMCALL *mvc;
    char v[64];
    int error;

    diag(COMPONENT, DIAG_DEBUG, 
	 "release_resolved_mgroup: looking up the local table...");
    /* is it really a multicast address ? */
    if (ip6m == NULL) {
	diag(COMPONENT, DIAG_ERROR, "release_resolved_mgroup: NULL address");
	return(-1);
    }
    if (IN6_IS_ADDR_MULTICAST(ip6m) == 0) {
	diag(COMPONENT, DIAG_ERROR,  
	 "release_resolved_mgroup: not a multicast address [%s]", 
	     inet_ntop(AF_INET6, ip6m, v, sizeof(v)));
	return(-1);
    }

    res_e = lookup_ip6m(ip6m);
    if (res_e != NULL) {
	/* is there any attached multicast VC */
	if ((mvc = res_e->mvc) != NULL) {
	    if (mvc->state != connected) {
		free_atmcall(mvc);
		NbCalls--;
	    } else {
		/* active close MVC */
		error = active_close_atmcall(0, mvc->vcc, mvc);
		if (error != 0) {
		    diag(COMPONENT, DIAG_ERROR,
			 "release_resolved_mgroup: failed to close multicast "
			 "VC [vcc = %d] for mgroup [%s]", mvc->vcc,
			 inet_ntop(AF_INET6, &res_e->ia6, v, sizeof(v)));
		    return(-1);
		}
		return(0);
		/* reslove entry will be released upon 'close' confirm */
	    }
	}
	/* stop potential timer */
	STOP_TIMER(res_e);
	/* put resolve entry in invalid state */
	res_e->rstate = rs_none;
	/* release resolve entry */
	LIST_REMOVE(res_e, links);
	free_resolv_entry(res_e);
	return(0);		
	
    } else {
	diag(COMPONENT, DIAG_ERROR,  
	     "release_resolved_mgroup: no matching reslove entry for "
	     "multicast address [%s]", 
	     inet_ntop(AF_INET6, ip6m, v, sizeof(v)));
	return(-1);
    }
    return(0);
}

	
void free_resolv_entries(struct resolv_head *res_head)
{
    struct resolv_entry *walk, *next;

    
    if (res_head != NULL) {
	for (walk = LIST_FIRST(res_head); walk; walk = next) {
	    next = LIST_NEXT(walk, links);
	    STOP_TIMER(walk);
	    LIST_REMOVE(walk, links);
	    free_ept_entries(&walk->ept_head);
	    free_ept_entries(&walk->new_head);
	    free(walk);
	}
    }
}


int update_membership(struct unacked_req *unacked, u_int16_t fl_layer3grp)
{
    enum mars_opcode opcode;
    struct minmax_pair *pair;
    int error = 0, cmp;

    pair = &unacked->pair;
    if (is_multicast_pair(&pair->min, &pair->max) == 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "update_membership: not a multicast minmax_pair");
	return(-1);
    }
    opcode = (enum mars_opcode) unacked->op;

    switch (opcode) {
    case mars_join:
    case mars_mserv:
	insert_new_mgroup(&Minmax_Table, &pair->min, fl_layer3grp);
	diag(COMPONENT, DIAG_DEBUG, 
	     "update_membership: new minmax pair inserted");
	break;
	
    case mars_leave:
    case mars_unserv:
	/* is it a single group (min == max) */
	cmp = in6_addr_cmp(&pair->min, &pair->max);
	if (cmp == 0) {
	    error = remove_mgroup(&Minmax_Table, &pair->min);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "update_membership: failed to remove min_max pair");
		break;
	    }
	
	    diag(COMPONENT, DIAG_DEBUG, 
		 "update_membership: minmax pair removed");
	} else { /* cmp != 0 */
	    struct in6_addr min_multi, max_multi;
	    int ret1, ret2, cmp1, cmp2;
	    struct minmax_entry *walk, *next;
	    
	    ret1 = inet_pton(AF_INET6, MIN_MULTI_ADDR, &min_multi);
	    ret2 = inet_pton(AF_INET6, MAX_MULTI_ADDR, &max_multi);
	    if ((ret1 < 0) || (ret2 < 0)) {
		diag(COMPONENT, DIAG_ERROR,
		     "update_membership: inet_pton: %s\n", strerror(errno));
		return(-1);
	    }
	    if ((ret1 == 0) || (ret2 == 0)) {
		/* paranoid, unlikely to happen */
		diag(COMPONENT, DIAG_ERROR, 
		     "update_membership: unparsable multicast min_max pair");
		return(-1);
	    }
	    cmp1 = in6_addr_cmp(&pair->min, &min_multi);
	    cmp2 = in6_addr_cmp(&pair->max, &max_multi);
	    if ((cmp1 ==0) && (cmp2 == 0)) {
		/* mars client is leaving all that it has joined */
		for (walk = TAILQ_FIRST(&Minmax_Table); walk; walk = next) {
		    next = TAILQ_NEXT(walk, links);
		    TAILQ_REMOVE(&Minmax_Table, walk, links);
		    free(walk);
		}
	    } else {
		/* not supported yet */
		diag(COMPONENT, DIAG_ERROR,
		     "update_membership: not supported yet");
		error = -1;
	    }
	}

	break;

    default:
	diag(COMPONENT, DIAG_ERROR, 
		 "update_membership: unexpected mars code %d", opcode);
	error = -1;
    }
    LIST_REMOVE(unacked, links);
    free(unacked);
    return(error);
}

/* 
 * When the client receives MULTI message(s) and if the the p2mp VC is 
 * under validation, a new mapping ATM address list is supposed to replace 
 * the old one. The p2mp update is performed with a sequence of DROP_PARTY (for 
 * nodes which are no longer member of the group) and ADD_PARTY (for new nodes).
 * The routine below constructs the two corresponding lists.
 */

void build_drop_add_lists(struct resolv_entry *entry, 
			  struct ept_head *to_drop_head)
{
    struct ept_entry *walk1, *walk2, *next;
    struct ept_head *head, *new_head;

    head = &entry->ept_head;
    new_head = &entry->new_head;
    /* 
     * at the end of this loop, the new_list will be reduced to 
     * the set of nodes to ADD 
     */
    for (walk1 = LIST_FIRST(head); walk1; walk1 = next) {
	for (walk2 = LIST_FIRST(new_head); 
	     walk2; 
	     walk2 = LIST_NEXT(walk2, links))
	    if (atm_equal(&walk2->addr, &walk1->addr, 0, 0) != 0)
		break;
	next = LIST_NEXT(walk1, links);
	if (walk2 != NULL) {
	    /* the node is still member of the group,  shrink "new_list" */
	    LIST_REMOVE(walk2, links);
	} else {
	    /* entry walk1 doesn't belong to the new resolve list 
	       ==> to be deleted */
	    LIST_REMOVE(walk1, links);
	    LIST_INSERT_HEAD(to_drop_head, walk1, links);
	}
    }
    return;
}

/* Build p2mp VC with the corresponding QoS (if any) */
int build_mvc(struct resolv_entry *entry)
{
    struct ept_entry *walk, *start;
    struct ept_head *head;
    u_long rnd_delay;
    int error, retry = 1;
    

    /* 
     * find the first non tagged address entry and attempt to initiate 
     * the p2mp call with it (cf 5.1.3) 
     */
    head = &entry->ept_head;
    start = LIST_FIRST(head);
    while (retry) {
	for (walk = start; walk; walk = LIST_NEXT(walk, links))
	    if (walk->tag_flag == 0)
		break;
	if (walk != NULL) {
	    error = atmcall_setup(&walk->addr, NULL, &entry->ia6, 
				  entry->qos, 0, walk, NULL);
	    if (error != 0) {
		/* next time we'll start retrying from here */
		start = LIST_NEXT(walk, links);
		diag(COMPONENT, DIAG_ERROR, 
		     "build_mvc: failed to initiate p2mp VC");
		/* drop this endpoint from the list */
		LIST_REMOVE(walk, links);
		free(walk);
		continue;
	    } else {
		diag(COMPONENT, DIAG_DEBUG, 
		     "build_mvc: waiting for response from atmsigd");
		return(0);
	    }
	} else {
	    diag(COMPONENT, DIAG_ERROR, 
		 "build_mvc: failed, no more endpoint to be tried: "
		 "all addresses are tagged");
	    /* start random timer on all tagged ept entries to be retried */
	    for (walk = LIST_FIRST(head); walk; walk = LIST_NEXT(walk, links)) {
		walk->new_delay = get_random_delay(5*T_SEC, 10*T_SEC);
		rnd_delay = walk->new_delay;
		start_timer(rnd_delay, on_RANDOM_FIRST_EPT, walk);
	    }
	    return(-1);
	}
    }
    return(0);
}

/* 
 * Upon building a new p2mp VC, it is possible to send a `bloc' 
 * of ADD_PARTY for all the ept entries
 */
int add_bloc_parties(ATMCALL *p2mcall, struct resolv_entry *res_e)
{
    struct ept_entry *walk;
    struct ept_head *head;
    PARTY *party;
    int i, error = 0, first = 1;
    
    head = &res_e->ept_head;
    for (walk = LIST_FIRST(head); walk; walk = LIST_NEXT(walk, links)) {
	party = p2mcall->parties;
	if (first && atm_equal(&party->rsvcaddr, &walk->addr, 0, 0)) {
	    /* it's the first initiated party */
	    first = 0;
	    continue;
	} else {
	    /* reset by the way tag_flag (for future use of res_e) */
	    walk->tag_flag = 0;
	    /* 
	     * party will be chained with ept entry 
	     */
	    party = active_add_party(p2mcall, &walk->addr, 
				     NULL, walk, NULL);
		if (party == NULL) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "add_bloc_parties: failed to initiate adding leaf: ");
		    for (i = 0; i < ATM_ESA_LEN; i++)
			printf("%02X", walk->addr.prv[i]);
		    printf("\n");
		    /* mark something has gone wrong */
		    error = -1;
		}
	}
    }
    return(error);
}

/* 
 * This is typically the case for MCS. close every multicast VC  
 * corresponding to a multicast address belonging to the min_max pair 
 */
int close_minmax_mvc(struct minmax_pair *pair) 
{
    struct in6_addr ip6m1, ip6m2;
    struct resolv_entry *res_e;
    int cmp, ret1, ret2, error, equal;
    char v[64];

    if (pair == NULL) {
	diag(COMPONENT, DIAG_ERROR, "close_minmax_mvc: minmax_pair is empty");
	return(-1);
    }
    
    ret1 = inet_pton(AF_INET6, MIN_MULTI_ADDR, &ip6m1);
    ret2 = inet_pton(AF_INET6, MAX_MULTI_ADDR, &ip6m2);

    if ((ret1 < 0) || (ret2 < 0)) {
	diag(COMPONENT, DIAG_ERROR,
	     "close_minmax_mvc: inet_pton: %s\n", strerror(errno));
	return(-1);
    }
    if ((ret1 == 0) || (ret2 == 0)) {
	/* paranoid, unlikely to happen */
	diag(COMPONENT, DIAG_ERROR, 
	     "close_minmax_mvc: unparsable multicast min_max pair");
	return(-1);
    }
    
    ret1 = in6_addr_cmp(&pair->min, &ip6m1);
    ret2 = in6_addr_cmp(&pair->max, &ip6m2);
    
    if ((ret1 == 0) && (ret2 == 0)) {
	
	/* Close all open multicast VCs */
	for (res_e = LIST_FIRST(&Resolv_Table); 
	     res_e; 
	     res_e = LIST_NEXT(res_e, links)) {
	    
	    diag(COMPONENT, DIAG_DEBUG, 
		 "close_minmax_mvc: closing mvc for mgroup [%s]", 
		 inet_ntop(AF_INET6, &res_e->ia6, v, sizeof(v)));
	    
	    error = active_close_atmcall(0, res_e->mvc->vcc, res_e->mvc);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "close_minmax_mvc: failed to close multicast "
		     "VC [vcc = %d] for mgroup [%s]", res_e->mvc->vcc,
		     inet_ntop(AF_INET6, &res_e->ia6, v, sizeof(v)));
		return(-1);
	    }
	    /* reslov entry will be released upon 'close' confirm */
	}

	return(0);
    } /* close all MVCs */
    
    bcopy(&pair->min, &ip6m1, sizeof(ip6m1));
    cmp = in6_addr_cmp(&pair->min, &pair->max);
    equal = (cmp == 0);
    while (cmp <= 0) {
	res_e = lookup_ip6m(&ip6m1);
	if (res_e != NULL) {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "close_minmax_mvc: closing mvc for mgroup [%s]", 
		 inet_ntop(AF_INET6, &ip6m1, v, sizeof(v)));
	    
	    error = active_close_atmcall(0, res_e->mvc->vcc, res_e->mvc);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "close_minmax_mvc: failed to close multicast "
		     "VC [vcc = %d] for mgroup [%s]", res_e->mvc->vcc,
		     inet_ntop(AF_INET6, &ip6m1, v, sizeof(v)));
		return(-1);
	    }
	    /* reslov entry will be released upon 'close' confirm */
	} else {
	    diag(COMPONENT, DIAG_WARN, 
		 "close_minmax_mvc: no matching resolve entry for "
		 "mgroup [%s]", 
		 inet_ntop(AF_INET6, &ip6m1, v, sizeof(v)));
	}
	if (equal)
	    break;
	error = increment_addr(&ip6m1, &ip6m2);
	if (error != 0) {
	    diag(COMPONENT, DIAG_WARN,
		 "close_minmax_mvc: increment failed");
	    return(-1);
	}
	bcopy(&ip6m2, &ip6m1, sizeof(ip6m1));
	cmp = in6_addr_cmp(&ip6m1, &pair->max);
    }
    return(0);
    
}

int update_revalid(struct resolv_entry *entry)
{
    struct ept_entry *walk, *next, *copy, *last = NULL;
    struct ept_head *to_drop_head, *to_add_head, *head;
    PARTY *party;
    int error, i, ret = 0;

    /* build "drop" and "add" lists */
    to_drop_head = alloc_t(struct ept_head);
    LIST_INIT(to_drop_head);
    build_drop_add_lists(entry, to_drop_head);
    to_add_head = &entry->new_head;
    head = &entry->ept_head;
    /* drop obsolete parties */
    for (walk = LIST_FIRST(to_drop_head); walk; walk = next) {
	next = LIST_NEXT(walk, links);
	if (next == NULL) {
	    /* 
	     * it's the last party, don't remove it now but mark it 
	     * and drop it later
	     * This is optimistic because we hope this last party wobn't 
	     * drop itself at the same moment :-)
	     */
	    last = walk;
	    break;
	}
	party = walk->party;
	if (party != NULL) {
	    error = active_drop_party(0, 0, party);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "update_revalid: failed to initiate dropping leaf: ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02X", walk->addr.prv[i]);
		printf("\n");
		/* rollback */
		LIST_INSERT_HEAD(head, walk, links);
		/* mark something has gone wrong */
		ret = -1;
	    }
	} else {
	    diag(COMPONENT, DIAG_WARN, 
		 "update_revalid: no party matching ept_e: ");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", walk->addr.prv[i]);
	    printf("\n");
	}
    }
    
    /* add new parties */
    for (walk = LIST_FIRST(to_add_head); walk; walk = LIST_NEXT(walk, links)) {
	party = active_add_party(entry->mvc, &walk->addr, NULL, walk, NULL);
	if (party == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "update_revalid: failed to initiate adding leaf : ");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", walk->addr.prv[i]);
	    printf("\n");
	    /* mark something has gone wrong */
	    ret = -1;
	} else {
	    /* copy this new ept entry to the resolve list */
	    copy = alloc_t(struct ept_entry);
	    bcopy(walk, copy, sizeof(struct ept_entry));
	    party->ept_e = copy;
	    LIST_INSERT_HEAD(head, copy, links);
	}
    }
    
    /* go back to the remaining party to drop (if any) */
    if (last != NULL) {
	party = last->party;
	if (party != NULL) {
	    error = active_drop_party(0, 0, party);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "update_revalid: failed to initiate "
		     "dropping leaf : ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02X", last->addr.prv[i]);
		printf("\n");
		/* rollback */
		LIST_INSERT_HEAD(head, last, links);
		/* mark something has gone wrong */
		ret = -1;
	    }
	} else {
	    diag(COMPONENT, DIAG_WARN, "no such party: ");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", last->addr.prv[i]);
	    printf("\n");
	}
    }
    
    /* free useless lists */
    free_ept_entries(to_drop_head);
    free(to_drop_head);
    free_ept_entries(to_add_head);
    entry->discard = 0;
    entry->conn_timer = NULL;
    return(ret);
}

/* 
 * add this cluster member given by its ATM address as a leaf to the 
 * multicast group(s) which the member has joined and which we have an open 
 * p2mp VC to. Update Resolv corresponding entries.
 */
void add_new_leaves(struct svc_addr *source_atm, struct minmax_head *mpairs)
{
    struct resolv_entry *res_e;
    struct ept_entry *ept_e;
    PARTY *party;
    int i, new = 0;
    
    for (res_e = LIST_FIRST(&Resolv_Table); 
	 res_e; 
	 res_e = LIST_NEXT(res_e, links))

	if ((res_e->rstate == rs_valid) 
	    && (belongs_to_block_list(&res_e->ia6, mpairs))) {
	    ept_e = find_ept_entry(source_atm, res_e);
	    if (ept_e == NULL) {
		/* create a new ept_entry */
		new = 1;
		ept_e = alloc_t(struct ept_entry);
		bzero(ept_e, sizeof(struct ept_entry));
		bcopy(source_atm, &ept_e->addr, sizeof(struct svc_addr));
	    } else {
		diag(COMPONENT, DIAG_WARN, 
		     "add_new_leaves: ept entry already exists: ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02X", source_atm->prv[i]);
		printf("\n");
		if (ept_e->party != NULL) {
		    diag(COMPONENT, DIAG_WARN, 
			 "add_new_leaves: party already exists: ");
		    for (i = 0; i < ATM_ESA_LEN; i++)
			printf("%02X", source_atm->prv[i]);
		    printf("\n");
		    return;
		}
	    }
	    /* no party associated yet, initiate adding party */
	    party = active_add_party(res_e->mvc, source_atm, NULL, ept_e, NULL);
	    if (party == NULL) {
		diag(COMPONENT, DIAG_ERROR, 
		     "add_new_leaves: failed to initiate adding leaf: ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02X", source_atm->prv[i]);
		printf("\n");
		if (new == 0) 
		    LIST_REMOVE(ept_e, links);
		free(ept_e);
	    } else {
		diag(COMPONENT, DIAG_DEBUG, 
		     "add_new_leaves: initiated adding leaf: ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02X", source_atm->prv[i]);
		printf("\n");
		party->ept_e = ept_e;
		ept_e->party = party;
		LIST_INSERT_HEAD(&res_e->ept_head, ept_e, links);
	    }
	}
    return;
}

void drop_leaves(struct svc_addr *source_atm, struct minmax_head *mpairs)
{
    struct resolv_entry *res_e;
    struct ept_entry *ept_e;
    int error, i;

    for (res_e = LIST_FIRST(&Resolv_Table); 
	 res_e; 
	 res_e = LIST_NEXT(res_e, links))

	if ((res_e->rstate == rs_valid) 
	    && (belongs_to_block_list(&res_e->ia6, mpairs))) {
	    ept_e = find_ept_entry(source_atm, res_e);
	    if (ept_e != NULL) {
		if (ept_e->party != NULL) {
		    error = active_drop_party(0, 0, ept_e->party);
		    if (error != 0) {
			diag(COMPONENT, DIAG_ERROR, 
			     "drop_leaves: failed to drop leaf: ");
			for (i = 0; i < ATM_ESA_LEN; i++)
			    printf("%02X", source_atm->prv[i]);
			printf("\n");
			return;
		    }
		} else {
		    diag(COMPONENT, DIAG_WARN, 
			 "drop_leaves: no party is associated to ept entry: ");
		    for (i = 0; i < ATM_ESA_LEN; i++)
			printf("%02X", source_atm->prv[i]);
		    printf("\n");
		    LIST_REMOVE(ept_e, links);
		    return;
		}
	    } else {
		diag(COMPONENT, DIAG_WARN, "no such ept entry: ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02X", source_atm->prv[i]);
		printf("\n");
	    }
	}
    return;
}

void revalidate_all(void)
{
    struct resolv_entry *walk;
    u_long rnd_delay;

    for (walk = LIST_FIRST(&Resolv_Table); walk; walk = LIST_NEXT(walk, links))
	if ((walk->mvc != NULL) && (walk->mvc->state == connected)) {
	    rnd_delay = get_random_delay(1*T_SEC, 10*T_SEC);	/* cf 5.1.5 */
	    start_timer(rnd_delay, on_RANDOM_REVALID, walk);
	}
    return;
}

void release_all_tables(void)
{

    struct unacked_req *walk_u, *next_u;
    struct minmax_entry *walk_m, *next_m;

    /* force releasing Minmax_Table */
    
    for (walk_m = TAILQ_FIRST(&Minmax_Table); walk_m; walk_m = next_m) {
	next_m = TAILQ_NEXT(walk_m, links);
	TAILQ_REMOVE(&Minmax_Table, walk_m, links);
	free(walk_m);
    }
    
    /* force releasing Resolv_Table */

    if (LIST_FIRST(&Resolv_Table) != NULL)
	free_resolv_entries(&Resolv_Table);
    
    /* force releasing Unacked table */
    
    for (walk_u = LIST_FIRST(&Unacked); walk_u; walk_u = next_u) {
	next_u = LIST_NEXT(walk_u, links);
	LIST_REMOVE(walk_u, links);
	STOP_TIMER(walk_u);
	free(walk_u);
    }
}

/* ****************** Display Internal Structures ********************* */


/* common to CLMB and MCS depending on 'mtype' variable */
void show_membership(enum mcl_type mtype)
{
    struct minmax_entry *mme;
    char v[64], v1[64];
    int fl3;

    printf("<%s state = %s>\t<CMI = %d>\n\n", mcl_name[mtype], 
	       mcl_state_name[mcl.mcl_state], mcl.cmi);
    if (p2p2mars != NULL)
	printf("P2P VC with MARS:\t<vcc, vpi, vci> = <%ld, %d, %d>\n", 
	       p2p2mars->vcc, p2p2mars->vpi, p2p2mars->vci);
    
    if (Cvc != NULL) {
	printf("%s Control VC:\t<vcc, vpi, vci> = <%ld, %d, %d>\n",
	       (mtype == clmb_type) ? "Cluster" : "Server",
	       Cvc->vcc, Cvc->vpi, Cvc->vci);
	printf("\t\t\tParty:\t[ep_ref = %d]\t<MSN = %d>\n\n", 
	       Cvc->parties->ep_ref, mcl.mcl_sn);
    }

    printf("\t*********     %s Multicast Groups    ***********\n",
	   (mtype == clmb_type) ? "Joined" : "Supported");
    printf("\tMin_Max pair\t\t\tLayer_3_Group flag\n\n");
    
    for (mme = TAILQ_FIRST(&Minmax_Table); mme; mme = TAILQ_NEXT(mme, links)) {
	fl3 = ((mme->fl_layer3grp &  FL_LAYER3GRP) != 0);
	printf("\t[%s, %s]\t\t%d\n", 
	       inet_ntop(AF_INET6, &mme->pair.min, v, sizeof(v)),
	       inet_ntop(AF_INET6, &mme->pair.max, v1, sizeof(v1)),
	       fl3);
    }
    return;
}


void show_resolved_mgroups(void)
{
    struct resolv_entry *res_e;
    struct ept_entry *ept_e;
    char v[64];
    int i; 

    printf("*******      Resolved Multicast Groups  **************\n\n");
    printf("Multicast Address\t\tResolve entry state\n\n");
    for (res_e = LIST_FIRST(&Resolv_Table); 
	 res_e; 
	 res_e =  LIST_NEXT(res_e, links)) {
	printf("%s\t\t\t\t%s\n",
	       inet_ntop(AF_INET6, &res_e->ia6, v, sizeof(v)),
	       resolv_state_name[res_e->rstate]);
	if (res_e->mvc != NULL)
	    printf("Multicast VC:\t<vcc, vpi, vci> = <%ld, %d, %d>\n",
		   res_e->mvc->vcc, res_e->mvc->vpi, res_e->mvc->vci);
	
	for (ept_e = LIST_FIRST(&res_e->ept_head);
	     ept_e;
	     ept_e = LIST_NEXT(ept_e, links)) {
	    printf("\t");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02x", ept_e->addr.prv[i]);
	    if (ept_e->party != NULL)
		printf("\t[ep_ref = %d]\n", ept_e->party->ep_ref);
	    else
		printf("\t[ep_ref = ???]\n");
	}
	printf("\n");
    }
    return;
}

void show_p2m_calls(void)
{
    ATMCALL *call;
    PARTY *party;
    char v[64];
    int i;

    printf("\n\t\t\t\tCURRENT p2m ATM CALLS\n\n");
    printf("\nCnf  Own   VCC   VPI   VCI   Rmt                EP_REF  C_State"
	   "   EP_State  Rt\n\n");
    for (call = AtmCalls; call != NULL; call = call->next) {
	if (call->conf != root)
	    continue;
	
	printf("root %-6s%-6ld%-6d%-6d****************           "
	       "%-10s          %-7s\n", call->owner ? "Y" : "N",
	       call->vcc, call->vpi, call->vci,
	       CallStateName[call->state],
	       call->route ? "Y" : "N");
	printf("\t\t%s\n",  inet_ntop(AF_INET6, &call->rin6, v, sizeof(v)));
	    for (party = call->parties; party != NULL; party = party->next) {
		if (party->rsvcaddr.prv[0] != 0) {
		    printf("\t");
		    for (i = 0; i < ATM_ESA_LEN; i++)
			printf("%02x", party->rsvcaddr.prv[i]);
		}
		printf("  %-8d        %-10s\n", party->ep_ref, 
		       PartyStateName[party->pstate]);
	    }
	    printf("\n");
    }
}

int close_all_p2m_calls(void)
{
    ATMCALL *call;
    int error;

    for (call = AtmCalls; call; call = call->next) {
	if (call->conf == root) {
	     if ((error = active_close_atmcall(0, call->vcc, call)) != 0) {
		 diag(COMPONENT, DIAG_ERROR, 
		      "close_all_p2m_calls: failed to close all p2m ATM calls");
		 return(error);
	     }
	 }
    }
    return(0);
}

