/* mcl_timeout.c - Processing of MARS client(CLMB/MCS) timeout events */
 
/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */
 
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <math.h>
#include <string.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"


#include "mars_msg.h"
#include "tables.h"

#include "mcl_tab.h"
#include "mcl_io.h"
#include "to_mars.h"
#include "mcl_timeout.h"

#define COMPONENT "TIMEOUT"

/* (cf 5.1.1) Upon receiving MARS_NAK, delay sending out a new MARS_REQUEST 
 * 5 <= T <= 10 seconds, choice : T = T_NEW_REQ = 10 s
 * During "holddown" period, every new resolution request will be ignored.
 * On timeout, the resolve entry will be simply removed
 */
void on_T_HOLD_REQ(void *user) 
{
    struct resolv_entry *entry = user;
    struct in6_addr *ip6m;
    ATMCALL *call;
    char v[64];

    ip6m = &entry->ia6;
    diag(COMPONENT, DIAG_DEBUG, "T_HOLD_REQ on resolve entry [%s]",
	 inet_ntop(AF_INET6, ip6m, v, sizeof(v)));
    entry->conn_timer = NULL;
    diag(COMPONENT, DIAG_DEBUG, "Releasing resolve entry [%s]", 
	 inet_ntop(AF_INET6, ip6m, v, sizeof(v)));
    LIST_REMOVE(entry, links);
    /* send a warning for every p2m call open toward this group */
    for (call = AtmCalls; call; call = call->next)
	if ((call->conf == root) && !bcmp(&call->rin6, ip6m, IN6_ADDR_LEN))
	    diag(COMPONENT, DIAG_WARN, 
		 "p2m Call with [vcc = %d] should be closed, its resolve "
		 "entry has become obsolete !!!", call->vcc);
    return;
}

/* (cf 5.1.1) Waiting for the last multi seq. Default = 10 s */
void on_T_LAST_SEQ(void *user) 
{
    struct resolv_entry *entry = user;
    char v[64];

    diag(COMPONENT, DIAG_DEBUG, "T_LAST_SEQ on entry [%s]",
	 inet_ntop(AF_INET6, &entry->ia6, v, sizeof(v)));
    if (entry->mvc != NULL) {
	if (entry->mvc->state != connected)
	    diag(COMPONENT, DIAG_WARN, "p2mp VC is not connected!!!");
	else {
	    entry->rstate = rs_revalid;
	    entry->mvc->revalidate = 1;
	}
    }
    discard_ept_list(entry);
    entry->conn_timer = NULL;
    (void) retransmit_mars_req(entry);
    return;
}

/* (cf 5.2.2) Waiting for join/leave response. 5 <= T, default = 10 s */
void on_T_MEMBERSHIP(void *user)
{
    struct unacked_req *unacked = user;

    
    diag(COMPONENT, DIAG_DEBUG, 
	 "T_MEMBERSHIP on JOIN/LEAVE request number [%d]", unacked->seq);
    unacked->conn_timer = NULL;
    retransmit_membership_req(unacked);
    return;
}

/* starts a random timer between t_min and t_max */
u_long get_random_delay(u_long t_min, u_long t_max) 
{
    double rnd;
    u_long rnd_delay;

    rnd = floor((t_max - t_min)*((double) random() / (double) 0xffffffff));
    rnd_delay = t_min + (u_long) rnd;
    return(rnd_delay);
}


void on_RANDOM_FIRST_EPT(void *user)
{
    struct ept_entry *ept_e = user, *walk;
    struct resolv_entry *res_e;
    ATMCALL *call;
    struct ept_head *head;
    int i, error;

    /* reset timer and ept_state */
    ept_e->conn_timer = NULL;
    ept_e->tag_flag = 0;
    diag(COMPONENT, DIAG_DEBUG, "RANODM_FIRST_EPT on endpoint entry");
    for (i = 0; i < ATM_ESA_LEN; i++)
	diag(COMPONENT, DIAG_DEBUG, "%02X", ept_e->addr.prv[i]);
    /* 
     * check if ever the p2m VC in question already exists 
     * (probably under construction) 
     */
    res_e = ept_e->res_e;
    call = res_e->mvc;
    if ((call != NULL) && (call->state == connecting)) {
	/* stop all the other timers */
	head = &res_e->ept_head;
	for (walk = LIST_FIRST(head); walk; walk = LIST_NEXT(walk, links)) {
	    STOP_TIMER(walk);
	    /* if the p2mp VC is currently under construction, double my 
	       timer delay */
	    if (walk == ept_e)
		ept_e->new_delay *= 2;
	    start_timer(ept_e->new_delay, on_RANDOM_FIRST_EPT, walk);
	}		/* else, we're done and no more timers */
	return;
    }

    error = atmcall_setup(&ept_e->addr, NULL, NULL, res_e->qos, 0, ept_e, NULL);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR, "build_mvc: failed to initiate p2mp VC");
	/* 
	 * if this is the default p2m VC (default qos))
	 * drop this endpoint from the list 
	 */
	LIST_REMOVE(ept_e, links);
	free(ept_e);
    } else
	diag(COMPONENT, DIAG_DEBUG, 
	     "build_mvc: waiting for response from atmsigd");
    return;
}   

/* 
 * (cf 5.1.3) Random time to wait before reissueing L_MULTI_ADD (or L_MULTI_RQ)
 * for tagged addresses. 5<= T <= 10 (seconds).
 */
void on_RANDOM_EPT(void *user)
{
    struct ept_entry *ept_e = user;
    PARTY *party;
    int i;

    ept_e->conn_timer = NULL;
    diag(COMPONENT, DIAG_DEBUG, "RANODM_EPT on endpoint entry");
    for (i = 0; i < ATM_ESA_LEN; i++)
	diag(COMPONENT, DIAG_DEBUG, "%02X", ept_e->addr.prv[i]);
    party = active_add_party(ept_e->res_e->mvc, &ept_e->addr, NULL, 
			     ept_e, NULL);
		if (party == NULL) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "active_add_party: failed to initiate adding leaf: ");
		    for (i = 0; i < ATM_ESA_LEN; i++)
			diag(COMPONENT, DIAG_DEBUG, "%02X", ept_e->addr.prv[i]);
		}
}   

/* 
 * (cf 5.1.5) Random time to wait before setting REVALIDATE flag 
 * of one specific p2mp VC or of all open p2mp VCs. 1<= T <= 10 (seconds).
 */
void on_RANDOM_REVALID(void *user)
{
    struct resolv_entry *res_e = user;
    char v[64];

    diag(COMPONENT, DIAG_DEBUG, "RANODM_REVALID on entry [%s]",
	 inet_ntop(AF_INET6, &res_e->ia6, v, sizeof(v)));
    res_e->conn_timer = NULL;
    if (res_e->mvc != NULL) {
	if (res_e->mvc->state != connected)
	    diag(COMPONENT, DIAG_WARN, "p2mp VC is not connected!!!");
	else {
	    res_e->rstate = rs_revalid;
	    res_e->mvc->revalidate = 1;
	}
    }
    discard_ept_list(res_e);
    (void) retransmit_mars_req(res_e);
    return;
}
