/* to_mars.c - Generate common MCL/MCS outgoing MARS messages */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>

#include <net/if.h>
#include <net/if_atm.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>


#include "../linux-atm.h"
#include "atm.h"
#include "atmsap.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"

#include "mars_def.h"
#include "mars_msg.h"
#include "tables.h"

#include "mcl_tab.h"
#include "mcl_timeout.h"
#include "to_mars.h"
#include "mcl_io.h"

#define COMPONENT "TO_MARS"


/* registering a mars client (CLMB/MCS), depending on mars_opcode */
int register_mars_client(enum mars_opcode opcode)
{
    u_int8_t *msg;
    struct mars_memb *memb_msg;
    u_int16_t msg_len;
    u_int8_t *start;
    u_int16_t op, pnum, flags;
    struct unacked_req *unacked;
    int error;


    diag(COMPONENT, DIAG_DEBUG, 
	 "register_mars_client: %s registering with MARS...", 
	 mcl_name[mcl.mcl_type]); 

    if (mcl.mcl_state == mcl_registered) {
	diag(COMPONENT, DIAG_WARN, 
	 "register_mars_client: %s already registered. No message sent.",
	     mcl_name[mcl.mcl_type]);
	return(0);
    }

    /* change mars client state */
    mcl.mcl_state = mcl_registering;

    /* 
     * if p2p2mars is not opened yet, start by establishing it 
     * and come back later when it is active 
     */

    if (p2p2mars == NULL) {

	if ((error = atmcall_setup(&mcl.mars_svc_addr, NULL, 
				   &mcl.mars_ia6, &qos_default_p2p, 
				   -1, NULL, NULL)) != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
	     "register_mars_client: failed to initiate connection to MARS");
	    exit(-1);
	}
	/* come back later */
	return(0);
    }

    if (p2p2mars->state != connected) {
	diag(COMPONENT, DIAG_ERROR, 
	     "register_mars_client: p2p VC with MARS is not "
	     "in `connected' state");
	exit(-1);
    }

    /* create the message to send */
    msg_len = ATMLLC_LEN + sizeof(struct mars_memb) + 
	IN6_ADDR_LEN + ATM_ESA_LEN;
    msg = (u_int8_t *) malloc(msg_len);
    bzero(msg, msg_len);
    bcopy(&llc_mars_ctl, msg, ATMLLC_LEN);
    /* Prepare at the same time a structure to keep track of the request */
    unacked = alloc_t(struct unacked_req);
    bzero(unacked, sizeof(*unacked));
    unacked->op = op = (u_int16_t) opcode;
    unacked->regist = FL_REGISTER;
    unacked->seq = 0;		/* disable this field for the moment */
    unacked->pnum = pnum = 0;
    bcopy(&mcl.svc_addr, &unacked->source_atm, sizeof(struct svc_addr));
    unacked->retries = MAX_RETRIES;
    flags = FL_REGISTER;	/* all other flags are reset */
    memb_msg = (struct mars_memb *) (msg + ATMLLC_LEN);
    format_memb_msg(memb_msg, op, pnum, flags);
    start = ((u_int8_t *) memb_msg) + sizeof(struct mars_memb);
    set_info(&start, mcl.svc_addr.prv, ATM_ESA_LEN);
    set_info(&start, &mcl.ia6, IN6_ADDR_LEN);

    /* copy the message in unacked structure for potential retransmission */
    unacked->msg = (u_int8_t *) malloc(msg_len);
    bcopy(msg, unacked->msg, msg_len);
    unacked->msg_len = msg_len;

    if ((error = send_msg_to_mars(msg, msg_len, opcode)) != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "register_mars_client: failed to send message to MARS");
	/* try to retransmit immediately as if timer has just expired */
	on_T_MEMBERSHIP(unacked);
	return(error);
    }
    START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
    LIST_INSERT_HEAD(&Unacked, unacked, links);
    if (msg != NULL)
	free(msg);
    return(0);
}


int deregister_mars_client(enum mars_opcode opcode)
{
    u_int8_t *msg;
    struct mars_memb *memb_msg;
    u_int16_t msg_len;
    u_int8_t *start;
    u_int16_t op, pnum, flags;
    struct unacked_req *unacked;
    int error;

    diag(COMPONENT, DIAG_DEBUG, 
	 "deregister_mars_client: %s de-registering with MARS...",
	 mcl_name[mcl.mcl_type]); 

    if (mcl.mcl_state == mcl_none) {
	diag(COMPONENT, DIAG_WARN, 
	 "deregister_mars_client: %s already de-registered. No message sent.",
	     mcl_name[mcl.mcl_type]);
	return(0);
    }

    if (p2p2mars == NULL) {
	diag(COMPONENT, DIAG_ERROR, 
	     "deregister_mars_client: p2p VC with MARS no longer exists");
	return(-1);
    }

    /* change mars client state */
    mcl.mcl_state = mcl_deregistering;

    if (p2p2mars->state != connected) {
	diag(COMPONENT, DIAG_ERROR, 
	     "deregister_mars_client: p2p VC with MARS is not "
	     "in `connected' state");
	return(-1);
    }

    /* create the message to send */
    msg_len = ATMLLC_LEN + sizeof(struct mars_memb) + 
	IN6_ADDR_LEN + ATM_ESA_LEN;
    msg = (u_int8_t *) malloc(msg_len);
    bzero(msg, msg_len);
    bcopy(&llc_mars_ctl, msg, ATMLLC_LEN);
    /* Prepare at the same time a structure to keep track of the request */
    unacked = alloc_t(struct unacked_req);
    bzero(unacked, sizeof(*unacked));
    unacked->op = op = (u_int16_t) opcode;
    unacked->regist = FL_REGISTER;
    unacked->seq = 0;		/* disable this field for the moment */
    unacked->pnum = pnum = 0;
    bcopy( &mcl.svc_addr, &unacked->source_atm, sizeof(struct svc_addr));
    unacked->retries = MAX_RETRIES;
    flags = FL_REGISTER;	/* all other flags are reset */
    memb_msg = (struct mars_memb *) (msg + ATMLLC_LEN);
    format_memb_msg(memb_msg, op, pnum, flags);
    start = ((u_int8_t *) memb_msg) + sizeof(struct mars_memb);
    set_info(&start, mcl.svc_addr.prv, ATM_ESA_LEN);
    set_info(&start, &mcl.ia6, IN6_ADDR_LEN);

    /* copy the message in unacked structure for potential retransmission */
    unacked->msg = (u_int8_t *) malloc(msg_len);
    bcopy(msg, unacked->msg, msg_len);
    unacked->msg_len = msg_len;

    if ((error = send_msg_to_mars(msg, msg_len, opcode)) != 0)
	{
	    diag(COMPONENT, DIAG_ERROR, 
		 "deregister_mars_client: failed to send msg %s to MARS",
		 mop_name[opcode]);
	    /* try to retransmit immediately as if timer has just expired */
	    on_T_MEMBERSHIP(unacked);
	    return(error);
	}
    START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
    LIST_INSERT_HEAD(&Unacked, unacked, links);
    if (msg != NULL)
	free(msg);
    return(0);
}


/* 
 * this routine can be called either by a cluster member in order 
 * to join a given min_max pair or by an MCS to serve a given min_max pair
 */
int join_serve_mgroup(struct in6_addr *ip6m, 
		      enum mars_opcode opcode,
		      u_int16_t fl_layer3grp)
{
    u_int8_t *msg;
    struct mars_memb *memb_msg;
    u_int16_t msg_len;
    u_int8_t *start;
    u_int16_t op, pnum, flags = 0;
    struct unacked_req *unacked;
    char v[64];
    int error;

    /* check mars client has already registered with marsd */
    if (mcl.mcl_state != mcl_registered) {
	diag(COMPONENT, DIAG_WARN, 
	     "MARS client isn't in 'registered' state, no message sent ");
	return(0);
    }

    if (ip6m == NULL) {
	diag(COMPONENT, DIAG_ERROR, "join_serve_mgroup: NULL address");
	return(-1);
    }
    if (IN6_IS_ADDR_MULTICAST(ip6m) == 0) {
	diag(COMPONENT, DIAG_ERROR,  
	 "join_serve_mgroup: not a multicast address %s", 
	     inet_ntop(AF_INET6, ip6m, v, sizeof(v)));
	return(-1);
    }

    diag(COMPONENT, DIAG_DEBUG, 
	 "join_serve_mgroup: %s multicast group : %s", 
	 (opcode == mars_join) ? "joining" : "serving",
	 inet_ntop(AF_INET6, ip6m, v, 64));
    /* create the message to send */
    msg_len = ATMLLC_LEN + sizeof(struct mars_memb) + 
	3*IN6_ADDR_LEN + ATM_ESA_LEN;
    msg = (u_int8_t *) malloc(msg_len);
    bzero(msg, msg_len);
    bcopy(&llc_mars_ctl, msg, ATMLLC_LEN);
    /* Prepare at the same time a structure to keep track of the request */
    unacked = alloc_t(struct unacked_req);
    bzero(unacked, sizeof(*unacked));
    unacked->op = op = (u_int16_t) opcode;
    unacked->regist = 0;
    unacked->seq = 0;		/* disable this field for the moment */
    unacked->pnum = pnum = 1;
    bcopy(&mcl.svc_addr, &unacked->source_atm, sizeof(struct svc_addr));
    bcopy(ip6m, &unacked->pair.min, IN6_ADDR_LEN);
    bcopy(ip6m, &unacked->pair.max, IN6_ADDR_LEN);
    unacked->retries = MAX_RETRIES;
    flags = fl_layer3grp;
    memb_msg = (struct mars_memb *) (msg + ATMLLC_LEN);
    format_memb_msg(memb_msg, op, pnum, flags);
    start = ((u_int8_t *) memb_msg) + sizeof(struct mars_memb);
    set_info(&start, mcl.svc_addr.prv, ATM_ESA_LEN);
    set_info(&start, &mcl.ia6, IN6_ADDR_LEN);
    set_info(&start, ip6m, IN6_ADDR_LEN);	/* min */
    set_info(&start, ip6m, IN6_ADDR_LEN);	/* max = min */

    /* copy the message in unacked structure for potential retransmission */
    unacked->msg = (u_int8_t *) malloc(msg_len);
    bcopy(msg, unacked->msg, msg_len);
    unacked->msg_len = msg_len;

    if ((error = send_msg_to_mars(msg, msg_len, op)) !=0)
	{
	    diag(COMPONENT, DIAG_ERROR, 
		 "join_serve_mgroup: failed to send message to MARS");
	    /* try to retransmit immediately as if timer has just expired */
	    on_T_MEMBERSHIP(unacked);
	    return(error);
	}
    START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
    LIST_INSERT_HEAD(&Unacked, unacked, links);

    if (msg != NULL)
	free(msg);
    return(0);
}

/* 
 * this routine can be called either by a cluster member in order 
 * to leave a given min_max pair or by an MCS to unserve a given min_max pair
 */
int leave_unserve_minmax(struct in6_addr *min, 
			 struct in6_addr *max, 
			 enum mars_opcode opcode,
			 u_int16_t fl_layer3grp)
{
    u_int8_t *msg;
    struct mars_memb *memb_msg;
    u_int16_t msg_len;
    u_int8_t *start;
    u_int16_t op, pnum, flags = 0;
    struct unacked_req *unacked;
    char v[64], v1[64];
    int error;

    /* check mars client has already registered with marsd */
    if (mcl.mcl_state != mcl_registered) {
	diag(COMPONENT, DIAG_WARN, 
	     "MARS client isn't in 'registered' state, no message sent ");
	return(0);
    }

    if ((min == NULL) || (max == NULL)) {
	diag(COMPONENT, DIAG_ERROR, "leave_unserv_minmax: NULL address");
	return(NULL);
    }
    if (is_multicast_pair(min, max) == 0) {
	diag(COMPONENT, DIAG_ERROR,  
	     "leave_unserv_minmax: not a multicast min_max pair [%s, %s] ", 
	     inet_ntop(AF_INET6, min, v, sizeof(v)),
	     inet_ntop(AF_INET6, max, v1, sizeof(v1)));
	return(NULL);
    }
    diag(COMPONENT, DIAG_DEBUG, 
	 "leave_unserv_minmax: %s multicast pair : [%s, %s] ", 
	 (opcode == mars_leave) ? "leaving" : "unserving",
	 inet_ntop(AF_INET6, min, v, sizeof(v)),
	 inet_ntop(AF_INET6, max, v1, sizeof(v1)));
    /* create the message to send */
    msg_len = ATMLLC_LEN + sizeof(struct mars_memb) + 
	3*IN6_ADDR_LEN + ATM_ESA_LEN;
    msg = (u_int8_t *) malloc(msg_len);
    bzero(msg, msg_len);
    bcopy(&llc_mars_ctl, msg, ATMLLC_LEN);
    /* Prepare at the same time a structure to keep track of the request */
    unacked = alloc_t(struct unacked_req);
    bzero(unacked, sizeof(*unacked));
    unacked->op = op = (u_int16_t) opcode;
    unacked->regist = 0;
    unacked->seq = 0;		/* disable this field for the moment */
    unacked->pnum = pnum = 1;
    bcopy( &mcl.svc_addr, &unacked->source_atm, sizeof(struct svc_addr));
    bcopy(min, &unacked->pair.min, IN6_ADDR_LEN);
    bcopy(max, &unacked->pair.max, IN6_ADDR_LEN);
    unacked->retries = MAX_RETRIES;
    flags = fl_layer3grp;
    memb_msg = (struct mars_memb *) (msg + ATMLLC_LEN);
    format_memb_msg(memb_msg, op, pnum, flags);
    start = ((u_int8_t *) memb_msg) + sizeof(struct mars_memb);
    set_info(&start, mcl.svc_addr.prv, ATM_ESA_LEN);
    set_info(&start, &mcl.ia6, IN6_ADDR_LEN);
    set_info(&start, min, IN6_ADDR_LEN);	/* min */
    set_info(&start, max, IN6_ADDR_LEN);	/* max */

    /* copy the message in unacked structure for potential retransmission */
    unacked->msg = (u_int8_t *) malloc(msg_len);
    bcopy(msg, unacked->msg, msg_len);
    unacked->msg_len = msg_len;

    if ((error = send_msg_to_mars(msg, msg_len, op)) !=0)
	{
	    diag(COMPONENT, DIAG_ERROR, 
		 "leave_unserv_minmax: failed to send message to MARS");
	    /* try to retransmit immediately as if timer has just expired */
	    on_T_MEMBERSHIP(unacked);
	    return(error);
	}

    START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
    LIST_INSERT_HEAD(&Unacked, unacked, links);

    if (msg != NULL)
	free(msg);
    return(0);
}


int resolve_mgroup(struct resolv_entry **rslv, 
		   struct in6_addr *ip6m, 
		   u_int8_t rtm_op)
{
    u_int8_t *msg;
    struct mars_req *req_msg;
    u_int16_t msg_len;
    u_int8_t *start;
    struct resolv_entry *res_e;
    char v[64];
    int error;

    /* make sure mars client has already registered with marsd */
    if (mcl.mcl_state != mcl_registered) {
	diag(COMPONENT, DIAG_WARN, 
	     "MARS client isn't in 'registered' state, no message sent ");
	return(0);
    }

    diag(COMPONENT, DIAG_DEBUG, 
	 "resolve_mgroup: looking up the local table...");
    /* is it really a multicast address ? */
    if (ip6m == NULL) {
	diag(COMPONENT, DIAG_ERROR, "resolve_mgroup: NULL address");
	return(-1);
    }
    if (IN6_IS_ADDR_MULTICAST(ip6m) == 0) {
	diag(COMPONENT, DIAG_ERROR,  
	 "resolve_mgroup: not a multicast address [%s]", 
	     inet_ntop(AF_INET6, ip6m, v, sizeof(v)));
	return(-1);
    }

    res_e = lookup_ip6m(ip6m);
    if (res_e != NULL) {
	switch (res_e->rstate) {
	case rs_revalid:
	    /*
	     * even when a p2mp VC is under revalidation, its resolve entry 
	     * can still be used
	     */
	    diag(COMPONENT, DIAG_WARN, "Multicast VC is being revalidated");

	    /* fall through */

	case rs_valid:   
	    *rslv = res_e;
	    return(0);

	case rs_resolv:
	case rs_holddown:
	case rs_migr:
	    /* ignore resolve request */
	    diag(COMPONENT, DIAG_WARN, 
		     "mgroup [%s] is in %s state. No message sent.",
		 inet_ntop(AF_INET6, ip6m, v, sizeof(v)),
		 resolv_state_name[res_e->rstate]);
	    return(0);
	    
	case rs_none:		/* this is an invalid entry */
	     diag(COMPONENT, DIAG_WARN,
		  "a matching resolve entry exists and is in <%s> state",
		  resolv_state_name[res_e->rstate]);
	    res_e->conn_timer = NULL;
	    /* is there any attached multicast VC */
	    if (res_e->mvc != NULL) {
		if (res_e->mvc->state != connected) {
		    diag(COMPONENT, DIAG_WARN, 
			 "multicast VC is not connected!!!");
		    free_atmcall(res_e->mvc);
		    NbCalls--;
		} else {
		    /* revalidate entry */
		    res_e->rstate = rs_revalid;
		    res_e->mvc->revalidate = 1;
		    discard_ept_list(res_e);
		    error = retransmit_mars_req(res_e);
		    if (error != 0)
			diag(COMPONENT, DIAG_ERROR,
			     "resolve_mgroup: failed to retransmit REQUEST "
			     "message for mgroup [%s]", 
			     inet_ntop(AF_INET6, ip6m, v, sizeof(v)));
		    return(error);
		}
	    }
		
	    /* release resolve entry and create a new one later */
	    LIST_REMOVE(res_e, links);
	    free_resolv_entry(res_e);
	    break;	/* go and create a new resolve entry */
		
	default:	/* unlikely to happen */
	    diag(COMPONENT, DIAG_ERROR, 
		 "resolve_mgroup: invalid resolve state code %d", 
		 res_e->rstate);
	    return(-1);
	    
	}
    }

    /* here res_e is NULL */
    /* create a resolve entry for this new group */
    res_e = new_resolv_entry(ip6m, rs_resolv, NULL);
    res_e->rtm_op = rtm_op;
    /* insert the new res_e in the resolve table */
    LIST_INSERT_HEAD(&Resolv_Table, res_e, links);
   

   /* create and format the request to send */
    msg_len = ATMLLC_LEN + sizeof(struct mars_req) 
	+ 2*IN6_ADDR_LEN + ATM_ESA_LEN;
    msg = (u_int8_t *) malloc(msg_len);
    bzero(msg, msg_len);
    bcopy(&llc_mars_ctl, msg, ATMLLC_LEN);
    req_msg = (struct mars_req *) (msg + ATMLLC_LEN);
    format_req_msg(req_msg);
    start = ((u_int8_t *) req_msg) + sizeof(struct mars_req);
    set_info(&start, mcl.svc_addr.prv, ATM_ESA_LEN);
    set_info(&start, &mcl.ia6, IN6_ADDR_LEN);
    set_info(&start, ip6m, IN6_ADDR_LEN);
    
    /* send the request msg */
    if((error = send_msg_to_mars(msg, msg_len, mars_request)) != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "resolve_mgroup: failed to send message to MARS");
    }
    free(msg);
    return(error);
}


int resolve_minmax_mgroups(struct minmax_pair *pair) 
{
    struct in6_addr ip6m1, ip6m2;
    int cmp , error, equal;
    char v[64];

    if (pair == NULL) {
	diag(COMPONENT, DIAG_ERROR, 
	     "resolve_minmax_mgroup: minmax_pair is empty");
	return(-1);
    }

    
    bcopy(&pair->min, &ip6m1, sizeof(ip6m1));
    cmp = in6_addr_cmp(&pair->min, &pair->max);
    equal = (cmp == 0);
    while (cmp <= 0) {
	diag(COMPONENT, DIAG_DEBUG, 
	     "reslove_minmax_mgroups: resolving mgroup [%s]", 
	     inet_ntop(AF_INET6, &ip6m1, v, 64));
	error = resolve_mgroup(&Curr_Resolved, &ip6m1, RTM_ADD);
	if (error != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "reslove_minmax_mgroups: failed to resolve mgroup [%s]", 
		 inet_ntop(AF_INET6, &ip6m1, v, 64));
	    return(-1);
	}
	if (equal)
	    break;
	error = increment_addr(&ip6m1, &ip6m2);
	if (error != 0) {
	    diag(COMPONENT, DIAG_WARN,
		 "resolve_minmax_mgroups: increment failed");
	    return(-1);
	}
	bcopy(&ip6m2, &ip6m1, sizeof(ip6m1));
	cmp = in6_addr_cmp(&ip6m1, &pair->max);
    }
    return(0);
    
}

void retransmit_membership_req(struct unacked_req *unacked)
{
    u_int16_t msg_len;
    u_int8_t *msg;
    int error;

    msg = unacked->msg;
    msg_len = unacked->msg_len;
    if (--unacked->retries == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "No more possible retransmissions for JOIN/LEAVE "
	     "request number [%d]", 
	     unacked->seq);
	diag(COMPONENT, DIAG_ERROR, "MARS failure...reconnecting to MARS...");
	(void) reopen_mars();
	LIST_REMOVE(unacked, links);
	if (msg != NULL)
	    free(msg);
	free(unacked);
	return;
    }
    if ((error = send_msg_to_mars(msg, msg_len, unacked->op)) !=0)
	{
	    diag(COMPONENT, DIAG_ERROR, 
		 "retransmit_membership_req: failed to send message to MARS");
	    /* try to retransmit immediately as if timer has just expired */
	    on_T_MEMBERSHIP(unacked);
	    return;
	}
    START_TIMER(T_MEMBERSHIP, on_T_MEMBERSHIP, unacked);
    return;
}

int retransmit_mars_req(struct resolv_entry *res_e)
{
    u_int8_t *msg;
    struct mars_req *req_msg;
    int msg_len;
    u_int8_t *start;
    char v[64];
    int error;

    if (res_e->rstate == rs_revalid)
	free_ept_entries(&res_e->new_head);
    diag(COMPONENT, DIAG_DEBUG, 
	 "retransmit_mars_req: Retransmitting MARS REQUEST for "
	 "multicast group: %s",
	 inet_ntop(AF_INET6, &res_e->ia6, v, sizeof(v)) );
    
    /* create and format the request to send */
    msg_len = ATMLLC_LEN + sizeof(struct mars_req) + 
	2*IN6_ADDR_LEN + ATM_ESA_LEN;
    msg = (u_int8_t *) malloc(msg_len);
    bzero(msg, msg_len);
    bcopy(&llc_mars_ctl, msg, ATMLLC_LEN);
    req_msg = (struct mars_req *) (msg + ATMLLC_LEN);
    format_req_msg(req_msg);
    start = ((u_int8_t *) req_msg) + sizeof(struct mars_req);
    set_info(&start, mcl.svc_addr.prv, ATM_ESA_LEN);
    set_info(&start, &mcl.ia6, IN6_ADDR_LEN);
    set_info(&start, &res_e->ia6, IN6_ADDR_LEN);
    

    /* send the request msg */
    if ((error = send_msg_to_mars(msg, msg_len, mars_request)) != 0)
	diag(COMPONENT, DIAG_ERROR, 
	     "retransmit_mars_req: failed to send message to MARS");

    free(msg);
    res_e->seq = 0;
    return(error);
}


/* specific to CLMB */
int join_all_nodes(void)
{
    struct in6_addr ia6all;
    int error;

    error = inet_pton(AF_INET6, IP6_ALL_NODES_GRP, &ia6all);
    if (error < 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "join_all_nodes: inet_pton: %s\n", strerror(errno));
	return(-1);
    }
    if (error == 0) {
	/* paranoid, unlikely to happen */
	diag(COMPONENT, DIAG_ERROR, 
	     "join_all_nodes: unparsable multicast address");
	return(-1);
    }
    
    error = join_serve_mgroup(&ia6all, mars_join, FL_LAYER3GRP);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "failed to initiate joining all_nodes group");
	return(-1);
    }
    return(0);
}


/* 
 * Specific to CLMB 
 * This routine causes a newly registered CLMB to get all multicast 
 * addresses of mgroups the interface has already joined and to send a 
 * MARS_JOIN for each of them
 */
int join_former_mgroups(int itf)
{
    struct in6_addr_head *in6h = NULL;
    struct in6_addr_entry *in6e;
    int error;
    char v[64];
    
    error = get_former_mgroups(itf, &in6h);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "join_former_mgroups: failed to get formerly joined mgroups");
	return(-1);
    }
    
    for (in6e = TAILQ_FIRST(in6h); in6e; in6e = TAILQ_NEXT(in6e, links)) {
	error = join_serve_mgroup(&in6e->ia6, mars_join, FL_LAYER3GRP);
	if (error != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "join_former_mgroups: failed to initiate joining mgroup %s",
		 inet_ntop(AF_INET6, &in6e->ia6, v, sizeof(v)));
	    return(-1);
	}
    }
    free_in6_addr_entries(in6h);
    free(in6h);
    return(0);
}


/* 
 * called by CLMB to leave all joined mgroups and by MCS to 
 * unserve all supported mgroups
 */
int leave_unserve_all(enum mars_opcode opcode, u_int16_t fl_layer3grp)
{
    struct in6_addr min, max;
    int error, ret1, ret2;

    ret1 = inet_pton(AF_INET6, MIN_MULTI_ADDR, &min);
    ret2 = inet_pton(AF_INET6, MAX_MULTI_ADDR, &max);
    if ((ret1 < 0) || (ret2 < 0)) {
	diag(COMPONENT, DIAG_ERROR,
	     "leave_unserve_all: inet_pton: %s\n", strerror(errno));
	return(-1);
    }
    if ((ret1 == 0) || (ret2 == 0)) {
	/* paranoid, unlikely to happen */
	diag(COMPONENT, DIAG_ERROR, 
	     "leave_unserve_all: unparsable multicast min_max pair");
	return(-1);
    }
    
    error = leave_unserve_minmax(&min, &max, opcode, fl_layer3grp);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "leave_userve_all: %s failed", opcode);
	return(-1);
    }

    /* change mars client state */
    mcl.mcl_state = mcl_deregistering;
    return(0);
}

int signoff_cluster(void)
{
    ATMCALL *call;
    int error;
    enum mars_opcode opcode;

    opcode = (mcl.mcl_type == clmb_type) ? mars_leave : mars_unserv;

    /* 
     * leave (unserve) all joined (supported) multicast groups (if any)
     * otherwise send 'de-register' request 
     */
    if (TAILQ_FIRST(&Minmax_Table) == NULL) {
	error = deregister_mars_client(opcode);
	if (error != 0)
	    diag(COMPONENT, DIAG_ERROR,
		 "signoff_cluster: failed to send de-registration request");
    } else {
	error = leave_unserve_all(opcode, FL_LAYER3GRP);
	if (error != 0)
	    diag(COMPONENT, DIAG_ERROR,
		 "signoff_cluster: failed to %s all %s mgroups",
		 (opcode == mars_leave) ? "leave" : "unserve",
		 (opcode == mars_leave) ? "joined" : "supported");
    }

    /* close all open multicast VCs (if any) */
    for (call = AtmCalls; call; call = call->next) {
	if ((call->conf == root) && (call->res_e != NULL)) {
	    error = active_close_atmcall(0, call->vcc, call);
	    if (error != 0)
		diag(COMPONENT, DIAG_ERROR,
		     "signoff_cluster: failed to close multicast VC [vcc = %d]",
		     call->vcc);
	    /* resolv_entry will be automatically freed upon close confirm. */
	}
    }
    /* force changing mclient state to 'deregsitering' if not 'deregsitered' */
    if (mcl.mcl_state != mcl_none)
	mcl.mcl_state = mcl_deregistering;

    return(0);
}
