/* mars_msg.c - MARS messages strucure and basic operations */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/param.h>
#include <sys/socket.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <net/if.h>
#include <net/if_atm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include "atmd.h"
#include "ipv6_tools.h"

#include "../linux-atm.h"

#include "mars_def.h"
#include "mars_msg.h"

#define COMPONENT "MARS_MSG"

struct atmllc llc_mars_ctl = LLC_MARS_CTL;

/* MARS operations (textual format) */
const char *mop_name[] = MOP_NAME;

void format_fixed_hdr(struct fixed_hdr *fxh, u_int16_t op)
{
    fxh->afn = htons(AFN);
    fxh->pro = htons(ETHERTYPE_IPV6);
    fxh->pro_snap1 = 0;
    fxh->pro_snap2 = 0;
    fxh->hdrrsv[2] = fxh->hdrrsv[1] = fxh->hdrrsv[0] = 0;
    fxh->extoff = fxh->chksum = 0;
    fxh->op = htons(op);
    fxh->shtl = ATM_ESA_LEN;
    fxh->sstl = 0;
}

int check_fxh(u_int8_t *mars_buf, int mars_len)
{
    struct fixed_hdr *fxh;

    fxh = (struct fixed_hdr *) mars_buf;
    
    /* check the Address Family (0x000f) */
    if (ntohs(fxh->afn) != AFN) {
	diag(COMPONENT, DIAG_ERROR, 
	     "pre_process_membership: unknown address family 0x%04x",
             ntohs(fxh->afn));
	return(-1);
    }
    /* check the Ether Type (0x86dd) */
    if (ntohs(fxh->pro) != ETHERTYPE_IPV6) {
	diag(COMPONENT, DIAG_ERROR, 
	     "pre_process_membership: unknown upper protocol 0x%04x",
             ntohs(fxh->pro));
	return(-1);
    }
    
    /* @@@ check the address format, see later */
    if (fxh->shtl & TL_164) {
	diag(COMPONENT, DIAG_ERROR, 
	     "pre_process_membership: can't handle E.164 format yet");
	return(-1);
    }
    
    /* @@@ disable sub-address field temporarily */
    if (fxh->sstl & TL_LEN) {
	diag(COMPONENT, DIAG_ERROR, 
	     "pre_process_membership: can't handle sub-address field yet");
	return(-1);
    }

    /* compute the checksum accross entire MARS message */

    /* @@@ CHECKSUM see later */

    return(0);
}


void format_memb_msg(struct mars_memb *memb_msg, 
		     u_int16_t op, u_int16_t  pnum, u_int16_t  flags)
{
    struct fixed_hdr *fxh;
    
    bzero(memb_msg, sizeof(struct mars_msg));
    fxh = (struct fixed_hdr *) memb_msg;
    format_fixed_hdr(fxh, op);
    memb_msg->spln = IN6_ADDR_LEN;
    if (flags & FL_REGISTER)
	memb_msg->tpln = 0;
    else
	memb_msg->tpln = IN6_ADDR_LEN;
    memb_msg->pnum = htons(pnum);
    memb_msg->flags = htons(flags);
    memb_msg->cmi =0;
    memb_msg->msn =0;
}

void format_req_msg(struct mars_req *req_msg)
{
    struct fixed_hdr *fxh;
    
    bzero(req_msg, sizeof(struct mars_msg));
    fxh = (struct fixed_hdr *) req_msg;
    format_fixed_hdr(fxh, (u_int16_t) mars_request);
    req_msg->spln = IN6_ADDR_LEN;
    req_msg->thtl = req_msg->tstl = 0;
    req_msg->tpln = IN6_ADDR_LEN;
}

void format_multi_msg(struct mars_req *req_msg,
                      struct mars_multi *multi_msg,
		      u_int16_t tnum,
                      u_int16_t seqxy,
		      u_int32_t msn)
{
    struct fixed_hdr *fxh;
    
    bzero(multi_msg, sizeof(struct mars_msg));
    fxh = (struct fixed_hdr *) multi_msg;
    format_fixed_hdr(fxh, (u_int16_t) mars_multi);
    multi_msg->spln = req_msg->spln;	/* IN6_ADDR_LEN */
    multi_msg->thtl = ATM_ESA_LEN;
    multi_msg->tstl = 0;
    multi_msg->tpln = req_msg->spln;	/* IN6_ADDR_LEN */
    multi_msg->tnum = htons(tnum);
    multi_msg->seqxy = htons(seqxy);
    multi_msg->msn = htonl(msn);
}

void format_migr_msg(struct mars_migr *migr_msg,
		     u_int16_t tnum,
		     u_int32_t msn)
{
    struct fixed_hdr *fxh;
    
    bzero(migr_msg, sizeof(struct mars_msg));
    fxh = (struct fixed_hdr *) migr_msg;
    format_fixed_hdr(fxh, (u_int16_t) mars_migrate);
    migr_msg->spln = IN6_ADDR_LEN;	/* IN6_ADDR_LEN */
    migr_msg->thtl = ATM_ESA_LEN;
    migr_msg->tstl = 0;
    migr_msg->tpln = IN6_ADDR_LEN;	/* IN6_ADDR_LEN */
    migr_msg->tnum = htons(tnum);
    migr_msg->msn = htonl(msn);
}

void format_redir_msg(struct mars_redir *redir_msg,
		      u_int8_t redirf,
		      u_int16_t tnum,
                      u_int16_t seqxy,
		      u_int32_t msn)
{
    struct fixed_hdr *fxh;
    
    bzero(redir_msg, sizeof(struct mars_msg));
    fxh = (struct fixed_hdr *) redir_msg;
    format_fixed_hdr(fxh, (u_int16_t) mars_redirect_map);
    redir_msg->thtl = ATM_ESA_LEN;
    redir_msg->redirf = redirf;
    redir_msg->tnum = htons(tnum);
    redir_msg->seqxy = htons(seqxy);
    redir_msg->msn = htonl(msn);
}

/* fill in message field */
void set_info(u_int8_t **start, void *info, int len)
{
    if (!len)
        return;
    bcopy(info, *start, len);
    (*start) += len;
    return;
}

/* get message field */
void *get_info(u_int8_t **start, int len)
{
    if (!len)
        return NULL;
    (*start) += len;
    return (*start - len);
}

