/* mars_msg.h - MARS messages strucure and basic operations */

/* Written 09-1998 by Mohsen.Souissi@inria.fr, INRIA */

#ifndef _MMSG_H
#define _MMSG_H

#define AFN 		0x000f

#define TL_LEN		0x3f	/* to compute length */
#define TL_164		0x40	/* to detect E.164 Address format */
#define SEQ_Y		0x7fff	/* mask to compute sequence number */
#define LAST_SEQ	0x8000	/* to check the last sequence in MARS_MULTI */
#define HARD_REDIRECT	0x80	/* to check if 'hard' redirect */

/* flags in JOIN/LEAVE msgs */
#define FL_LAYER3GRP	0x8000
#define FL_COPY		0x4000
#define FL_REGISTER	0x2000
#define FL_PUNCHED	0x1000
#define FL_SEQUENCE	0xff	/* it's also the mask */

enum mars_opcode {
    mars_request = 1,
    mars_multi = 2,
    mars_mserv   = 3,
    mars_join    = 4,
    mars_leave   = 5,
    mars_nak     = 6,
    mars_unserv  = 7,
    mars_sjoin   = 8,
    mars_sleave  = 9,
    mars_grouplist_request = 10,
    mars_grouplist_reply   = 11,
    mars_redirect_map      = 12,
    mars_migrate = 13
};



struct fixed_hdr {
    u_int16_t  afn;        /* Address Family (0x000F) */
    u_int16_t  pro;        /* Protocol identification (type=0x86DD for IPv6) */
    u_int32_t  pro_snap1;  /* SNAP zeroed when pro.type != 0x80 */
    u_int8_t   pro_snap2;
    u_int8_t   hdrrsv[3];  /* Reserved. Unused by MARS control protocol */
    u_int16_t  chksum;     /* Checksum across entire MARS message */
    u_int16_t  extoff;     /* Extensions offset: for supplementary list */ 
    u_int16_t  op;         /* Operation code (version and type) */ 
    u_int8_t   shtl;       /* Type/Length of source ATM number */
    u_int8_t   sstl;       /* Type/Length of source ATM sub-address */
};

/*
 * MARS message structure for MARS_REQUEST and MARS_NAK
 */

struct mars_req {
    struct fixed_hdr fxh;  /* Fixed header */

    u_int8_t   spln;	/* Length of source protocol address (16) (s) */
    u_int8_t   thtl;	/* Type/Length of target ATM number  (x), (zero) */
    u_int8_t   tstl;	/* Type/Length of target ATM sub-address (y) (zero) */
    u_int8_t   tpln;	/* Length of target group address (16) (z) */
    u_int32_t  pad1;	/* Padding to align with MARS_MULTI */
    u_int32_t  pad2;	/* Second word of padding */
    
    /* sha: source ATM number (r octets) */
    /* ssa: source ATM sub-address (q octets) */
    /* spa  source protocol address (s octets)*/
    
    /* tpa: target protocol address (z octets) */
    /* tha: target ATM number (x octets) */
    /* tsa: target ATM sub-address (y octets) */
};

/*
 * MARS message structure for MARS_MULTI
 */

struct mars_multi {
    struct fixed_hdr fxh;  /* Fixed header */

    u_int8_t   spln;	/* Length of source protocol address (16) (s) */
    u_int8_t   thtl;	/* Type/Length of target ATM number  (x) */
    u_int8_t   tstl;	/* Type/Length of target ATM sub-address (y) */
    u_int8_t   tpln;	/* Length of target group address (16) (z) */
    u_int16_t  tnum;	/* No. of target ATM addresses returned (N) */
    u_int16_t  seqxy;	/* Boolean flag x and sequence number y */
    u_int32_t  msn;	/* MARS sequence number */
    
    /* sha: source ATM number (r octets) */
    /* ssa: source ATM sub-address (q octets) */
    /* spa  source protocol address (s octets)*/
    
    /* tpa: target multicast group address (z octets) */

    /* tha.1: target ATM number 1 (x octets) */
    /* tsa.1: target ATM sub-address 1 (y octets) */
    /* ....*/
    /* tha.N: target ATM number N (x octets) */
    /* tsa.N: target ATM sub-address N (y octets) */
};

/*
 * MARS message structure MARS_REDIRECT_MAP
 * based on MARS_MULTI with mar$redirf instead of mar$tpln and mar$spln
 */

struct mars_redir {
    struct fixed_hdr fxh;  /* Fixed header */

    u_int8_t   spln;	/* Length of source protocol address (16) RESERVED */
    u_int8_t   thtl;	/* Type/Length of target ATM number  (x) */
    u_int8_t   tstl;	/* Type/Length of target ATM sub-address (y) */
    u_int8_t   redirf;	/* Flag controling client redirect behaviour */
    u_int16_t  tnum;	/* No. of target ATM addresses returned (N) */
    u_int16_t  seqxy;	/* Boolean flag x and sequence number y */
    u_int32_t  msn; 	/* MARS sequence number */
    
    /* sha: source ATM number (r octets) */
    /* ssa: source ATM sub-address (q octets) */
    /* spa  source protocol address (s octets)*/
    
    /* tpa: target multicast group address (z octets) */
    
    /* tha.1: target ATM number 1 (x octets) */
    /* tsa.1: target ATM sub-address 1 (y octets) */
    /* ....*/
    /* tha.N: target ATM number N (x octets) */
    /* tsa.N: target ATM sub-address N (y octets) */
};

/*
 * MARS message structure for MARS_MIGRATE
 * This format is similar to MARS_MULTI except for seqxy/resv fields
 */

struct mars_migr {
    struct fixed_hdr fxh;  /* Fixed header */

    u_int8_t   spln;	/* Length of source protocol address (16) (s) */
    u_int8_t   thtl;	/* Type/Length of target ATM number  (x) */
    u_int8_t   tstl;	/* Type/Length of target ATM sub-address (y) */
    u_int8_t   tpln;	/* Length of target group address (16) (z) */
    u_int16_t  tnum;	/* No. of target ATM addresses returned (N) */
    u_int16_t  resv;	/* Reserved */
    u_int32_t  msn;	/* MARS sequence number */
  
    /* sha: source ATM number (r octets) */
    /* ssa: source ATM sub-address (q octets) */
    /* spa  source protocol address (s octets)*/

    /* tpa: target multicast group address (z octets) */

    /* tha.1: target ATM number 1 (x octets) */
    /* tsa.1: target ATM sub-address 1 (y octets) */
    /* ....*/
    /* tha.N: target ATM number N (x octets) */
    /* tsa.N: target ATM sub-address N (y octets) */
};

/*
 * MARS message structure for MARS_JOIN/ LEAVE/ SJOIN/ SLEAVE/ MSERV/ UNSERV
 *                            GROUPLIST_REQ/ GROUPLIST_REPLY
 */

struct mars_memb {
    struct fixed_hdr fxh;  /* Fixed header */

    u_int8_t   spln;	/* Length of source protocol address (16) (s) */
    u_int8_t   tpln;	/* Length of group address (16) (z) */
    u_int16_t  pnum;	/* No. of group address pairs (N) */
    u_int16_t  flags;	/* layer3grp, copy, and register flags */
    u_int16_t  cmi;	/* Cluster member ID */
    u_int32_t  msn;	/* MARS sequence number */

  
    /* sha: source ATM number (r octets) */
    /* ssa: source ATM sub-address (q octets) */
    /* spa  source protocol address (s octets)*/

    /* minimum multicast group address pair 1 (z octets) */
    /* Maximum multicast group address pair 1 (z octets) */
    /* ....*/
    /* minimum multicast group address pair N (z octets) */
    /* Maximum multicast group address pair N (z octets) */
};

struct mars_msg {
    union {
	struct mars_req req;
	struct mars_multi multi;
	struct mars_redir redir;
	struct mars_migr migr;
	struct mars_memb memb;

    } u_msg;
};

extern struct atmllc llc_mars_ctl;
extern const char *mop_name[];

void format_fixed_hdr(struct fixed_hdr *fxh, u_int16_t op);
int check_fxh(u_int8_t *mars_buf, int mars_len);

void format_memb_msg(struct mars_memb *memb_msg, 
		     u_int16_t op, u_int16_t  pnum, u_int16_t  flags);

void format_req_msg(struct mars_req *req_msg);

void format_multi_msg(struct mars_req *req_msg,
                      struct mars_multi *multi_msg,
		      u_int16_t tnum,
                      u_int16_t seqxy,
		      u_int32_t msn);

void format_migr_msg(struct mars_migr *migr_msg,
		      u_int16_t tnum,
		      u_int32_t msn);

void format_redir_msg(struct mars_redir *redir_msg,
		      u_int8_t redirf,
		      u_int16_t tnum,
                      u_int16_t seqxy,
		      u_int32_t msn);

void set_info(u_int8_t **start, void *info, int len);
void *get_info(u_int8_t **start, int len);


#endif

