/* tables.c - MARSD/CLMB/MCS common table management */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>

#include <net/if.h>
#ifndef if_type
#include <net/if_var.h> 
#endif

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "../linux-atm.h"

#include "atmd.h"
#include "ipv6_tools.h"

#include "mars_def.h"
#include "mars_msg.h"
#include "tables.h"

#define COMPONENT "COMMON_TAB"

/* MARS client type (textual format) */
const char *mcl_name[] = MCL_NAME;
const char *mcl_state_name[] = MCL_STATE_NAME;

struct svc_addr_entry *new_addr_entry(u_int8_t *tha)
{
    struct svc_addr_entry *new;
    
    new = alloc_t(struct svc_addr_entry);
    bzero(new, sizeof(struct svc_addr_entry));
    bcopy(tha, &new->addr.prv, ATM_ESA_LEN);
    return(new);
}

void free_svc_addr_entries(struct svc_addr_head *head)
{
    struct svc_addr_entry *walk, *next;
    
    for (walk = LIST_FIRST(head); walk; walk = next) {
	next = LIST_NEXT(walk, links);
	LIST_REMOVE(walk, links);
	free(walk);
    }

}

void free_in6_addr_entries(struct in6_addr_head *in6h)
{
    struct in6_addr_entry *walk, *next;
    
    for (walk = TAILQ_FIRST(in6h); walk; walk = next) {
        next = TAILQ_NEXT(walk, links);
        TAILQ_REMOVE(in6h, walk, links);
        free(walk);
    }
}

void free_minmax_entries(struct minmax_head *mmh)
{
    struct minmax_entry *walk, *next;
    
    for (walk = TAILQ_FIRST(mmh); walk; walk = next) {
        next = TAILQ_NEXT(walk, links);
        TAILQ_REMOVE(mmh, walk, links);
        free(walk);
    }
}

/*  
 * update the client's minmax_list when a single group is being 
 * joined (min = max)
 */
void insert_new_mgroup(struct minmax_head *mmh, 
		       struct in6_addr *ip6m,
		       u_int16_t fl_layer3grp)
{
    struct minmax_entry *walk, *first = NULL, *new_minmax_e;
    int cmp = -1;
    
    /* determine position of the new address */
    first = TAILQ_FIRST(mmh);
    for(walk = first; walk; walk = TAILQ_NEXT(walk, links))
	if ((cmp = in6_addr_cmp(ip6m, &walk->pair.min)) <= 0)
	    break;
    
    if (cmp == 0)	/* we've found a matching entry */
	return;

    /* create a new minmax entry for the new multicast address */
    new_minmax_e = alloc_t(struct minmax_entry);
    bcopy(ip6m, &new_minmax_e->pair.min, sizeof(struct in6_addr));
    bcopy(ip6m, &new_minmax_e->pair.max, sizeof(struct in6_addr));
    new_minmax_e->fl_layer3grp = fl_layer3grp;
   
    /* insert the new entry at its place */
    if (first == NULL) {
	TAILQ_INSERT_HEAD(mmh, new_minmax_e, links);
	return;
    }

    if (walk != NULL)
	TAILQ_INSERT_BEFORE(walk, new_minmax_e, links);
    else
	TAILQ_INSERT_TAIL(mmh, new_minmax_e, links);
    
    return;
}

/*  
 * update the client's minmax_list when a single group is being 
 * left/unserved (min = max)
 */
int remove_mgroup(struct minmax_head *mmh, 
		  struct in6_addr *ip6m)
{
    struct minmax_entry *walk, *new_minmax_e;
    int cmp_max, cmp_min = -1;
    struct in6_addr ia6_min, ia6_max;
    char v[64];
    int error;

    /* find the pair that encompasses the address */
    for(walk = TAILQ_FIRST(mmh); walk; walk = TAILQ_NEXT(walk, links))
	if ((cmp_max = in6_addr_cmp(ip6m, &walk->pair.max)) <= 0) {
	    cmp_min = in6_addr_cmp(ip6m, &walk->pair.min);
	    break;
	}
    
    if (walk == NULL) {
	diag(COMPONENT, DIAG_WARN,
	     "single_update_minmax_leave: no such a group [%s] minmax list",
	     inet_ntop(AF_INET6, ip6m, v, 64));
	return(0);
    }
    
    if ((cmp_max == 0) || (cmp_min == 0)) {
	if ((cmp_min == 0) && (cmp_max == 0)) {
	    /* this is the most likely case */
	    TAILQ_REMOVE(mmh, walk, links);
	    free(walk);
	    return(0);
	}
	if (cmp_min == 0) {
	    error = increment_addr(ip6m, &ia6_min);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "single_update_minmax_leave: increment failed");
		return(-1);
	    }
	    /* shift min boundary in `walk' entry */
	    bcopy(&ia6_min, &walk->pair.min, sizeof(ia6_min));
	    return(0);
	}
	if (cmp_max == 0) {
	    error = decrement_addr(ip6m, &ia6_max);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "single_update_minmax_leave: decrement failed");
		return(-1);
	    }
	    /* shift min boundary in `walk' entry */
	    bcopy(&ia6_max, &walk->pair.max, sizeof(ia6_max));
	    return(0);
	}
    }
	
    /* split `walk' entry */
    error = increment_addr(ip6m, &ia6_min);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "single_update_minmax_leave: increment failed");
	return(-1);
    }
    error = decrement_addr(ip6m, &ia6_max);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "single_update_minmax_leave: decrement failed");
	return(-1);
    }
    
    /* create a new minmax entry and insert it */
    new_minmax_e = alloc_t(struct minmax_entry);
    bcopy(&ia6_min, &new_minmax_e->pair.min, sizeof(ia6_min));
    bcopy(&walk->pair.max, &new_minmax_e->pair.max, sizeof(ia6_min));
    new_minmax_e->fl_layer3grp = walk->fl_layer3grp;
    bcopy(&ia6_max, &walk->pair.max, sizeof(ia6_min));
    TAILQ_INSERT_AFTER(mmh, walk, new_minmax_e, links);
    return(0);
}
