/* atmcall.c - atm call management */
/* Written 1998-1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#ifdef NDP_CLIENT
#include "ndpd.h"
#endif

#ifdef MARS_CLIENT
#include "tables.h"
#include "mcl_tab.h"
#include "mars_msg.h"
#include "mcl_io.h"
#endif

#ifdef MARSD
#include "tables.h"
#include "mars_msg.h"
#include "marsd_tab.h"
#include "marsd_io.h"
#endif

#define COMPONENT "ATMCALL"

ATMCALL *AtmCalls = NULL;
int NbCalls;
u_long CurrVcc;
const char *CallStateName[] = { "null", "connecting", "accepting",
				"connected", "disconnecting" };
const char *PartyStateName[] = { "ps_null", "ps_add_init", "ps_add_recv", 
				 "ps_active", "ps_drop_init", "ps_dro_recv" };

/* for atm qos_default_p2[pm] */
struct atm_qos qos_default_p2p = {{ATM_UBR, 0, 0, 0, 0, DEFAULT_MAX_SDU},
				  {ATM_UBR, 0, 0, 0, 0, DEFAULT_MAX_SDU},
				  ATM_AAL5};

struct atm_qos qos_default_p2m = {{ATM_UBR, 0, 0, 0, 0, DEFAULT_MAX_SDU},
				  {ATM_NONE, 0, 0, 0, 0, 0},
				  ATM_AAL5};

ATMCALL *new_atmcall(u_long vcc, u_int8_t vpi, u_int16_t vci, 
		     CONN_CONF conf, u_int8_t owner, ATMCALL_STATE state, 
		     struct svc_addr *remote_atm, char *remote_name, 
		     struct in6_addr *remote_in6, struct atm_qos *qos,
		     void *clnt_e)
{
    ATMCALL *call;
  
    call = alloc_t(ATMCALL);
    if (call == NULL) {
	diag(COMPONENT, DIAG_ERROR, 
	     "new_atmcall: failed to allocate a new atmcall structure");
	return(NULL);
    }

    bzero(call, sizeof(ATMCALL));
    call->vcc = vcc;
    call->vpi = vpi;
    call->vci = vci;
    call->conf = conf;
    call->owner = owner;
    call->state = state;
    call->sd = -1;
    if (remote_atm != NULL)
	bcopy(remote_atm, &call->rsvcaddr, sizeof(struct svc_addr));
    if (remote_name != NULL)
	strcpy(call->rname, remote_name);
    if (remote_in6 != NULL)
	bcopy(remote_in6, &call->rin6, sizeof(struct in6_addr));
    /* setting QoS */
    if (qos != NULL)
	bcopy(qos, &call->qos, sizeof(struct atm_qos));
    else {
	switch (conf) {
	case p2p:
	    bcopy(&qos_default_p2p, &call->qos, sizeof(struct atm_qos));
	    break;
	case root:
	    bcopy(&qos_default_p2m, &call->qos,  sizeof(struct atm_qos));
	    break;
	default: /* should never happen */
	    diag(COMPONENT, DIAG_WARN, "new_atmcall: Leaf QoS misconfigured");
	}
    }

#ifdef NDP_CLIENT
    /*
     * if we are in NDP CLIENT link the call to the neighbor entry
     */
    if (clnt_e != NULL) {
	call->ndp = (struct neighbor *) clnt_e;
	call->ndp->circuit = call;
    }
#endif

#ifdef MARS_CLIENT
    /*
     * if we are in MARS CLIENT and in root config,
     * link the call with the corresponding resolve entry if non null
     */
    if ((conf == root) && (clnt_e != NULL)) {
	call->res_e = (struct resolv_entry *) clnt_e;
	call->res_e->mvc = call;
    }
#endif

    return(call);
}

PARTY *new_party(ATMCALL *call, u_int16_t ep_ref, PARTY_STATE pstate, 
		 struct svc_addr *remote_atm, char *remote_name, 
		 void *ept_e, void *mcl_e)
{
    PARTY *party;
    int copied = 0;
     
    party = alloc_t(PARTY);
    if (party == NULL) {
	diag(COMPONENT, DIAG_ERROR, 
	     "new_party: failed to allocate a new party structure");
	return(NULL);
    }

    bzero(party, sizeof(PARTY));
    party->ep_ref = ep_ref;
    party->pstate = pstate;

#ifdef MARS_CLIENT
    /* if party structure is built from an ept_entry (e.g. MARS client...) */
    if (ept_e != NULL) {
        struct ept_entry *epe = (struct ept_entry *) ept_e;
	bcopy(&epe->addr, &party->rsvcaddr, sizeof(struct svc_addr));
	party->ept_e = epe;
	copied = 1;
    }
#endif

#ifdef MARSD
    /* if MARS daemon is creating the party */
    if (mcl_e != NULL) {
	struct mcl_entry *clmbe = (struct mcl_entry *) mcl_e;
	
	bcopy(&clmbe->svc_addr, &party->rsvcaddr, sizeof(struct svc_addr));
	party->mcl_e = clmbe;
	clmbe->party = party;
	copied = 1;
    }
#endif

    if ((copied == 0) && (remote_atm != NULL))
	bcopy(remote_atm, &party->rsvcaddr, sizeof(struct svc_addr));
    if (remote_name != NULL)
	strcpy(party->rname, remote_name);
    party->call = call;
    party->next = call->parties;
    party->prev = NULL;
    if (call->parties != NULL)
	(call->parties)->prev = party;
    call->parties = party;

    return (party);
}

void append_atmcall(ATMCALL *call)
{

    if (AtmCalls != NULL) {
	AtmCalls->prev = call;
	call->next = AtmCalls;
    }
    AtmCalls = call;
} 

void show_atmcalls(void)
{
    ATMCALL *call;
    PARTY *party;
    char rname[20];
    char *idx;
    int len = 19, i;

    printf("\n\t\t\t\tCURRENT ATM CALLS\n\n");
    printf("\nCnf  Own   VCC   VPI   VCI   Rmt                EP_REF  C_State"
	   "   EP_State  Rt\n\n");
    for (call = AtmCalls; call != NULL; call = call->next) {
	if ((idx = index(call->rname, '.')) != NULL) {
	    len = idx - call->rname;
	    if (len > 19)
		len = 19;
	}
	strncpy(rname, call->rname, len);
	rname[19] = '\0';
	switch (call->conf) {
	case p2p:
	    printf("p2p  %-6s%-6ld%-6d%-6d%-19.19s        %-10s          %-4s"
		   "\n", call->owner ? "Y" : "N",
		   call->vcc, call->vpi, call->vci, rname, 
		   CallStateName[call->state],
#ifndef NDP_CLIENT
		   call->route ? "Y" : "N"
#else
		   call->ndp ? "Y" : "N"
#endif
		   );
	    if (call->rsvcaddr.prv[0] != 0) {
		    printf("                             To:    ");
		    for (i = 0; i < ATM_ESA_LEN; i++)
			printf("%02x", call->rsvcaddr.prv[i]);
		    printf("\n");
		}
	    printf("\n");
	    break;
	case root:
	    printf("root %-6s%-6ld%-6d%-6d****************           "
		   "%-10s          %-7s\n", call->owner ? "Y" : "N",
		   call->vcc, call->vpi, call->vci,
		   CallStateName[call->state],
#ifndef NDP_CLIENT
		   call->route ? "Y" : "N"
#else
		   call->ndp ? "Y" : "N"
#endif
		   );
	    for (party = call->parties; party != NULL; party = party->next) {
		if (party->rsvcaddr.prv[0] != 0) {
		    printf("\t");
		    for (i = 0; i < ATM_ESA_LEN; i++)
			printf("%02x", party->rsvcaddr.prv[i]);
		}
		printf("  %-8d        %-10s\n", party->ep_ref, 
		       PartyStateName[party->pstate]);
	    }
	    printf("\n");
	    break;
	case leaf:
	    party = call->parties;
	    printf("leaf       %-6ld%-6d%-6d%-19s%-8d%-10s%-10s\n",
		   call->vcc, call->vpi, call->vci, party->rname, 
		   party->ep_ref, CallStateName[call->state], 
		   PartyStateName[party->pstate]);
	    if (party->rsvcaddr.prv[0] != 0) {
		printf("                             From:  ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02x", party->rsvcaddr.prv[i]);
		printf("\n");
	    }
	    printf("\n");
	}
    }
}


ATMCALL *atmcall_lookup(u_long vcc)
{
    ATMCALL *call;

    for (call = AtmCalls; call; call = call->next) {
	if (call->vcc == vcc)
	    return(call);
    }
    return(NULL);
}

/* another lookup function. Search by socket descriptor. */
ATMCALL *atmcall_lookup2(int sd)
{
    ATMCALL *call;

    for (call = AtmCalls; call; call = call->next) {
	if (call->sd == sd)
	    return(call);
    }
    return(NULL);
}

#ifdef NDP_CLIENT
/* yet another lookup function. Search unbound by svc_addr. */
ATMCALL *atmcall_lookupa(struct svc_addr *addr)
{
    ATMCALL *call;

    for (call = AtmCalls; call; call = call->next) {
	if ((call->ndp == NULL) &&
	    (call->qos.txtp.traffic_class == ATM_UBR) &&
	    atm_equal(addr, &call->rsvcaddr, 0, 0))
	    return(call);
    }
    return(NULL);
}
#endif

PARTY *party_lookup(u_long vcc, u_int16_t ep_ref)
{
    ATMCALL *call;
    PARTY *party;

    if ((call = atmcall_lookup(vcc)) == NULL) {
	return(NULL);
    }
    for (party = call->parties; party != NULL; party = party->next)
	if (party->ep_ref == ep_ref)
	    break;
    return(party);
}


void free_party(PARTY *party)
{
    ATMCALL *call;
    PARTY  *hold;

    if (party == NULL)
	return;

    if ((call = party->call) == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "free_party: party <EPR = %d> is not attached to any ATM call", 
	     party->ep_ref);
	return;
    }
    diag(COMPONENT, DIAG_DEBUG, 
	 "Freeing party [Vcc = %ld, EPR = %d]", 
	 party->call->vcc, party->ep_ref);
    call->nbactp--;

#ifdef MARS_CLIENT
    /* cut link with ept_entry if any */
    if (party->ept_e != NULL)
	party->ept_e->party = NULL;
#endif

#ifdef MARSD
    /* cut link with mcl_entry if any */
    if (party->mcl_e != NULL)
	party->mcl_e->party = NULL;
#endif

    if (party == call->parties) {
	if (party->next == NULL) { /* we are the only party */
	    call->parties = NULL;
	    free(party);
	    return;
	} /* else */
	hold = party->next;
	call->parties = hold;
	hold->prev = NULL;
	free(party);
	return;
    }
    hold = party->prev;
    hold->next = party->next;
    if (party->next != NULL)
	(party->next)->prev = hold;
    free(party);
    return;
}

void free_all_parties(ATMCALL *call)
{
    PARTY  *walk, *next;

    diag(COMPONENT, DIAG_DEBUG, 
	 "Freeing all parties of ATM call <vcc = %ld>", call->vcc);
    for (walk = call->parties; walk != NULL; walk = next) {
	next = walk->next;

#ifdef MARS_CLIENT
	/* disconnect ept_entry if any */
	if (walk->ept_e != NULL)
	    walk->ept_e->party = NULL;
#endif

#ifdef MARSD
	/* cut link with mcl_entry if any */
	if (walk->mcl_e != NULL)
	    walk->mcl_e->party = NULL;
#endif

	free(walk);
    }
    call->nbactp = 0;
    call->parties = NULL;
}


void free_atmcall(ATMCALL *call)
{
    if ((call->prev == NULL) &&
	(call->next == NULL) &&
	(call != AtmCalls)) {	/* not inserted yet */
	goto free_c;
    }
    if (call == AtmCalls)
	AtmCalls = call->next;
    else {
	(call->prev)->next = call->next;
	if (call->next != NULL)
	    (call->next)->prev = call->prev;
    }
    if (call->parties != NULL)
	free_all_parties(call);

#ifdef NDP_CLIENT
    /* notify ndp daemon if any neighbor entry is using this */
    if (call->ndp != NULL)
	closing_circuit(call, 1);
#endif

#ifdef MARS_CLIENT
    /* cut the link between the call and the resolv_entry structure */
    /* 
     * see later: should we remove the resolv_entry at the same time ? 
     *                   or keep it in the cache ? 
     */
    if (call->res_e != NULL)
	call->res_e->mvc = NULL;
#endif

#ifdef MARSD
    /* cut the link between the call and the cluster member entry if any */
    if (call->mcl_e != NULL) {
	call->mcl_e->p2p = NULL;
    }

    /* if it is CCVC/SCVC, reset it */
    if (call == Ccvc)
	Ccvc = NULL;
    
    if (call == Scvc)
	Scvc = NULL;
#endif

 free_c:
    free(call);
}

int atmcall_setup(struct svc_addr *remote_atm, char *remote_name, 
		  struct in6_addr *remote_in6, struct atm_qos *qos,
		  int16_t ep_ref, void *ept_e, void *mcl_e)
{
    int i, ret;
    ATMCALL *call;
    int sd;
    CONN_CONF conf;
    PARTY *party = NULL;
#ifdef MARS_CLIENT
    struct ept_entry *epe;
#endif
    
    
    if (ep_ref > 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "atmcall_setup: %d : bad end_point reference value", ep_ref);
	return(-1);
    }
    conf = (ep_ref < 0) ? p2p : root;
    printf("\nREMOTE ADDR = ");
    for (i = 0; i < ATM_ESA_LEN; i++)
	printf("%02X", remote_atm->prv[i]);
    printf("\n");
    if (++CurrVcc == 0xffffff) 
	CurrVcc = 1;
#ifdef MARS_CLIENT
    if (ept_e != NULL) {
	epe = (struct ept_entry *) ept_e;
        call = new_atmcall(CurrVcc, 0, 0, conf, 1, connecting, remote_atm, 
                           remote_name, remote_in6, qos, epe->res_e);
    } else
#endif
	call = new_atmcall(CurrVcc, 0, 0, conf, 1, connecting,
                               remote_atm, remote_name, remote_in6, qos, NULL);
    
    if (call == NULL) {
	return(-1);
    }
    if (conf == root) {
	party = new_party(call, (u_int16_t) 0, ps_add_init, remote_atm, 
			  remote_name, ept_e, mcl_e);
	call->nbactp = 1;
    }

    ret = send_atmsigd(CurrVcc, 0, ep_ref, as_connect, 0, 
		       remote_atm, &my_local_atm_addr, NULL, &call->qos);
    if (ret != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "atmcall_setup: failed to communicate with atmsigd");
	if (party != NULL)
	    free_party(party);
	free(call);
	return(-1);
    }

    /* creating a NATM socket for the new ATM call */
    if ((sd = new_natm_sock(call)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "atmcall_setup: failed to create a NATM socket, clearing call");
	(void) active_close_atmcall(0, CurrVcc, call);
	if (party != NULL)
	    free_party(party);
	if (call != NULL)
	    free(call);
	return(-1);
    }

    call->sd = sd;
    append_atmcall(call);
    NbCalls++;
    return(0);
}

PARTY *active_add_party(ATMCALL *call, struct svc_addr *remote_atm,
                        char *remote_name, void *ept_e, void * mcl_e)
{
    int i, ret;
    u_int16_t ep_ref;
    PARTY *party;

    ep_ref = call->parties->ep_ref; /* the last assigned ep_ref */
    if (++ep_ref == 0x7ff)
	ep_ref = 1;
    printf("\nREMOTE ADDR = ");
    for (i = 0; i < ATM_ESA_LEN; i++)
	printf("%02X", remote_atm->prv[i]);
    printf("\n");
    printf("<EPR = %d>\n", ep_ref);
    party = new_party(call, ep_ref, ps_add_init, remote_atm, remote_name,
                      ept_e, mcl_e);
    if (party == NULL) {
	diag(COMPONENT, DIAG_ERROR, 
	     "active_add_party: failed to allocate PARTY structure");
	return(NULL);
    }
    call->nbactp++;
    ret = send_atmsigd(call->vcc, 0, (int16_t) ep_ref, as_add, 0, 
		       remote_atm, &my_local_atm_addr, NULL, NULL);
    if (ret != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "active_add_party: failed to initiate adding party");
	free_party(party);
	return(NULL);
    }
    return(party);
}


int new_natm_sock(ATMCALL *call) 
{
    struct sockaddr_natm pvc = { sizeof(pvc), AF_NATM };
    int s;

    (void) snprintf(pvc.snatm_if, IFNAMSIZ, ENNAME, myitf);
    pvc.snatm_vpi = call->vpi;
    pvc.snatm_vci = call->vci;
  
    if ((s = socket(PF_NATM, SOCK_DGRAM, PROTO_NATMAAL5)) < 0) {
	perror("socket");
	return(-1);
    }
    return(s);
}

int active_drop_party(u_long vcc, u_int16_t ep_ref, PARTY *partyarg)
{

    int ret;
    PARTY *party;
    
    if ((party = partyarg) == NULL)
  	if ((party = party_lookup(vcc, ep_ref)) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "active_drop_party: no matching party for "
		 "<VCC = %ld, EPR = %d>", vcc, ep_ref);
	    return(-1);
	}
    if (party->call->conf == leaf) {
	diag(COMPONENT, DIAG_WARN, 
	     "call is in leaf conf, use CLOSE instead");
	return(0);
    }
    if ((ret = send_atmsigd(party->call->vcc, 0, party->ep_ref, as_drop,
			    0, NULL, NULL, NULL, NULL)) != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "active_drop_party: send_atmsigd failed");
	abort();
    }
    party->pstate = ps_drop_init;
    return(0);
}



int active_close_atmcall(int sig_down, u_long vcc, ATMCALL *callarg)  
{
    ATMCALL *call;

    /* delete ipatm route if any, and retrieve atmcall */
    call = delete_ipatmroute(vcc, 1, NULL, 0, callarg);
    if (call == NULL)
	return(-1);

    /* close ATM call itself */
    if (sig_down == 0) {
	if (send_atmsigd(call->vcc, 0, -1, as_close,
			 0, NULL, NULL, NULL, NULL) != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "active_close_atmcall: send_atmsigd failed");
	    abort();
	}
    }
    call->state = disconnecting;
    return(0);
}

int close_all(int sig_down)
{
    int error;
    ATMCALL *rm = AtmCalls;

    while (rm) {
	if ((error = active_close_atmcall(sig_down, 0, rm)) != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "close_all: failed to close all ATM calls");
	    return(error);
	}
	rm = rm->next;
    }
#ifdef MARS_CLIENT
    release_all_tables();
    mcl.mcl_state = mcl_none;
#endif
    return(0);
}

/* ------------------ route management on atm calls ------------------- */
#ifndef NDP_CLIENT
int add_ipatmroute(int is_host, 
		   struct in6_addr *rin6, 
		   int plen, 
		   ATMCALL *call,
		   u_int8_t rtm_op)
{
    int error;
    char v[64];
    struct in6_addr *ria6;

    (void) setuid(0);

    if ((ria6 = rin6) == NULL) {
	ria6 = &call->rin6;
	if (call->route != 0)  {
	    diag(COMPONENT, DIAG_WARN, 
		 "add_ipatmroute: route to destination %s already exists", 
		 inet_ntop(AF_INET6, ria6, v, sizeof(v)));
	    return(0);
	}
    }

    error = sendrtmsg(rtm_op, is_host, ria6, plen, call->vpi, call->vci);
    if (error != 0)
	diag(COMPONENT, DIAG_ERROR, 
	     "add_ipatmroute: failed to add route to destination %s: %s",
	     inet_ntop(AF_INET6, ria6, v, sizeof(v)), strerror(error));
    else { 
	diag(COMPONENT, DIAG_DEBUG, 
	     "Added route to destination %s", 
	     inet_ntop(AF_INET6, ria6, v, sizeof(v)));
	if (rin6 == NULL)
	    call->route = 1;
#ifdef MARS_CLIENT
	call->rt_count++;
#endif
    }

    return(error);
}
#endif

ATMCALL *delete_ipatmroute(u_long vcc, 
			   int is_host, 
			   struct in6_addr *rin6, 
			   int plen, 
			   ATMCALL *callarg)  
{

    int error;
    ATMCALL *call;
    char v[64];
    struct in6_addr *ria6;

    (void) setuid(0);
    if ((call = callarg) == NULL)
	if ((call = atmcall_lookup(vcc)) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "delete_ipatmroute: no matching atm call for VCC %ld", 
		 vcc);
	    return(NULL);
	}
    
#ifdef NDP_CLIENT
    /* notify ndp daemon if any neighbor entry is using this */
    if (call->ndp != NULL)
	closing_circuit(call, 0);
#else
    if ((ria6 = rin6) == NULL)
	ria6 = &call->rin6;

    if (call->route == 0) {
#ifdef MARS_CLIENT
	if ((rin6 == NULL) || (call->rt_count == 0))
	    return(call);		/* nothing to be done */
#else
	return(call);		/* nothing to be done */
#endif
    }

    error = sendrtmsg(RTM_DELETE, is_host, ria6, plen, call->vpi, call->vci);
    if (error != 0) 
	diag(COMPONENT, DIAG_ERROR, 
	     "delete_ipatmroute: failed to delete route to "
	     "destination %s: %s", 
	     inet_ntop(AF_INET6, ria6, v, sizeof(v)), strerror(error));
    else {
	diag(COMPONENT, DIAG_DEBUG, 
	     "Deleted route to destination %s", 
	     inet_ntop(AF_INET6, ria6, v, sizeof(v)));

#ifdef MARS_CLIENT    
	if (rin6 != NULL) {
	    call->rt_count--;
	    return(call);
	}
#endif
	call->route = 0;
    }
#endif
    return(call);
}

#ifndef NDP_CLIENT

#define ROUNDUP(a) \
((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n)	(x += ROUNDUP(n))

int sendrtmsg(int op, 
	      int is_host, 
	      struct in6_addr *rin6, 
	      int plen, 
	      u_int8_t vpi,
	      u_int16_t vci)
{
    struct {
	struct rt_msghdr m_rtm;
	char m_space[128];
    } m_rtmsg;
    static int s = -1;
    int len;
    char *cp;
    struct sockaddr_in6 *sin6;
    struct sockaddr_dl *sdl;
    char aph[4];
    static int seq = 1;
    char str_name[IFNAMSIZ + 1];
    unsigned int if_index;

    if (s < 0) {
	if ((s = socket(PF_ROUTE, SOCK_RAW, 0)) < 0)
	    return (errno);
    }
    (void)shutdown(s, 0);
    (void)bzero(&m_rtmsg, sizeof(m_rtmsg));

    bzero(str_name, IFNAMSIZ + 1);
    (void) snprintf(str_name, IFNAMSIZ, ENNAME, myitf);
    if_index = if_nametoindex(str_name);

    m_rtmsg.m_rtm.rtm_index = if_index;
    m_rtmsg.m_rtm.rtm_version = RTM_VERSION;
    m_rtmsg.m_rtm.rtm_type = op;
    if (is_host) {
	m_rtmsg.m_rtm.rtm_flags = RTF_UP | RTF_HOST;
	m_rtmsg.m_rtm.rtm_addrs = RTA_DST | RTA_GATEWAY;
    } else {
	m_rtmsg.m_rtm.rtm_flags = RTF_UP | RTF_STATIC;
	m_rtmsg.m_rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
    }
    m_rtmsg.m_rtm.rtm_pid = getpid();
    m_rtmsg.m_rtm.rtm_seq = seq++;

    cp = m_rtmsg.m_space;

    sin6 = (struct sockaddr_in6 *)cp;
    sin6->sin6_len = sizeof(*sin6);
    sin6->sin6_family = AF_INET6;
/*
 *    if (inet_pton(AF_INET6, callarg->ipaddr, &sin6->sin6_addr) <= 0) {
 *	printf("sendrtmsg: illegal host %s\n", callarg->rname);
 *	return (-1);
 *   }
 */
    bcopy(rin6, &sin6->sin6_addr, sizeof(struct in6_addr));
    ADVANCE(cp, sizeof(*sin6));

    sdl = (struct sockaddr_dl *)cp;
    sdl->sdl_len = sizeof(*sdl);
    sdl->sdl_family = AF_LINK;
    (void) strncpy(sdl->sdl_data, str_name, IFNAMSIZ);
    sdl->sdl_index = if_index;
    sdl->sdl_nlen = 3;

    sdl->sdl_alen = sizeof(aph);
    aph[0] = 3;
    aph[1] = vpi;
    aph[2] = vci >> 8;
    aph[3] = vci & 0xff;
    bcopy(aph,  LLADDR(sdl), sizeof(aph));
    ADVANCE(cp, sizeof(*sdl));

    if (!is_host) {
	sin6 = (struct sockaddr_in6 *)cp;
	sin6->sin6_len = sizeof(*sin6);
	sin6->sin6_family = AF_UNSPEC;
	memset(&sin6->sin6_addr, -1, plen);
	ADVANCE(cp, sizeof(*sin6));
    }

    if (cp > (char *)(&m_rtmsg + 1)) {
	diag(COMPONENT, DIAG_ERROR, "sendrtmsg: message overflow");
	return (-1);
    }

    m_rtmsg.m_rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

    if (write(s, (char *)&m_rtmsg, len) < 0)
	return (errno);
    return (0);
}
#endif
