/* iponatm.c - ip/atm traffic and route management */
/* Written 1998-1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/route.h>
#ifndef if_type
#include <net/if_var.h> 
#endif

#include <netinet/in.h>
#include <netinet/in6_var.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "ndp.h"

#include "atmcall.h"
#include "iponatm.h"
#include "scl_io.h"

#define PATHCONF "ip6atm.conf"	/* file containing IP6ATM addresses */

const char *AddrConf = PATHCONF;

int ipoveratm(int opcode, u_long vcc, int dump)
{
    int ret;
    ATMCALL *call;
    struct in6_ifreq ifr;
    struct sockaddr_in6 *s_in = NULL;
    struct in6_addr r_in6;
    struct hostent *hp;
    int s, on = 1;
    char str_addr[64];
  
    if ((opcode != NDP_ACTIVE) && (opcode != NDP_PASSIVE)) {
	printf("ERROR: ipoveratm: %d : unknown opcode for ipoveratm\n", opcode);
	return(-1);
    }
    if ((call= atmcall_lookup(vcc)) == NULL) {
	printf("\nERROR: ipoveratm: atmcall_lookup: no matching atm call for "
	       "VCC %ld\n", vcc);
	return(-2);
    }

    /* if client is not the call owner and is attempting an active add route */
    if ((opcode == NDP_ACTIVE) && (call->owner == 0)) {	
	printf("WARNING: ipoveratm: Not call owner (initiator)\n");
	return(0);
    }

    if (call->sd <= 0) {
	printf("ERROR: ipoveratm: no available native ATM socket for "
	       "ATM call\n");
	return(-3);
    }

    if (call->conf != p2p) {
	printf("WARNING: ipoveratm: : attempting to add route to p2m "
	       "call <VCC= %ld>.\nWARNING: Not implemented yet", vcc);
	return(0);
    }
    /* Retrieving IPv6 address of the ATM interface */
  

    bzero(&ifr, sizeof(ifr));
    (void) snprintf(ifr.ifr_name, IFNAMSIZ, ENNAME, myitf);
    if ((s = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
	perror("socket(inet)");
	return(-1);
    }
    ret = ioctl(s, SIOCGIFADDR6, (char *)&ifr);
    (void) close(s);
    if (ret < 0) {
	perror("SIOCGIFADDR6");
	return(ret);
    }
    s_in = (struct sockaddr_in6 *)&ifr.ifr_Addr;
  
/*
 * if active: test if source and destination have the same IPv6/ATM address.
 * If it's the case, recommend loopback!
 */
    if (opcode == NDP_ACTIVE) {
	hp = gethostbyname2(call->rname, AF_INET6);
	if (hp ==NULL) {
	    printf("%s: no such host or address\n", call->rname);
	    return(0);
	}
	bcopy(hp->h_addr, &r_in6, hp->h_length);
	if (bcmp(&r_in6, &s_in->sin6_addr, sizeof(struct in6_addr)) == 0) {
	    printf("\n !!! Source and destination have the same "
		   "IPv6/ATM address: ");
	    printf("%s", inet_ntop(AF_INET6, &s_in->sin6_addr, str_addr, 
				   sizeof(str_addr)));
	    printf("\nUSE LOOPBACK!!!");
	    printf("\nNo IPv6 route added to local host\n");
	    return(0);
	}
	bcopy(&r_in6, &call->rin6, sizeof(struct in6_addr));

/*
 * set the route then send a sollicitation.
 * atmndp() will wait for an advertisement sent by peer kernel
 * when the route will be set on the peer.
 */
	
	if ((ret = add_ipatmroute(1, NULL, 0, call, RTM_ADD)) != 0) {
	    printf("ERROR: ipoveratm: failed to add route to host %s\n", 
		   call->rname);
	    return(ret);
	}
	if ((ret = atmndp(opcode, s_in, &r_in6, dump)) != 0) {
	    printf("ERROR: ipoveratm: atmndp failed\n");
	    return(ret);
	}
	return(ret);
    } /* else if (opcode == NDP_PASSIVE) */

/*
 * if passive: set IP mode,
 * wait for a sollicitation with the peer address
 * then set the route (kernel will do advertisement(s).)
 */

    ret = setsockopt(call->sd, PROTO_NATMAAL5, NATM_IPMODE, &on, sizeof(int));
    if (ret != 0) {
	perror("setsockopt NATM_IPMODE");
	return(ret);
    }
    if ((ret = atmndp(opcode, s_in, &r_in6, dump)) != 0) {
	printf("ERROR: ipoveratm: atmndp failed\n");
	return(ret);
    }
    if (IN6_IS_ADDR_UNSPECIFIED(&r_in6)) {
	/* nothing changed in r_in6 structure */
	printf("ERROR: ipoveratm: haven't got peer IP address\n");
	return(-4);
    }
    hp = gethostbyaddr((char *) &r_in6, sizeof(r_in6), AF_INET6);
    if (hp == NULL) {
	printf("gethostbyaddr failed\n");
	return(0);
    }
    bcopy(&r_in6, &call->rin6, sizeof(r_in6));
    strncpy(call->rname, hp->h_name, NAME_LEN);
    if ((ret = add_ipatmroute(1, NULL, 0, call, RTM_ADD)) != 0)
	printf("ERROR: ipoveratm: failed to add route to host %s\n", 
	       call->rname);
    return(ret);
}

/* -------------------- Native ATM communication ------------------------- */

int recv_natm_peer(ATMCALL *call, int dump)
{
    int ret;

    if ((call->route != 0) || (call->conf != p2p)) {
	char buf[1024];

	ret = read(call->sd, buf, sizeof(buf));
	if (ret < 0) {
	    fprintf(stderr, "read on ATM vcc %ld failed: %s\n",
		    call->vcc, strerror(errno));
	    return(ret);
	}
	if (ret == 0) {
	    printf("ATM vcc %ld closed ?\n", call->vcc);
	    return(ret);
	}
	printf("flush %d bytes from ATM vcc %ld\n", ret, call->vcc);
	return(0);
    }
    if ((ret = ipoveratm(NDP_PASSIVE, call->vcc, dump)) != 0) {
	printf("recv_natm_peer: ipoveratm failed\n");
	return(ret);
    }
    return(0);
}
