/* ndp.c - ATM Neighbor Discovery tool */

/* Written 08-1998 by Francis.Dupont@inria.fr, INRIA */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#define STEVENS_API
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <errno.h>

#include "ndp.h"

void dodump(u_char *buf, u_char *buf2, int len)
{
	register struct ip6_hdr *ip = NULL;
	register struct nd_neighbor_solicit *nsp;
	struct timeval tv;
	struct tm *t;
	time_t clk;
	char v[64];

	if (!buf)
		nsp = (struct nd_neighbor_solicit *)buf2;
	else {
		ip = (struct ip6_hdr *)buf;
		nsp = (struct nd_neighbor_solicit *)(ip + 1);
		len -= sizeof(*ip);
	}

	(void) gettimeofday(&tv, NULL);
	clk = (time_t)tv.tv_sec;
	t = localtime(&clk);
	printf("%02d:%02d:%02d.%06u ",
	       t->tm_hour,
	       t->tm_min,
	       t->tm_sec,
	       (u_int)tv.tv_usec);

	switch (nsp->nd_ns_type) {
	    case ND_NEIGHBOR_SOLICIT:
		printf("request: ");
		break;
	    case ND_NEIGHBOR_ADVERT:
		printf("reply: ");
		break;
	}

	if (buf) {
		printf("src=%s ", inet_ntop(AF_INET6, &ip->ip6_src, v, 64));
		printf("dst=%s ", inet_ntop(AF_INET6, &ip->ip6_dst, v, 64));
	}
	printf("tgt=%s ", inet_ntop(AF_INET6, &nsp->nd_ns_target, v, 64));

	if (len == sizeof(struct nd_neighbor_solicit))
		printf("\n");
	else
		printf("...\n");
}

void dorequest(int s, struct in6_addr *peer, int dump)
{
	register int len, cc;
	struct nd_neighbor_solicit ns;

	len = sizeof(struct nd_neighbor_solicit);
	bzero(&ns, sizeof(ns));
	ns.nd_ns_type = ND_NEIGHBOR_SOLICIT;
	bcopy(peer, &ns.nd_ns_target, 16);

	cc = write(s, &ns, len);
	if (cc < 0) {
		perror("write(request)");
		exit(1);
	}
	if (cc != len) {
		fprintf(stderr,
			"truncated write(request) (%d != %d)\n",
			cc, len);
		exit(1);
	}
	if (dump) {
		printf("send request\n");
		dodump(NULL, (u_char *)&ns, len);
	}
}

int atmndp(int opcode, struct sockaddr_in6 *s_in, 
	   struct in6_addr *ia, int dump)
{
	register int s, cc;
	struct sockaddr_in6 peer;
	struct icmp6_filter filt;
	struct ip6_hdr *ip;
	struct nd_neighbor_solicit *nsp;
	u_char buf[1024];
	int request, reply;

	if (dump)
		setlinebuf(stdout);
	switch (opcode) {
	    case NDP_ACTIVE:
		request = 1;
		reply = 0;
		break;
	    case NDP_PASSIVE:
		request = 0;
		reply = 1;
		break;
	    default:
		printf("%d : bad ATM ND opcode\n", opcode);
		return(-1);
	}

	(void) setuid(0);

	s = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
	if (s < 0) {
		perror("socket ICMPv6");
		return(-1);
	}
	if (reply)
		(void) shutdown(s, 1);
	ICMP6_FILTER_SETBLOCKALL(&filt);
	if (request)
		ICMP6_FILTER_SETPASS(ND_NEIGHBOR_ADVERT, &filt);
	if (reply)
		ICMP6_FILTER_SETPASS(ND_NEIGHBOR_SOLICIT, &filt);
	cc = setsockopt(s, IPPROTO_ICMPV6, ICMP6_FILTER,
			&filt, sizeof(filt));
	if (cc < 0) {
		perror("setsockopt ICMPv6 filter");
		return(-1);
	}
	cc = bind(s, (struct sockaddr *)s_in, sizeof(*s_in));
	if (cc < 0) {
		perror("bind ICMPv6");
		return(-1);
	}
	bzero(&peer, sizeof(peer));
	peer.sin6_family = AF_INET6;
	peer.sin6_len = sizeof(peer);
	bcopy(ia, &peer.sin6_addr, 16);
	if (request)
		cc = connect(s, (struct sockaddr *)&peer, sizeof(peer));
	if (cc < 0) {
		perror("connect ICMPv6");
		return(-1);
	}

    retry:
	if (request) {
		fd_set set;
		struct timeval tv;

		dorequest(s, ia, dump);

		FD_ZERO(&set);
		FD_SET(s, &set);
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		cc = select(s + 1, &set, NULL, NULL, &tv);
		if (cc < 0) {
			perror("NDP select");
			if (errno == EINTR)
				goto retry;
			return(-1);
		}
		if (cc == 0) {
			printf("NDP timeout...\n");
			goto retry;
		}
		if (!FD_ISSET(s, &set)) {
			printf("NDP select botch\n");
			goto retry;
		}
	}

	cc = read(s, buf, sizeof(buf));
	if (cc < 0) {
		perror("read");
		return(-1);
	}
	if (cc == 0) {
		fprintf(stderr, "closed ?\n");
		return(-1);
	}

	ip = (struct ip6_hdr *)buf;
	nsp = (struct nd_neighbor_solicit *)(ip + 1);
	if (ip->ip6_plen + sizeof(*ip) != cc) {
		printf("IPv6 packet too short (%d != %d)\n",
		       ip->ip6_plen + sizeof(*ip), cc);
		goto retry;
	}
	if (ip->ip6_nxt != IPPROTO_ICMPV6) {
		printf("not an ICMPv6 packet (proto=%d)?\n", ip->ip6_nxt);
		return(-1);
	}
	if ((request && (nsp->nd_ns_type != ND_NEIGHBOR_ADVERT)) ||
	    (reply && (nsp->nd_ns_type != ND_NEIGHBOR_SOLICIT))) {
		printf("got bad ICMPv6 type %d\n", nsp->nd_ns_type);
		return(-1);
	}
	if (cc < sizeof(struct nd_neighbor_solicit)) {
		printf("NDP packet too short (%d)\n", cc);
		return(-1);
	}
	if (bcmp(&ip->ip6_dst, &s_in->sin6_addr, 16)) {
		char v1[64], v2[64];

		printf("NDP packet not for me (%s != %s)\n",
		       inet_ntop(AF_INET6, &ip->ip6_dst, v1, 64),
		       inet_ntop(AF_INET6, &s_in->sin6_addr, v2, 64));
		return(-1);
	}
	if (request && bcmp(&ip->ip6_src, ia, 16)) {
		char v1[64], v2[64];

		printf("NDP packet not from my peer (%s != %s)\n",
		       inet_ntop(AF_INET6, &ip->ip6_src, v1, 64),
		       inet_ntop(AF_INET6, ia, v2, 64));
		return(-1);
	}
	if (request && bcmp(&nsp->nd_ns_target, ia, 16)) {
		char v1[64], v2[64];

		printf("NDP packet target mismatch (%s != %s)\n",
		       inet_ntop(AF_INET6, &nsp->nd_ns_target, v1, 64),
		       inet_ntop(AF_INET6, ia, v2, 64));
		goto retry;
	}
	if (reply && bcmp(&nsp->nd_ns_target, &s_in->sin6_addr, 16)) {
		char v1[64], v2[64];

		printf("NDP packet not targeted to us (%s != %s)\n",
		       inet_ntop(AF_INET6, &nsp->nd_ns_target, v1, 64),
		       inet_ntop(AF_INET6, &s_in->sin6_addr, v2, 64));
		goto retry;
	}
	if (dump)
		dodump(buf, NULL, cc);
	if (reply)
		bcopy(&ip->ip6_src, ia, 16);
	return(0);
}
