/* scl_io.c - Signaling Client I/O operations */

/* Written 1997-1999 by Mohsen.Souissi@inria.fr, INRIA */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include <sys/time.h>
#include <sys/types.h>
#include <netdb.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmsap.h"
#include "atmd.h"
#include "localnsap.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#define COMPONENT "SCL_IO"


#define SERV_IP "0.0.0.0"
#define ATMSIG_PORT 6666


struct timeval now;
int sig_sd;


/* ---------------------------- addresses ------------------------------*/
/* we assume for the moment there is only one sap per interface */

struct svc_addr my_local_atm_addr;
struct in6_addr my_local_ipv6_addr;

struct svc_addr *get_addr(u_int8_t *aesa, u_int8_t sel)
{
    struct svc_addr *addr;
  
    addr = alloc(sizeof(struct svc_addr));
    *addr->pub = 0;
    bcopy(aesa, addr->prv, PREFIX_LEN + ESI_LEN);
    addr->prv[ATM_ESA_LEN -1] = sel; /* SEL zeroed */
    return(addr);
}

/* init local ATM and IPv6-ATM addresses */
void init_addr(int itf, u_int8_t sel)
{
    char str_addr[64];
    int error, i;

    get_local_addr(&my_local_atm_addr, 1, itf);

    my_local_atm_addr.prv[PREFIX_LEN + ESI_LEN] = 0x00 + sel;

    error = get_my_ipv6(&my_local_ipv6_addr, itf);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "init_addr: failed to get local ATM Interface IPv6 address");
	exit;
    }
    
    diag(COMPONENT, DIAG_INFO, "Signaling client's local addresses:");
    printf("\tATM:\t\t");
    for (i = 0; i < ATM_ESA_LEN; i++)
	printf("%02X", my_local_atm_addr.prv[i]);
	printf("\n");
    printf("\tIPv6-ATM:\t");
    printf("%s\n\n", inet_ntop(AF_INET6, &my_local_ipv6_addr, str_addr, 64));
}

/* ---------------------- signal handling ------------------------------ */

void sig_pipe(int signo)
{
    printf("SIGPIPE caught\n");
    exit(1);
}

/* ----------------------------- signaling messages -------------------- */

const char *as_name[] = { 
    "<invalid>",	"as_bind",	"as_connect",	"as_add",
    "as_accept",	"as_reject",	"as_listen",	"as_okay", 
    "as_error",		"as_indicate",	"as_drop",	"as_close",
    "as_itf_notify",	"as_modify",	"as_identify",	"as_terminate"
};



/* -------------------- Interface with atmsigd ------------------------- */

/* 
 * open a TCP socket to communicate with the signaling daemon (atmsigd) 
 * then register as signaling client 
 */
int register_with_atmsigd(void)   
{
    int s;
    struct sockaddr_in server_in;
    static long abuffer[sizeof(struct atmsvc_msg) / sizeof(long) + 1];
    u_int8_t *buffer = (u_int8_t *)abuffer;
    int size, error;
    struct atmsvc_msg *msg;
    struct atm_qos qos;
    

    if ((s = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	return (1);
    }
    memset(&server_in, 0, sizeof(server_in));
    server_in.sin_len = sizeof(server_in);
    server_in.sin_family = AF_INET;
    server_in.sin_addr.s_addr = inet_addr(SERV_IP);
    server_in.sin_port =  htons(ATMSIG_PORT);
    if (connect(s, (struct sockaddr *) &server_in, sizeof(server_in)) < 0) {
	perror("connect");
	return (1);
    }
    /* intializing sig_sd */
    sig_sd = s;

    /* let qos be compatible with every incoming qos */
    bzero(&qos, sizeof(qos));
    qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_ANYCLASS;
    qos.rxtp.max_sdu = qos.txtp.max_sdu = DEFAULT_MAX_SDU;
    qos.aal = ATM_AAL5;
    /* start registering with atmsigd */
    error = send_atmsigd(0, 0, -1, as_listen, 0, &my_local_atm_addr, 
			 NULL, NULL, &qos);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "register_with_atmsigd: failed to send as_listen\n");
	exit(error);
    }
    if ((size = read(sig_sd, buffer, sizeof(struct atmsvc_msg))) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "register_with_atmsigd: %s\n", strerror(errno));
	exit(-1);	
    }
    if (size == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "register_with_atmsigd: tcp_socket (%d) closed by peer\n", 
	     sig_sd);
	exit(-2);
    }
    if (size != sizeof(struct atmsvc_msg)) {
	diag(COMPONENT, DIAG_ERROR,
	     "register_with_atmsigd: atmsigd message has bad "
	     "length (got %d < %d)\n",
	     size, sizeof(struct atmsvc_msg));
	exit(-3);
    }
    msg = (struct atmsvc_msg *) buffer;
    switch (msg->type) {
    case as_okay:
	diag(COMPONENT, DIAG_DEBUG,
	     "register_with_atmsigd: Signaling client registered successfully");
	return(0);
    case as_error: 
	diag(COMPONENT, DIAG_ERROR, 
	     "register_with_atmsigd: Clearing end point\n");
	diag(COMPONENT, DIAG_ERROR, "register_with_atmsigd: failed");
	exit(-4);
    default:
	diag(COMPONENT, DIAG_ERROR, 
	     "Received %s message, unexpected\n", as_name[msg->type]);
	printf("register_sclient: Clearing end point\n");
	printf("register_with_atmsigd: failed");
	exit(-4);
    }
    return(0);
}

void deregister_with_atmsigd(void)   
{
    int error;

    error = send_atmsigd(0, 0, -1, as_close, 0, NULL, NULL, NULL, NULL);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "deregister_with_atmsigd: failed to send <as_close> to sigd\n");
	exit;
    }
    return;
}


int recv_atmsigd(enum atmsvc_msg_type *type)
{
    static long abuffer[sizeof(struct atmsvc_msg) / sizeof(long) + 1];
    u_int8_t *buffer = (u_int8_t *)abuffer;
    int size, ret;

    if ((size = read(sig_sd, buffer, sizeof(struct atmsvc_msg))) < 0) {
	printf("ERROR:  recv atmsigd: %s\n", strerror(errno));
	return(-1);	
    }
    if (size == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "recv atmsigd: tcp_socket (%d) closed by peer\n", 
	       sig_sd);
	return(-2);
    }
    if (size != sizeof(struct atmsvc_msg)) {
	diag(COMPONENT, DIAG_ERROR,
	     "recv atmsigd: atmsigd message has bad length (got %d < %d)\n",
	       size, sizeof(struct atmsvc_msg));
	return(-3);
    }
    ret = from_atmsigd((struct atmsvc_msg *) buffer, type);
    if (ret < 0)
	return(-4);
    return(ret);
}

int send_atmsigd(u_long vcc, u_long listen_vcc, int16_t ep_ref,  
		 enum atmsvc_msg_type type, int reply, 
		 const struct svc_addr *svc, const struct svc_addr *local, 
		 const struct atm_sap *sap, const struct atm_qos *qos)
{
    struct atmsvc_msg *msg;
    int ret;

    msg = alloc(sizeof(struct atmsvc_msg));
    memset(msg, 0, sizeof(*msg));
    msg->vcc = vcc;  
    msg->listen_vcc = listen_vcc;
    msg->ep_ref = ep_ref;
    msg->type = type;
    msg->reply = reply;
    memset(&msg->pvc, 0, sizeof(msg->pvc));
    if (qos != NULL) 
	msg->qos = *qos;
    if (local != NULL)
	msg->local = *local;
    if (svc != NULL) 
	msg->svc = *svc;
    if (sap == NULL)
	msg->sap.bhli.hl_type = ATM_HL_NONE;
    else
	msg->sap = *sap;
    ret = to_atmsigd(msg);
    free(msg);
    return(ret);
}

int to_atmsigd(struct atmsvc_msg *msg)
{
    int wrote;
   
    diag(COMPONENT, DIAG_DEBUG,
	 "TO ATMSIGD  >>>---------------->>>");
    diag(COMPONENT, DIAG_DEBUG,
	 "\t%s, <reply %d>, <vcc %ld, listen_vcc %ld> <EPR = %d>",
	 as_name[msg->type], 
	 msg->reply, msg->vcc, msg->listen_vcc, msg->ep_ref);
    if ((wrote = write(sig_sd, msg, sizeof(*msg))) == sizeof(*msg))
	return(0);
    if (wrote < 0) {
	fprintf(stderr, "to_atmsigd write: %s", strerror(errno));
	if (errno == EPIPE)
	    printf("ERROR:  atmsigd is down\n");
	return(wrote);
    }
    diag(COMPONENT, DIAG_ERROR,
	 "to_atmsigd: bad atmsigd write: wanted %d, wrote %d\n",
	   sizeof(*msg), wrote);
    return(wrote);
}


/* ----- common part ------------------------------------------------------- */

void init_current_time(void)
{
    (void) gettimeofday(&now, NULL);
}
