/* test.c - test signaling client */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <sys/wait.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "atm.h"


#include "ndp.h"
#include "ipatmlookup.h"
#include "atmcall.h"
#include "iponatm.h"
#include "scl_io.h"

#define COMPONENT	"TEST"

int debug = 0, ndp_dump;
int myitf = 0;

const char *sig_qos = NULL;
const char *dump_dir = NULL;


/* ------------------------------- main ...  ------------------------------- */

static void usage(const char *name)
{
    fprintf(stderr,
	    "usage: %s [ -s selector ] \n", name);
    exit(1);
}

static int main_loop(void)
{
    int ret = 0, fin = 0, fds;
    pid_t pid;
    fd_set set, perm;
    ATMCALL *call;
    int sig_down = 0;
    enum atmsvc_msg_type type;
    
    if (signal(SIGPIPE, sig_pipe) == SIG_ERR)
	printf("sig_pipe: error\n");
  
    FD_ZERO(&perm);
    FD_SET(sig_sd, &perm);
    fds = FD_SETSIZE;
    (void) gettimeofday(&now, NULL);

    if ((ret = socketpair(AF_UNIX, SOCK_STREAM, 0, psd)) < 0) {
	printf("main_loop: failed to create a socketpair\n");
	return(-1);
    }
    if ((pid = fork()) < 0) {
	fprintf(stderr, "fork error: %s\n", strerror(errno));
	exit(1);
    }
    if (pid == 0) {		/* child */
	(void) close(sig_sd);	/* useless for child */
	if ((fin = user_input()) == 1) {
	    printf("Child terminating session, exiting normally\n");
	    exit(0);
	}
	exit(1);
    }
    if (pid > 0) {		/* parent */
	while (fin <= 0) {
	    set = perm;
	    FD_SET(psd[0], &set);
	    for (call = AtmCalls; call; call =  call->next)
                if (call->sd > 0)
                    FD_SET(call->sd, &set);
	    ret = select(fds, &set, NULL, NULL, next_timer());
	    if (ret < 0) {
		if (errno != EINTR) {
		    perror("select");
		    if (kill(pid, SIGTERM) < 0) /* kill the child  */
			perror("kill");
		    abort();
		}
		continue;
	    }

            (void) gettimeofday(&now, NULL);

	    if (FD_ISSET(psd[0], &set))
		fin = process_cmd();
	    if (FD_ISSET(sig_sd, &set)) {
		if ((ret = recv_atmsigd(&type)) < 0) {
		    if (ret == -2) { 
			/* sigd has closed its tcp socket with client */
			sig_down = 1;
			printf("\n\nmain_loop: !!! CONNECTION WITH ATMSIGD "
			       "IS DOWN !!!\n");
			printf("TEARING DOWN CURRENT ATM CALLS\n");
			(void) close_all(sig_down);
			(void) close(sig_sd);
			if (kill(pid, SIGTERM) < 0) /* kill the child  */
			    perror("kill");
		    }
		    return(-2);
		}
	    }
	    for (call = AtmCalls; call; call =  call->next) {
                if ((call->sd > 0) && (FD_ISSET(call->sd, &set))) {
                    if ((ret = recv_natm_peer(call, ndp_dump)) != 0) {
                        printf("recv_natm_peer: failed to read message "
                               "from peer\n");
                        /* return(-1); */       /* Don't return here */
                    }
                }
            }
	    /* expire timers after handling messges to make sure we don't
               time out unnecessarily because of scheduling delays */
            expire_timers();

	}
    }
    return(0);
}

int main(int argc, char **argv)
{
    int c, error;
    char *end;
    long long_sel;
    u_int8_t selector;


    ndp_dump = 1;
    set_application("SCLIENT");
    dump_dir = NULL;

    set_verbosity(NULL, DIAG_DEBUG);
    /* initialize default selector value */
    selector = 0x00;

    if (argc > 1) {
	while ((c = getopt(argc, argv, "s:")) != EOF)
	    switch (c) {
	    case 's':		/* choose a selector for local ATM address */
		long_sel = strtol(optarg, &end, 0);
		if (*end)
		    usage(argv[0]);
		if (long_sel > 0xff) {
		    diag(COMPONENT, DIAG_ERROR,
			 "Bad selector value!  0x00 <= selector <= 0xff");
		    exit(-1);
		}
		selector = (u_int8_t) long_sel;
		break;
	    default:
		usage(argv[0]);
	    }
    }
    
    diag(COMPONENT, DIAG_INFO, "\nATM signaling client "
#ifdef UNI30
	 "UNI 3.0"
#endif
#ifdef UNI31
	 "UNI 3.1"
#ifdef ALLOW_UNI30
	 "+3.0compat"
#endif
#endif
#ifdef UNI40
	 "UNI 4.0"
#ifdef Q2963_1
	 "+Q.2931.1"
#endif
#endif
	 "/AAL5, version " VERSION "\n");
  
    init_current_time();
    init_addr(myitf, selector);
    CurrVcc = 0;

    if ((error = register_with_atmsigd()) != 0) {
	diag(COMPONENT, DIAG_ERROR, "main: failed to register with atmsigd");
	exit(error);
    }
    diag(COMPONENT, DIAG_DEBUG, "\tWaiting for incoming messages.....");
    error = main_loop();
    (void) close(sig_sd);
    return (error);
}

